/*
 * xtcp_mod_cyc.c
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * for any purpose with or without fee is hereby granted, provided
 * that the above copyright notice and this permission notice
 * appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE UNIVERSITY OF UTAH
 * DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING
 * ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO
 * EVENT SHALL THE UNIVERSITY OF UTAH BE LIABLE FOR ANY SPECIAL,
 * DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER
 * IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
 * THIS SOFTWARE.
 */

/*
 * Kernel module entry point for cyclone protocols.
 *
 * Imports:
 *
 *  Cyc_xtcp_sw - The tfi_protosw_t containing the protocol entry points.
 *  KMOD_HASH - A preprocessor symbol that maps to the protocol hash in hex.
 *              (Should be in the kmod_data.h file generated by the Makefile)
 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/module.h>
#include <sys/conf.h>

#include <netinet/xtcp/tfi.h>
#include <netinet/xtcp/tfc.h>
#include <netinet/xtcp/tfci.h>

#include "kmod_data.h"

#define CYCLONE_SYM(sym) Cyc_##sym
#define XCYCLONE_SYM(sym) CYCLONE_SYM(sym)

extern struct tfi_protosw_cyc XCYCLONE_SYM(xtcp_sw);

static struct tfi_proto xtcp_pr;

static int
xtcp_modevent(module_t mod, int type, void *data)
{
	int retval = 0;
	
	switch (type) {
	case MOD_LOAD:
		printf("\nXTCP protocol: " __XSTRING(KMOD_HASH) "\n");
		/*
		 * Add the protocol and have it initialized (done by
		 * tfi_addproto_cyc).
		 */
		retval = tfi_addproto_cyc(IPPROTO_TFC,
					  KMOD_HASH,
					  &XCYCLONE_SYM(xtcp_sw),
					  &xtcp_pr);
		printf("xtcp_init(NULL) = %d\n", retval);
		break;

	case MOD_UNLOAD:
#if !defined(KLD_MODULE)
		printf("xtcp proto tcp statically compiled, cannot unload\n");
		return EINVAL ;
#else
		printf("tfi_delproto(%p)\n", &xtcp_pr);
		retval = tfi_delproto(&xtcp_pr);
		printf("%d\n", retval);
#endif
		break ;
	default:
		break ;
	}
	return retval;
}

static moduledata_t xtcp_proto_cyctcp_mod = {
	"xtcp_proto_" __XSTRING(KMOD_HASH),
	xtcp_modevent,
	NULL
};
DECLARE_MODULE(xtcp_proto, xtcp_proto_cyctcp_mod, SI_SUB_PROTO_DOMAIN, SI_ORDER_ANY);
MODULE_VERSION(xtcp_proto, 1);
