/*
 * scoreboard.h
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * for any purpose with or without fee is hereby granted, provided
 * that the above copyright notice and this permission notice
 * appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE UNIVERSITY OF UTAH
 * DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING
 * ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO
 * EVENT SHALL THE UNIVERSITY OF UTAH BE LIABLE FOR ANY SPECIAL,
 * DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER
 * IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
 * THIS SOFTWARE.
 */
#ifndef TFC_SCOREBOARD_H
#define TFC_SCOREBOARD_H

#define SBSIZE 2048
/*
 * TCP sequence numbers are 32 bit integers operated
 * on with modular arithmetic.  These macros can be
 * used to compare such integers.
 */

typedef u_int32_t seq_t;

#define	TFC_SEQ_LT(a,b)	  ((int)((a)-(b)) < 0)
#define	TFC_SEQ_LEQ(a,b)  ((int)((a)-(b)) <= 0)
#define	TFC_SEQ_GT(a,b)	  ((int)((a)-(b)) > 0)
#define	TFC_SEQ_GEQ(a,b)  ((int)((a)-(b)) >= 0)

#define SB_DIFF(a, b)     (((a) + (SBSIZE+1) - (b)) % (SBSIZE+1))
#define SB_INC(a) (((a)+1) % (SBSIZE+1))
#define SB_DEC(a) SB_DIFF((a), 1)

/* Pass it the ECT code points and this function gets one bit nonce */
#define NONCE_SUM(a, b) (((a)+(b)) & 0x03)

#define PPL_INC(p) (((p) + 1) % (WEIGHTS + 1))
#define PPL_DEC(p) ((p) == 0 ? WEIGHTS : ((p)-1))

#define SBSTATE_SENT   0x01
#define SBSTATE_RESENT 0x02
#define SBSTATE_ACKED  0x04
#define SBSTATE_LOST   0x08

struct sbnode {
	seq_t seq_no;
	seq_t bseq;
	seq_t eseq;
	u_char nonce;
	tick_t ts; /* When the packet was sent */
	u_char state;
#ifdef TFC_CUT
	u_char nonce_sum;	
	u_char resent;
	u_char valid;
	tick_t to; /* when does this entry timeout */
	tick_t rtt;
#endif
};

struct tfc_cb;

extern int sb_insert(seq_t n, u_char nonce, struct tfc_cb* tp, tick_t ts);
extern int sb_matchsum(int eseq, u_char nonce_sum, tick_t ts, struct tfc_cb *tp);
extern int sb_lookup(seq_t seq_no, struct tfc_cb *tp);
extern int register_loss(seq_t seq_no, struct tfc_cb *tp, int resend);
extern void sb_drain(struct tfc_cb *tp, tick_t ts);
extern void sb_cleanup(struct tfc_cb *tp);
extern int sb_update(seq_t seq_no, tick_t ts, struct tfc_cb *tp, u_char *nonce);

extern int register_loss_leq(seq_t seq_no, struct tfc_cb *tp, int resend);
extern int sb_matchsum_leq(int eseq, u_char nonce_sum, tick_t ts, struct tfc_cb *tp);
extern int sb_lookup_leq(seq_t seq_no, struct tfc_cb *tp);
extern int sb_update_leq(seq_t seq_no, tick_t ts, struct tfc_cb *tp, u_char *nonce);

/* Range functions */
extern int sb_insert_range(struct tfc_cb* tp, seq_t bseq, seq_t eseq, u_char nonce, tick_t ts);
extern int sb_matchsum_range(struct tfc_cb* tp, seq_t bseq, seq_t eseq, u_char nonce_sum, tick_t ts);
extern int register_loss_range(struct tfc_cb *tp, seq_t bseq, seq_t eseq, int resend);
extern int sb_lookup_range(struct tfc_cb *tp, seq_t bseq, seq_t eseq, int *bidx, int *isbeg,
			   int *eidx, int *isend);
extern int sb_advance(struct tfc_cb *tp, seq_t eseq, int idx, int isboundary);
extern int sb_update_range(struct tfc_cb *tp, seq_t bseq, seq_t eseq, tick_t ts, u_char *nonce);


#endif /* TFC_SCOREBOARD_H */
