/*
 * tfc_cb.c
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * for any purpose with or without fee is hereby granted, provided
 * that the above copyright notice and this permission notice
 * appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE UNIVERSITY OF UTAH
 * DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING
 * ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO
 * EVENT SHALL THE UNIVERSITY OF UTAH BE LIABLE FOR ANY SPECIAL,
 * DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER
 * IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
 * THIS SOFTWARE.
 */
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/malloc.h>
#include <sys/kernel.h>
#include <sys/sysctl.h>
#include <sys/mbuf.h>
#include <sys/socket.h>
#include <sys/socketvar.h>
#include <sys/protosw.h>
#include <sys/proc.h>
#include <sys/jail.h>
#include <sys/queue.h>

#include <vm/vm_zone.h> /* memory for tfc_cb and the likes */

#include <net/if.h>
#include <net/route.h>

#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/in_pcb.h>
#include <netinet/in_var.h>
#include <netinet/ip_var.h>
#include <netinet/tcp.h>
#include <netinet/tcp_fsm.h>
#include <netinet/tcp_seq.h>
#include <netinet/tcp_timer.h>
#include <netinet/tcp_var.h>
#include <netinet/tcpip.h>

#include <netinet/xtcp/tfc.h>
#include <netinet/xtcp/tfci.h>
#include <netinet/xtcp/tfi.h>
#include <netinet/xtcp/tfc_stats.h>
#include <netinet/xtcp/tfc_error.h>
#include <netinet/xtcp/lossfraction.h>
#include <netinet/xtcp/scoreboard.h>
#include <netinet/xtcp/tfc_cb.h>
#include <netinet/xtcp/tfi_pcb.h>

MALLOC_DEFINE(M_TFCMBUFQ, "tfc_mbufs", "TFC protocol mbufs");

struct tfc_cb *
tfc_newtfccb(struct inpcb *inp)
{
	struct inp_xtcb *ix;
	register struct tfc_cb *tp;
	struct sbnode *sb;

	MALLOC(sb, struct sbnode *, sizeof (struct sbnode) * (SBSIZE + 1), M_TFCPROTO,
	       M_ZERO | M_NOWAIT);

	if (NULL == sb) {
		printf("Unable to allocate memory\n");
		return NULL;
	}

	ix = (struct inp_xtcb *)inp;
	tp = &ix->xtcb;
	bzero((char *) tp, sizeof(struct tfc_cb));

	tp->inpcb = inp;
	tp->scoreboard = sb;
	tp->checknow = CHECKFREQ;
	callout_init(tp->tt_2msl = &ix->inp_xtcb_2msl);
	callout_init(tp->tt_nofeed = &ix->inp_xtcb_nofeed);

	/*
	 * initialization of variables related to the equation
	 */
	tp->rtt_ignore = INIT_RTT_IGN;
        /*
	 * IPv4 TTL initialization is necessary.
	 */
	inp->inp_ip_ttl = ip_defttl;

	/*
	 * Association from the inpcb to tfc_cb
	 */
	inp->inp_ppcb = (caddr_t)tp;
	return (tp);
}
