/*
 * tfc_cb.h
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * for any purpose with or without fee is hereby granted, provided
 * that the above copyright notice and this permission notice
 * appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE UNIVERSITY OF UTAH
 * DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING
 * ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO
 * EVENT SHALL THE UNIVERSITY OF UTAH BE LIABLE FOR ANY SPECIAL,
 * DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER
 * IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
 * THIS SOFTWARE.
 */
#ifndef TFC_CB_H
#define TFC_CB_H

/* declare common variables */
#define	ALIGNMENT	32
#define	ALIGNM1		(ALIGNMENT - 1)

/* Congestion control states */
#define TFCS_INIT          0
#define TFCS_ATTACHED      1
#define TFCS_INITRTT       2
#define TFCS_SLOWSTART     3
#define TFCS_STEADYSTATE   4
#define TFCS_DISCONNECTING 5

/* Connection states */
#define TFCS_CONN_UNCONNECTED  0
#define TFCS_CONN_LISTENING    1
#define TFCS_CONN_USESENT      2
#define TFCS_CONN_USERCVD      3
#define TFCS_CONN_DONE         4

/* Protocol MBUF queue entry */
struct mbufq_ent {
	LIST_ENTRY(mbufq_ent) next;
	struct	mbuf	*m;		/* mbuf contains packet */
};

LIST_HEAD(mbufq_head, mbufq_ent);
MALLOC_DECLARE(M_TFCMBUFQ);

#define CHECKFREQ 8 /* we check for exact rate comliance only every 8 rtts,
		       very very liberal */
#define RTT_IGN_FACTOR 10 /* ignore extreme abberations in rtt */
#define MAX_VIOLATIONS 8
#define INIT_RTT_IGN 2
#define MAX_INIT_ALLOWED 4
#define MAX_INIT_INTERVAL 400 /* msec */
#define TFC_OPEN_SS 4 /* max slow start window size */
#define RTT_DECAY_SHIFT 3

#define INIT_NOFEED_TIMER 2000 /* msec. Initial timeout value for the timer */

struct tfc_cb {

	u_char state;
	/*
	 * parameters of the TCP rate control equation
	 */
	tick_t rtt; /* current estmate of rtt */
	tick_t rto; /* Timeout */
	u_char rtt_ignore;
	
	tick_t prev_ts; /* when was the prev rtt round */
	
	int rate_eq; /* current send rate as calculated by the equation */
	int rate_eqsum; /* Summation of the allowed sending rate */
	int rate_act; /* current actual send rate */
	int ack_prev; /* acks received in the previous rtt */
	int ack_cur;
	int sent_cur;
	int sent_prev;

        int slow_start_credits;
	int checknow; /* when was the last time, a check was done */
	int violations; /* how many violations has the protocol done */

	int t_nofeed; /* timeout before the no feeback timer expires */
	int ppl[WEIGHTS + 1]; /* WEIGHTS packets per loss intervals + last open
				       interval */
	int ppl_open; /* The open loss interval */

	seq_t last_ack_seq; /* the last packet acked */
	tick_t last_ack_ts; /* when did we see last acksum */
	/* an rto after this a loss should be assumed to have happened */

	seq_t prev_le_seq; /* seq no covered by prev loss event */
	tick_t prev_le_ts; /* time stamp that the prev loss event covers */

	int prev_le_idx; /* index of the previous loss event in the scoreboard */

	/* struct sbnode scoreboard[SBSIZE+1];*/ /* per-packet data such as nonce */
	struct sbnode *scoreboard; /* per-packet data such as nonce */
	int sb_lastplus1;  /* Next open entry in the scoreboard */
	int sb_first; /* first valid entry in the scoreboard */
	
	u_char nonce_sum;  /* nonce sum up to nonce_sum_seq */
	u_char nonce_sumseq; /* nonce sum is up to this seq number */
	/* sb_first should point to nonce_sumseq */
	u_char nonce_sumsynched; /* is noncesum synched with the sender/receiver */
	
	/* Related to normal working of the protocol */
	struct tfc_protodomain *dm;
	struct tfc_proto *xp; /* protocol that actually does something */
	struct tfi_protosw *psw; /* invoke functions, duplicated here for speed */
	struct tfi_protosw_cyc *cyc_psw; /* invoke functions, duplicated here for speed */
	struct callout *tt_2msl; /* Timeout to delete a connection after the user
				   has issued a "close" */
	struct callout *tt_nofeed; /* No feedback timer */
	struct mbufq_head mbufq; /* List of mbufs in use by this protocol */
	int t_softerror; /* some error */
	struct inpcb *inpcb; /* back pointer */
	void *pcb; /* A pointer to the user protocol CB */

	u_char conn_state; /* what state the connection is in */
	struct tfc_proto *requested_proto; /* the protocol requested by the user.
					      This protocol is not into effect until a
					      connection is established with the same
					      protocol on the other end.
					   */
};

/* This is what is actually allocated when we allocate a inpcb */
/* saves an extra allocation call */
struct inp_xtcb {
	union {
		struct inpcb *ip;
		char	align[(sizeof(struct inpcb) + ALIGNM1) & ~ALIGNM1];
	} inp;
	struct tfc_cb xtcb;
	struct callout inp_xtcb_2msl;
	struct callout inp_xtcb_nofeed;
};

#define intotfccb(in)   ((struct tfc_cb*)(in)->inp_ppcb)
#define sototfccb(so)   (intotfccb(sotoinpcb(so)))

struct tfc_cb * tfc_newtfccb(struct inpcb *inp);

#endif /* TFC_CB_H */
