/*
 * tfi.c
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * for any purpose with or without fee is hereby granted, provided
 * that the above copyright notice and this permission notice
 * appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE UNIVERSITY OF UTAH
 * DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING
 * ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO
 * EVENT SHALL THE UNIVERSITY OF UTAH BE LIABLE FOR ANY SPECIAL,
 * DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER
 * IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
 * THIS SOFTWARE.
 */
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/malloc.h>
#include <sys/kernel.h>
#include <sys/sysctl.h>
#include <sys/mbuf.h>
#include <sys/socket.h>
#include <sys/socketvar.h>
#include <sys/protosw.h>
#include <sys/proc.h>
#include <sys/jail.h>
#include <sys/queue.h>
#include <sys/errno.h>

#include <vm/vm_zone.h> /* memory for tfc_cb and the likes */

#include <net/if.h>
#include <net/route.h>

#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/in_pcb.h>
#include <netinet/in_var.h>
#include <netinet/ip_var.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <netinet/tcp_fsm.h>
#include <netinet/tcp_seq.h>
#include <netinet/tcp_timer.h>
#include <netinet/tcp_var.h>
#include <netinet/tcpip.h>

#include <netinet/xtcp/tfc.h>
#include <netinet/xtcp/tfci.h>
#include <netinet/xtcp/tfi.h>
#include <netinet/xtcp/tfc_stats.h>
#include <netinet/xtcp/tfc_error.h>
#include <netinet/xtcp/lossfraction.h>
#include <netinet/xtcp/scoreboard.h>
#include <netinet/xtcp/tfc_cb.h>

/* debug */
/* #define TFC_DEBUG 1 */

#define FAST_SS

/* globals */
MALLOC_DEFINE(M_TFCPROTO, "tfc_proto", "TFC protocol entry");

/* All the protocols that want to look like TCP, go here. */
#define XTCP_DEFAULT_HASH 170 /* hash of the default protocol */
struct tfc_protodomain tfc_tcpd = { TFCCB_HASHTABLE_SIZE, 0, IPPROTO_XTCP,
				    { 0 }, { 0 }, NULL, NULL, XTCP_DEFAULT_HASH,
				    NULL, 1 };

/* This domain should contain all the protocols that are not TCP compatible */
struct tfc_protodomain tfc_defaultd = { TFCCB_HASHTABLE_SIZE, 0,
					IPPROTO_TFC, { 0 }, { 0 },
					NULL, NULL, 0, NULL, 0 };

struct tfc_protodomain *tfc_domains[TFC_NUM_DOMAINS] = {&tfc_tcpd,
						       &tfc_defaultd};

/* ECN code points */
#define ECT0 0x01 
#define ECT1 0x02
static u_char ect[2] = { ECT0, ECT1 };

static int tfii_send(struct mbuf *m, struct inpcb *inp, u_char ect);
static int rate_check(struct tfc_cb *tp, tick_t ts);
static void update_rate(struct tfc_cb *tp);
static void per_rtt_bookkeeping (struct tfc_cb* tp) ;

#ifdef TFC_LATER
static void tfc_timer_nofeed(void *vtp);
#endif
static int round;

/* Cyclone fat pointer. This should move into cmap.c file */
#include <netinet/xtcp/xudp.h>
#include <netinet/xtcp/xudp_var.h>

struct _dyneither_ptr {
  unsigned char *curr; 
  unsigned char *base; 
  unsigned char *last_plus_one; 
};  
struct ip *tfi_mtoip(struct mbuf *m);
extern struct _dyneither_ptr tfi_mtodfat(struct mbuf *m);
struct xudphdr *tfi_strtoxudphdr(struct _dyneither_ptr data);
struct xtcphdr;
struct xtcphdr *tfi_strtoxtcphdr(struct _dyneither_ptr data);
struct xudpiphdr *tfi_strtoxudpiphdr(struct _dyneither_ptr data);
struct ipovly *tfi_strtoipovly(struct _dyneither_ptr data);
struct xtcptemp; /* size 64 */
struct xtcptemp *tfi_mtoxtcptemp(struct mbuf *m);
extern struct in_addr tfi_inpladdr(struct inpcb *inp);
extern struct in_addr tfi_inpfaddr(struct inpcb *inp);
extern u_int16_t tfi_inplport(struct inpcb *inp);
extern u_int16_t tfi_inpfport(struct inpcb *inp);
struct pkthdr* tfi_m_pkthdr(struct mbuf *m);
int tfi_msetdata(struct mbuf *m, char *data);
extern u_char tfi_inpttl(struct inpcb *inp);
extern u_char tfi_inptos(struct inpcb *inp);
int tfi_in_cksum(struct mbuf *m, int len);
u_short tfi_in_pseudo(u_int32_t a, u_int32_t b, u_int32_t c);
/* Cyclone runtime support */
extern void *tfi_calloc(int size, int mult);
extern void tfi_free(void *ptr);
extern void tfi_exit(int status);
extern u_int32_t tfi_htonl(u_int32_t x);
extern u_int16_t tfi_htons(u_int16_t x);
extern u_int32_t tfi_ntohl(u_int32_t x);
extern u_int16_t tfi_ntohs(u_int16_t x);
extern int tfi_printf(const char *);
struct rt_metrics *tfi_getrtmx(struct rtentry *rte);
int tfi_mhlen = MHLEN;

/* BEGIN - Parveen */
extern int tfi_soshiwat(struct socket *so);
struct rtentry *tfi_rtlookup(struct in_conninfo *inc);
int tfi_isrtup(struct rtentry *rt);
int tfi_issolinger(struct socket *so);
void tfi_setsolinger(struct socket *so, int period);
void tfi_sbrflush(struct socket * so);
void tfi_soisdisconnecting(struct socket * so);
void tfi_socantsendmore(struct socket * so);
int tfi_sogetscc(struct socket *so);
int tfi_sogetrcc(struct socket *so);
int tfi_in_pcbladdr(struct inpcb *inp, struct sockaddr *nam,
		    struct sockaddr_in **ifaddr);
struct inpcb *tfi_in_pcblookup_hash(struct inpcb *inp, struct in_addr faddr, u_int fport,
				    struct in_addr laddr, u_int lport, u_int wildcard,
				    struct ifnet *ifp);
void tfi_in_pcbrehash(struct inpcb *inp);
void tfi_soisconnecting(struct socket * so);
int tfi_sorhiwat(struct socket *so);
void tfi_soisdisconnected(struct socket * so);
short tfi_getsostate(struct socket * so);
short tfi_mflags(struct mbuf *m);
int tfi_in_broadcast(struct in_addr in, struct ifnet *ifp);
extern void tfi_sbsreserve(struct socket *so, u_long cc);
extern void tfi_sbrreserve(struct socket *so, u_long cc);
extern int tfi_rtgetmtu(struct rtentry *);
struct mbuf *tfi_mgetnext(struct mbuf *);
void tfi_socantrcvmore(struct socket *so);
void tfi_sohasoutofband(struct socket *so);
void tfi_sbsdrop(struct socket *so, int len);
int tfi_in_localaddr(struct in_addr in);
extern struct _dyneither_ptr tfi_m_copydata(struct socket *so, int off, int len,
					   struct _dyneither_ptr data);
extern struct mbuf *tfi_m_copy(struct socket *so, int off, int len);
extern void tfi_m_setnext(struct mbuf *, struct mbuf *);
extern u_short  tfi_in_addword(u_short, u_short);

/* END - Parveen */

short tfi_sogetoptions(struct socket *so);
struct mbuf *tfi_xtcptemptom(struct xtcptemp *);
long tfi_getrt_refcnt(struct rtentry *rte);
u_long tfi_getrt_flags(struct rtentry *rte);
struct sockaddr *tfi_getrt_gateway(struct rtentry *rte);
void tfi_rt_free(struct rtentry *rte);
struct sockaddr *tfi_rt_key(struct rtentry *rte);
struct sockaddr *tfi_rt_mask(struct rtentry *rte);
/* struct socket *tfi_sonewconn(struct socket *, int); */
void tfi_soabort(struct socket *so);

extern int Cyc_pru_abort(struct socket * so, struct tfi_protosw_cyc * psw);
extern int Cyc_pru_accept(struct socket * so, struct sockaddr * * nam, struct tfi_protosw_cyc * psw);
extern int Cyc_pru_attach( struct socket * so, int proto, struct proc * p, struct tfi_protosw_cyc * psw);
extern int Cyc_pru_connect( struct socket * so, struct sockaddr * nam, struct proc * p, struct tfi_protosw_cyc * psw);
extern int Cyc_pru_detach( struct socket * so, struct tfi_protosw_cyc * psw);
extern int Cyc_pru_disconnect( struct socket * so, struct tfi_protosw_cyc * psw);
extern int Cyc_pru_listen( struct socket * so, struct proc * p, struct tfi_protosw_cyc * psw);
extern int Cyc_pru_rcvd( struct socket * so, int flags, struct tfi_protosw_cyc * psw);
extern int Cyc_pru_rcvoob( struct socket * so, struct mbuf * m, int flags, struct tfi_protosw_cyc * psw);
extern int Cyc_pru_send( struct socket * so, int flags, struct mbuf * m, struct sockaddr * addr, struct mbuf * control, struct proc * p, struct tfi_protosw_cyc * psw);
extern int Cyc_pru_shutdown( struct socket * so, struct tfi_protosw_cyc * psw);
extern int Cyc_pr_input(struct mbuf *m, int off, int proto, struct socket *so, struct tfi_protosw_cyc * psw);
extern int Cyc_pr_output(struct mbuf * m, struct socket * so, struct tfi_protosw_cyc * psw);
extern int Cyc_pr_ctlinput(struct socket *  so, int flags, struct tfi_protosw_cyc * psw);
extern int Cyc_pr_ctloutput(struct socket * so, struct sockopt *sopt, struct tfi_protosw_cyc * psw);
extern int Cyc_pr_init(struct tfi_protosw_cyc * psw);
extern int Cyc_pr_fasttimo(struct tfi_protosw_cyc * psw);
extern int Cyc_pr_slowtimo(struct tfi_protosw_cyc * psw);
extern int Cyc_pr_drain(struct tfi_protosw_cyc * psw);
extern int Cyc_pr_copystate(struct tfi_tcpcb *ctp, struct socket *so,
			    struct tfi_protosw_cyc *psw);
extern int Cyc_pr_synsocket(struct tfi_syncache *sc, struct socket *so,
			    struct socket **nso, struct tfi_protosw_cyc *psw);


void tfi_minithdr(struct mbuf *m, int offset, int len);

static struct tfi_usrreqs cyc_uu = {
	Cyc_pru_abort,
	Cyc_pru_accept,
	Cyc_pru_attach,
	Cyc_pru_connect,
	Cyc_pru_detach,
	Cyc_pru_disconnect,
	Cyc_pru_listen,
	Cyc_pru_rcvd,
	Cyc_pru_rcvoob,
	Cyc_pru_send,
	Cyc_pru_shutdown
};

static struct tfi_protosw cyc_psw = {
	Cyc_pr_input,
	Cyc_pr_output,
	Cyc_pr_ctlinput,
	Cyc_pr_ctloutput,
	Cyc_pr_init,
	Cyc_pr_fasttimo,
	Cyc_pr_slowtimo,
	Cyc_pr_drain,
	&cyc_uu,
	Cyc_pr_copystate,
	Cyc_pr_synsocket
};

/* end of cyclone stuff */

/* Add a new protocol */
int
tfi_addproto(u_short proto, u_long hash, struct tfi_protosw *sw,
	       struct tfi_proto *pr)
{
	int s = splnet();
	struct tfc_protodomain *dm;
	struct tfc_proto *cp;
	int error = 0;

	/* Fetch the protocol domain */
	dm = protolookup_domain(proto);
	if (dm == NULL) {
		error = TFC_ENOSUPPORT;
		goto out;
	}
	
	/* XXX: Call the protocol init function when the protocol
	   is a loadable module */

	/* Check if the protocol is already loaded */
	if ((error = domain_addproto(dm, hash, sw, NULL, &cp)) != 0) {
		goto out;
	}

	/* Return the protocol handle */
	pr->p = cp;
	/* Done! */
 out:
	splx(s);
	return error;
}

int
tfi_addproto_cyc(u_short proto, u_long hash, struct tfi_protosw_cyc *sw,
	       struct tfi_proto *pr)
{
	int s = splnet();
	struct tfc_protodomain *dm;
	struct tfc_proto *cp;
	int error = 0;

	/* Fetch the protocol domain */
	dm = protolookup_domain(proto);
	if (dm == NULL) {
		error = TFC_ENOSUPPORT;
		goto out;
	}
	
	/* Check if the protocol is already loaded */
	if ((error = domain_addproto(dm, hash, &cyc_psw, sw, &cp)) == 0) {
		/* Return the protocol handle */
		pr->p = cp;
		error = cyc_psw.pr_init(sw); /* XXX */
	} else {
		goto out;
	}

	/* Done! */
 out:
	splx(s);
	return error;
}

int
tfi_delproto(struct tfi_proto *pr)
{
	int s = splnet();
	int error = 0;

	if (pr == NULL) {
		error = EINVAL;
		goto out;
	}

	if (pr->p == NULL) {
		/* Failed early in initialization... */
		goto out;
	}

	if (pr->p->ref_count > 0) {
		error = TFC_EINUSE;
		printf ("Use count = %ld\n", pr->p->ref_count);
		goto out;
	}
	
	if (! TFI_PROTO_INVARIANT(pr)) {
		error = TFC_INVALIDPROTO;
		goto out;
	}
	
	/* Delete the protocol from all tables */
	error = domain_deleteproto(pr->p->dm, pr->p);

	/* zero out everything, so tha the user protocol cannot reuse it */
	bzero(pr, sizeof(struct tfi_proto));
 out:	
	splx(s);
	return error;
}

/* Internal functions */

int
domain_addproto(struct tfc_protodomain *dm, u_long hash,
		struct tfi_protosw *psw, struct tfi_protosw_cyc *cyc_psw, struct tfc_proto **xpp)
{
	/* check if it already exists and then add it to the hashtable */
	/* Code hash to index in the routing table */

	int index = TFC_CHASH2INDEX(dm, hash);
	struct tfc_proto *e;
	struct tfc_protohead *xp = &dm->hashbase[index];
	SLIST_FOREACH(e, xp, next) {
		if (e->hash == hash)
			return TFC_EXISTS;
	}
	/* Not found, initialize, add it, and return */
	MALLOC(e, struct tfc_proto *, sizeof (struct tfc_proto), M_TFCPROTO,
	       M_NOWAIT);
	if (NULL == e) {
		tfc_stats.protomemfail++;
		return TFC_ENOMEM;
	}
	e->hash = hash;
	e->psw = psw;
	e->cyc_psw = cyc_psw;
	e->dm = dm;
	e->ref_count = 0;
	SLIST_INSERT_HEAD(xp, e, next);

	if (hash == dm->default_proto_hash)
		dm->default_protop = e;
	
	*xpp = e;
	return 0;
}

struct tfc_proto *
domain_findproto(struct tfc_protodomain *dm, u_long hash)
{
	int index = TFC_CHASH2INDEX(dm, hash);
	struct tfc_proto *e;
	struct tfc_protohead *xp = &dm->hashbase[index];

	if (xp == NULL)
		return NULL;
	
	SLIST_FOREACH(e, xp, next) {
		if (e && e->hash == hash)
			return e;
	}

	return NULL;
}


int
domain_deleteproto(struct tfc_protodomain *d, struct tfc_proto *p)
{
	struct tfc_protohead *xp = &d->hashbase[TFC_CHASH2INDEX(d, p->hash)];
	SLIST_REMOVE(xp, p, tfc_proto, next);
	if (d->default_protop == p)
		d->default_protop = NULL;
	FREE(p, M_TFCPROTO);
	return 0;
}

struct tfc_protodomain *
protolookup_domain(u_short proto)
{
	int i = 0;
	/* Search through the list of domains */
	for (i = 0; i < TFC_NUM_DOMAINS; i++) {
		if (proto == tfc_domains[i]->proto)
			return tfc_domains[i];
	}

	return NULL;
}

/* socket functions */
int
tfi_sbrappend(struct socket *so, struct mbuf *m)
{
	int error  = 0;
	/* check if there is rcv space in the socket */
	if (tfi_sbrspace(so) < m->m_pkthdr.len) {
#if 0
		printf("%s: no bufs\n", __func__);
#endif
		error = ENOBUFS;                           
		goto out;
	}
	sbappend(&so->so_rcv, m);
 out:
	return error;
}

int
tfi_sbsappend(struct socket *so, struct mbuf *m)
{
	int error  = 0;
	/* check if there is snd space in the socket */
	if (tfi_sbsspace(so) < m->m_pkthdr.len) {
#if 0
		printf("%s: no bufs\n", __func__);
#endif
		error = ENOBUFS;                           
		goto out; 
	}
	sbappend(&so->so_snd, m);
 out:
	return error;
}

#ifdef TFC_LATER
static void
tfc_timer_nofeed(void *vtp)
{
	struct tfc_cb *tp = (struct tfc_cb *)vtp;
	tick_t ts;

	get_tick(ts);

	if (tp->rtt) {
		tick_t diff;
		diff = TICK_DIFF(ts, tp->prev_ts);
		
		if (TICK_GT(diff, tp->rtt)) {
			/* no packets have been sent since last rtt */
			tp->ack_prev = tp->ack_cur;
			tp->sent_prev = tp->sent_cur;
			tp->sent_cur = 0;
			tp->ack_cur = 0;
			tp->prev_ts = ts;
		}
		
		if (TICK_GT(TICK2TIMER(diff), tp->t_nofeed) &&
		    tp->sent_prev == 0 &&
		    tp->ack_prev <= 4) {
			/* no packets have been sent since this timer
			   was set */
			;
		} else {

			if (tp->ack_prev) {
				if (tp->rate_eq >  2 * tp->ack_prev)
					tp->ack_prev = max(tp->ack_prev/2, 1);
				else
					tp->ack_prev = tp->rate_eq/4;
			}

			update_rate(tp);
		}
		
	} else {
		if (tp->rate_eq)
			tp->rate_eq /= 2;
		else
			tp->rate_eq = max(tp->sent_cur/2, 1);
	}
	
	tp->t_nofeed = max(tp->rto, MSEC2TICK(2000));
	tp->t_nofeed = TICK2TIMER(tp->t_nofeed);

	callout_reset(tp->tt_nofeed, tp->t_nofeed,
		      tfc_timer_nofeed, tp);

	/* sb_cleanup(tp); */
	tp->sent_cur = 0;
	tfc_stats.nofeed++;
}
#endif

/*
 * The user protocol should not try to free an mbuf for which no error
 * is returned.
 * tfc_usr_send should ensure that the user protocol cannot free the
 * mbuf passed via the upcall. In general, we need to be careful about
 * accounting mbufs. I learned it the hard way.
 */
int
tfi_netsend(struct mbuf *m, struct inpcb *inp, seq_t seq_no)
{
	int error = 0;
	tick_t ts;
	int ect_cp;
	int s;
	struct tfc_cb *tp = (struct tfc_cb *)inp->inp_ppcb;	

	if (tp == NULL) {
		error = EINVAL;
		goto release;
	}
		
	get_tick(ts);

	if (tp->state > TFCS_INITRTT &&
	    (TICK_GT(TICK_DIFF(ts, tp->prev_ts), tp->rtt) ||
	     tp->prev_ts == 0 ||
	     tp->checknow <= 0)) {
		tp->prev_ts = ts;
		round++;

			  
		error = rate_check(tp, ts);
	
		if (error)
			goto check_done;
	} else 	if (tp->state > TFCS_INITRTT) {
		int rate;

		if (tp->rate_eq)
			/* For now tighter control for xudp, we can be more lenient and allow upto twice the rate_eq for normal transports. */
			rate = max(min(tp->rate_eq, 2 * tp->ack_prev), TFC_OPEN_SS);
		else
#ifdef FAST_SS
			rate = max(max(2 * tp->ack_prev, TFC_OPEN_SS), 2 * tp->sent_prev);
#else
			rate = max(2 * tp->ack_prev, TFC_OPEN_SS);
#endif             /* Do a coarse grained check for each packet */

		
		if (tp->sent_cur + 1 > rate) {
			/* printf("seq no = %d, ack_prev = %d, sent_cur = %d\n", seq_no, tp->ack_prev, tp->ack_cur); */
			tfc_stats.ss_over++;
			error = TFC_WAIT;
			goto check_done;
		} else {
			tfc_stats.ss_ok++;
		}
	} else {
		/* if (tp->state <= TFCS_INITRTT)*/ 
		/* We don't know no rtt */

		if (TICK_GT(TICK_DIFF(ts, tp->prev_ts), MSEC2TICK(MAX_INIT_INTERVAL)))
			tp->sent_cur = 0;

		if (tp->sent_cur + 1 > MAX_INIT_ALLOWED) {
			/* There is no hope for this connection */
			error = TFC_WAIT;
			goto check_done;
		} else {
#ifdef TFC_LATER
			if (tp->t_nofeed == 0)
				tp->t_nofeed = MSEC2TIMER(INIT_NOFEED_TIMER);
			callout_reset(tp->tt_nofeed, tp->t_nofeed,
				      tfc_timer_nofeed, tp);
			/* tp->violations = 0; */
#endif
			tp->prev_ts = ts;
			tfc_stats.initsent++;
		}
	}

 check_done:
	if (tp->violations > MAX_VIOLATIONS) {
		error = TFC_GOAWAY;
	} 
	if (error)
		goto release;
	
	/*
	 * Generate a nonce for this packet
	 * Record the seqno and nonce
	 * Also update the "open" loss-interval
	 * 
	 */
	
	ect_cp = (ts & 0x01);

	/* ii_send will take care of freeing mbuf */
	error = tfii_send(m, inp, ect[ect_cp]);
	if (error == 0) {
		tp->sent_cur++;

		/* printf("Insert seq %d, with nonce %d\n", seq_no, ect[ect_cp]); */
		s = splnet();
		get_tick(ts);
		error = sb_insert(seq_no, ect[ect_cp], tp, ts);
		splx(s);
	}
	
	/* mbuf has been freed */
	goto out;
	
	/* tp->ppl[tp->ppl_open]++; */
 release:
	m_freem(m);

 out:
	return error;
}

int
tfi_netsend_range(struct mbuf *m, struct inpcb *inp, seq_t bseq, seq_t eseq)
{
	int error = 0;
	tick_t ts;
	int ect_cp;
	struct tfc_cb *tp = (struct tfc_cb *)inp->inp_ppcb;	

	if (tp == NULL) {
		error = EINVAL;
		goto release;
	}
		
	get_tick(ts);

	/* 
	 * If this is the first packet send, advance the connection
	 * into slow start, with an initial window of packet credits
	 */

	if (tp->state < TFCS_SLOWSTART) {

	  tp->ack_prev = 0;
	  tp->sent_cur = 0;
	  tp->ack_cur = 0;
	  tp->rate_act = 0;
	  tp->sent_prev = 0;
	  tp->prev_ts = ts;

	  tp->state = TFCS_SLOWSTART;
	  tp->slow_start_credits = TFC_OPEN_SS;
	}

	/*
	 * If an RTT has expired, perform per-RTT processing
	 */
	if (TICK_GT(TICK_DIFF(ts, tp->prev_ts), tp->rtt) ||
	    tp->prev_ts == 0 || tp->checknow <= 0) {
	  
	  tp->prev_ts = ts;
	  round++;
	  
	  /*	  printf("Round=%d\n",round);*/

	  if (tp->state <= TFCS_SLOWSTART) {
	    /*
	     * XXX is this necessary?
	     */
	    per_rtt_bookkeeping(tp);

	    /* add slow-start credit so we don't starve */
	    if (tp->slow_start_credits <= 0)
	      tp->slow_start_credits = TFC_OPEN_SS;
	  }

	  /*
	   * Otherwise, we're in steady-steady.  Perform a 
	   * rate check (which includes the per_rtt_bookkeeping)
	   */
	  else {
	    
	    error = rate_check(tp, ts);
	    
	    if (error) {
	      /*printf("rate_check error (1)\n");*/
	      goto check_done;
	    }	  
	  }
	}

	if (tp->state == TFCS_SLOWSTART) {
	  
	  if (tp->slow_start_credits < 1) {
	    printf("Error: no slow start credits\n");
	    
	    error = TFC_WAIT;
	    goto check_done;
	  }

	  else {
	    tp->slow_start_credits--;
	    tfc_stats.ss_ok++;
	  }
	}
       
	/*
	 * Otherwise, we're in steady-state.
	 * Perform a coarse-grained per-packet rate check.
	 */
	else {
	  int rate;

	  if (tp->rate_eq)	    
	    rate = max(min(tp->rate_eq, 2 * tp->ack_prev), 
		       TFC_OPEN_SS);
	  else {
	    printf("WEIRDNESS: this shouldn't happen!\n");
	    rate = max(max(2 * tp->ack_prev, TFC_OPEN_SS), 2 * tp->sent_prev);
	  }

	    
	  if (tp->sent_cur + 1 > rate) {
	      
	    tfc_stats.ss_over++;
	    error = TFC_WAIT;
	    goto check_done;
		
	  } 
	  else {
	    tfc_stats.ss_ok++;
	  }
	  
	}

	/* 
	 * XXX handle INITRTT case
	 */

 check_done:
	if (tp->violations > MAX_VIOLATIONS) {
		error = TFC_GOAWAY;
	}
	if (error)
		goto release;
	
	/*
	 * Generate a nonce for this packet
	 * Record the seqno and nonce
	 * Also update the "open" loss-interval
	 * 
	 */
	
	ect_cp = (ts & 0x01);

	/* ii_send will take care of freeing mbuf */
	error = tfii_send(m, inp, ect[ect_cp]);
	if (error == 0) {
		tp->sent_cur++;
		error = sb_insert_range(tp, bseq, eseq, ect[ect_cp], ts);
	}
	
	/* mbuf has been freed */
	goto out;
	
	/* tp->ppl[tp->ppl_open]++; */
 release:
	m_freem(m);

 out:
	return error;
}

static void
update_rate(struct tfc_cb *tp)
{
	long long lf;
	long rto, rtt;

	rtt = TICK2USEC(tp->rtt);

	/* our equation does not handle large rtts nicely,
	   so send as small as possible, usec or msec
	*/
	if (rtt > 1000) {
		rtt = TICK2MSEC(tp->rtt);
		rto = TICK2MSEC(tp->rto);
	} else {
		rto = TICK2USEC(tp->rto);
	}
	
	/* calculate the send rate */
	lf = lossfraction(tp);
	/* printf("loss fraction %lld\n", lf); */

	if (lf > 0) {
		int rate;
		rate = tf_rate(lf, rtt, rto);
		/* XXX: sometimes there is a rate of 0 for no reason.
		   Figure it out later
		*/
#ifdef TFC_DEBUG
		printf("Rate %d lf %lld = rtt = %ld rto %ld\n", rate, lf, rtt, rto);
#endif
		if (rate)
			tp->rate_eq = rate;
	} else {
#ifdef FAST_SS		
		tp->rate_eq = max(2 * tp->ack_prev, 2 * tp->sent_prev);
#else
		tp->rate_eq = min(2 * tp->ack_prev, 2 * tp->sent_prev);
#endif

#ifdef TFC_DEBUG
		printf("fr 0 Rate %d per rtt\n", tp->rate_eq);
#endif
	}
	tp->rate_eq = max(tp->rate_eq, TFC_OPEN_SS);
}

static void 
per_rtt_bookkeeping (struct tfc_cb* tp) 
{

	/* The book-keeping that needs to be done each rtt */
	tp->ack_prev = tp->ack_cur;
	tp->sent_prev = tp->sent_cur;
	tp->sent_cur = 0;
	tp->ack_cur = 0;
	
	update_rate(tp);
}

static int
rate_check(struct tfc_cb *tp, tick_t ts)
{
	int error = 0;

	/*
	if (tp->state == TFCS_SLOWSTART)
		goto out;
	*/

	tp->rate_act += tp->sent_cur;
	tp->rate_eqsum += tp->rate_eq;

	if (--tp->checknow <= 0) {
		/* check fine-grained. An average over last checkfreq intervals */
		if (tp->rate_act > tp->rate_eqsum) {
			tp->violations++;
			error = TFC_WAIT;

			printf("transport violated the rate, %d << %d, sent cur %d, ack_cur %d\n", tp->rate_eqsum, tp->rate_act, tp->sent_cur, tp->ack_prev);

			tfc_stats.rate_over++;

		} else {
			tp->violations = 0;
			tfc_stats.rate_ok++;
		}
		tp->checknow = CHECKFREQ;
		tp->rate_act = 0;
		tp->rate_eqsum = 0;
	}

	per_rtt_bookkeeping(tp);

	if (tp->last_ack_ts && TICK_GT(ts, tp->last_ack_ts)) {
		sb_drain(tp, ts);
		/* may be force the protocol into slowstart and
		   slowly open the window
		*/
	}

	return error;
}

int
tfi_acksum(seq_t seq_no, u_char nonce_sum, struct socket *so)
{
	struct tfc_cb *tp = sototfccb(so);
	tick_t ts;
	int error = 0;
	int s;

	get_tick(ts);
	if (tp == NULL) {
		printf("%s: Invalid Socket\n", __FILE__);
		return TFC_EINVAL;
	}
	/* If there has been a loss in the "recent" past
	   and the loss event covers this seq
	*/
	if (tp->nonce_sumsynched == 0 && TFC_SEQ_LEQ(seq_no, tp->prev_le_seq)) {
		goto out;
	}
	
	/* Check that the sum is valid */
	s = splnet();
	error = sb_matchsum(seq_no, nonce_sum, ts, tp);
	if (error) {
		int e;
		
		/* seq mismatch is tantamount to a loss */
		/* The transport must suffer for misreporting an ack */
		/* register loss */
		e = register_loss(seq_no, tp, 0);
#if 0
		if (e)
			printf("register_loss error %d for seq %d\n", e, seq_no);
#endif
	}
	splx(s);
 out:
	if (error == 0) {
		if (TFC_SEQ_GT(seq_no, tp->last_ack_seq))
			tp->last_ack_seq = seq_no;

#ifdef TFC_LATER
		tp->t_nofeed = TICK2TIMER(tp->rto);
		callout_reset(tp->tt_nofeed, tp->t_nofeed,
			      tfc_timer_nofeed, tp);
#endif
	}
	return error;
}

int
tfi_acksum_range(seq_t bseq, seq_t eseq, u_char nonce_sum, struct socket *so)
{
	int s = splnet();
	struct tfc_cb *tp = sototfccb(so);
	tick_t ts;
	int error = 0;

	if (tp == NULL) {
		printf("%s: Invalid Socket\n", __FILE__);
		splx(s);
		return TFC_EINVAL;
	}
	get_tick(ts);

	/* Check that the sum is valid */
	error = sb_matchsum_range(tp, bseq, eseq, nonce_sum, ts);
	if (error == TFC_BADNONCE) {
		int e;
		
		/* nonce mismatch indicates a loss */
		e = register_loss_range(tp, tp->scoreboard[tp->sb_first].bseq,
					eseq, 0);
		if (e)
			printf("register_loss_range error %d for seq %d, %d\n", e,
			       bseq, eseq);
	}
	
	splx(s);
	return error;
}

/*
 * Acknowledge the largest sequence number <= seq_no.  Useful
 * for TCP-style loss recovery.
 */
int tfi_acksum_leq(seq_t seq_no, u_char nonce_sum, struct socket *so)
{
  
	struct tfc_cb *tp = sototfccb(so);
	tick_t ts;
	int error = 0;
	int s;

	/*printf("tfi_acksum_leq: seq_no=%d nonce_sum=0x%x\n",
	       seq_no,nonce_sum); 
	       */

	get_tick(ts);
	if (tp == NULL) {
		printf("%s: Invalid Socket\n", __FILE__);
		return TFC_EINVAL;
	}
	/* If there has been a loss in the "recent" past
	   and the loss event covers this seq
	*/


	if (tp->nonce_sumsynched == 0 && TFC_SEQ_LEQ(seq_no, tp->prev_le_seq)) {
		goto out;
	}

	/* Check that the sum is valid */
	s = splnet();
	error = sb_matchsum_leq(seq_no, nonce_sum, ts, tp);
	if (error) {
		int e;
		
		printf("WARNING: bad nonce_sum: seqno=%d nonce_sum=0x%x\n",
		       seq_no,nonce_sum);

		/* seq mismatch is tantamount to a loss */
		/* The transport must suffer for misreporting an ack */
		/* register loss */
		e = register_loss_leq(seq_no, tp, 0);

#if 0
		if (e)
		  printf("register_loss error %d for seq %d\n", e, seq_no);
#endif
		
	}

	splx(s);
 out:
	if (error == 0) {
	  if (TFC_SEQ_GT(seq_no, tp->last_ack_seq))
			tp->last_ack_seq = seq_no;



#ifdef TFC_LATER
		tp->t_nofeed = TICK2TIMER(tp->rto);
		callout_reset(tp->tt_nofeed, tp->t_nofeed,
			      tfc_timer_nofeed, tp);
#endif
	}
	return error;
}

int
tfi_nack(seq_t seq_no, struct socket *so)
{
	int error = 0;
	int s;
	struct tfc_cb *tp = sototfccb(so);
	if (tp == NULL) {
		return EINVAL;
	}
	tfc_stats.nacktotal++;
	s = splnet();
	error = register_loss(seq_no, tp, 0);
	splx(s);
	return error;
}

int
tfi_sendack(struct mbuf *m, struct inpcb *inp)
{
	tfc_stats.acksent++;
        return ip_output(m, inp->inp_options, &inp->inp_route, 
		  (inp->inp_socket->so_options & (SO_DONTROUTE | SO_BROADCAST)), 
		  inp->inp_moptions); 
}

static int
tfii_send(struct mbuf *m, struct inpcb *inp, u_char ect)
{
	int error = 0;
	struct ip *ih = mtod(m, struct ip*);

	if (ih == NULL) {
		error = EINVAL;
		goto release;
	}
#if 0
	if (ect > 1) {
		error = EINVAL;
		printf("%s: %d oof\n", __func__, ect_cp);
		goto release;
	}
#endif

	ih->ip_tos |= ect;

	error = ip_output(m, inp->inp_options, &inp->inp_route, 
		  (inp->inp_socket->so_options & (SO_DONTROUTE | SO_BROADCAST)),
		  inp->inp_moptions);

	/* All done here, return status */
	goto out;

 release:
	m_freem(m);
 out:
	if (error == 0)
		tfc_stats.sendtotal++;
#if 0
	else if (error == ENOBUFS)
		printf("IP output error\n");
#endif
	return error;
	
}

struct mbuf *
tfi_mgethdr()
{
	struct mbuf *m;
	MGETHDR((m), M_DONTWAIT, MT_HEADER);
	if ((m) != NULL) {
		(m)->m_data += max_linkhdr;
		(m)->m_pkthdr.rcvif = (struct ifnet *)0;
	}
	return m;
}

/* This works only for header mbufs and their length is always less than 100 bytes */ 
int
tfi_msethlen(struct mbuf *m, int plen)
{
	if (! (m->m_flags & M_PKTHDR)) {
		printf("%s: not a packet\n", __func__);
		return -1;
	}
	m->m_pkthdr.len = plen;
	return 0;
}

int
tfi_msetlen(struct mbuf *m, int len)
{
	if (len > MLEN) {
		printf("%s: bad setlen\n", __func__);
		return -1;
	}
	m->m_len = len;
	return 0;
}

/* This is safe because we make the setlen check above */
int
tfi_msetdata(struct mbuf *m, char *data)
{
	if (! (m->m_flags & M_PKTHDR)) {
		printf("%s: bad setlen\n", __func__);
		return -1;
	}
	if (m->m_next)
		m_freem(m->m_next);
	m->m_next = NULL;
	m->m_data = data;
	return 0;
}

/*
 * Send a packet while reporting a loss for the largest sequence
 * number <= seq_no
 */
int
tfi_netresend(struct mbuf *m, struct inpcb *inp, seq_t seq_no)
{
	int error = 0;
	u_char ect_cp;
	tick_t ts;
	struct tfc_cb *tp = (struct tfc_cb *)inp->inp_ppcb;

	if (tp == NULL) {
		error = EINVAL;
		goto release;
	} 

	get_tick(ts);
	
	/* 1. Register loss or the protocol can fool us */
	error = register_loss_leq(seq_no, tp, 1);
	if (error)
		goto release;

	/* 2. Get nonce from the scoreboard */
	error = sb_update_leq(seq_no, ts, tp, &ect_cp);
	if (error)
		goto release;

	/* We do not generate a new nonce for this packet because it doesn't
	 * matter to us whether the protocol gets an ack for this one or the
	 * original packet.
	 *
	 * This can also handle protocols that suffer from retransmission ambiguity.
	 */

	/* 3. Just send it */
	
	error = tfii_send(m, inp, ect_cp);

	goto out;

 release:
	m_freem(m);
 out:
	return error;
}

/*
 * Send a packet while reporting a loss for the largest sequence
 * number <= seq_no
 */
int
tfi_netresend_range(struct mbuf *m, struct inpcb *inp, seq_t bseq, seq_t eseq)
{
	int error = 0;
	u_char ect_cp;
	tick_t ts;
	struct tfc_cb *tp = (struct tfc_cb *)inp->inp_ppcb;

	if (tp == NULL) {
		error = EINVAL;
		goto release;
	} 

	get_tick(ts);
	
	/* 1. Register loss or the protocol can fool us */
	error = register_loss_range(tp, bseq, eseq, 1);
	if (error)
		goto release;

	/* 2. Get nonce from the scoreboard */
	error = sb_update_range(tp, bseq, eseq, ts, &ect_cp);
	if (error)
		goto release;

	/* We do not generate a new nonce for this packet because it doesn't
	 * matter to us whether the protocol gets an ack for this one or the
	 * original packet.
	 *
	 * This can also handle protocols that suffer from retransmission ambiguity.
	 */

	/* 3. Just send it */
	error = tfii_send(m, inp, ect_cp);

	goto out;

 release:
	m_freem(m);
 out:
	return error;
}

#ifdef TFC_LATER

int
tfi_ack(struct socket *so, seq_t seq_no, u_char nonce)
{
	struct tfc_cb *tp = sototfccb(so);
	int sb_idx;
	if (tp == NULL)
		return EINVAL;
	/*
	  Ack a packet.
	  See if this packet acks the next expected ack in seq
	  Check if there has been a loss
	*/
	if (TFC_SEQ_LEQ(seq_no, tp->prev_le))
		return 0;

	sb_idx = lookup_sb(seq_no, tp);
	if (sb_idx < 0)
		return EINVAL;

	if (tp->scoreboard[sb_idx].nonce != nonce)
		return TFC_BADNONCE;

	if (tp->sb_start == seq_no)
		tp->sb_start = SB_INC(tp->sb_start);

	if (TFC_SEQ_GT(seq_no, tp->last_ack))
		tp->last_ack = seq_no;
	
	return 0;
}

int
tfi_ack_range()
{
	return 0;
}

#endif

/* All of the functions below should be inlined and put in a separate header file
   or it'll suck big time */
long tfi_sbrspace(struct socket *so) {
	return sbspace(&(so)->so_rcv);
}
long tfi_sbsspace(struct socket *so) {
	return sbspace(&(so)->so_snd);
} 

TFI_INLINE void tfi_sorwakeup(struct socket *so) { sorwakeup(so); }
TFI_INLINE void tfi_sowwakeup(struct socket *so) { sowwakeup(so); }
TFI_INLINE void tfi_soseterror(struct socket *so, int err) { so->so_error = err; }
TFI_INLINE void tfi_sosetstate(struct socket *so, short state) { so->so_state |= state; }
TFI_INLINE void tfi_sounsetstate(struct socket *so, short state) { so->so_state &= ~state; }
TFI_INLINE struct inpcb *tfi_sotoinpcb(struct socket *so) { return sotoinpcb(so); }
TFI_INLINE struct xtcpcb *tfi_sotouserpcb(struct socket *so) {
	return (struct xtcpcb *) sototfccb(so)->pcb;
}
TFI_INLINE struct xtcpcb *tfi_intouserpcb(struct inpcb *inp) {
	return (struct xtcpcb*)((struct tfc_cb *)inp->inp_ppcb)->pcb;
}
TFI_INLINE int tfi_mlen(struct mbuf *m) { return m ? m->m_len : 0; }

TFI_INLINE struct _dyneither_ptr tfi_mtodfat(struct mbuf *m)
{
	struct _dyneither_ptr p = { 0 };
	char *s = mtod(m, char *);
	if (s) {
		p.curr = p.base = s;
		p.last_plus_one = s + MLEN; /* XXX m->m_len; */
	}
	return p;
}

/* These definitions should go into cmap.c */
struct ip *tfi_mtoip(struct mbuf *m)
{
#if 0
	if ( < sizeof(struct ip)) {
		printf("%s: oof\n", __func__);
		return NULL;
	}
#endif
	return mtod(m, struct ip*);
}

struct xtcptemp *tfi_mtoxtcptemp(struct mbuf *m)
{
	/* XXX: Hard-coded structure size */
	if (m->m_len < 64) {
		printf("%s: oof\n", __func__);
		return NULL;
	}
	return mtod(m, struct xtcptemp*);
}

struct xudphdr *tfi_strtoxudphdr(struct _dyneither_ptr data)
{
	if (data.last_plus_one - data.curr < sizeof(struct xudphdr)) {
		printf("%s: oof\n", __func__);
		return NULL;
	}

	return (struct xudphdr *)data.curr;
}

struct xtcphdr *tfi_strtoxtcphdr(struct _dyneither_ptr data)
{
	if (data.last_plus_one - data.curr < 20) {
		printf("%s: oof\n", __func__);
		return NULL;
	}

	return (struct xtcphdr *)data.curr;
}


struct xudpiphdr *tfi_strtoxudpiphdr(struct _dyneither_ptr data)
{
	if (data.last_plus_one - data.curr < sizeof(struct xudpiphdr)) {
		printf("%s: oof\n", __func__);
		return NULL;
	}

	return (struct xudpiphdr *)data.curr;
}

struct ipovly *tfi_strtoipovly(struct _dyneither_ptr data)
{
	if (data.last_plus_one - data.curr < sizeof(struct ipovly)) {
		printf("%s: oof\n", __func__);
		return NULL;
	}

	return (struct ipovly *)data.curr;
}

struct in_addr tfi_inpladdr(struct inpcb *inp) {
	return inp->inp_laddr;
}
struct in_addr tfi_inpfaddr(struct inpcb *inp) {
	return inp->inp_faddr;
}

u_int16_t tfi_inplport(struct inpcb *inp) {
	return inp->inp_lport;
}
u_int16_t tfi_inpfport(struct inpcb *inp) {
	return inp->inp_fport;
}

struct pkthdr* tfi_m_pkthdr(struct mbuf *m) {
	/* XXX: Make sure that mbuf is of type M_PKTHDR */
	return &(m->m_pkthdr);
}

u_char tfi_inpttl(struct inpcb *inp) {
	return inp->inp_ip_ttl;
}
u_char tfi_inptos(struct inpcb *inp) {
	return inp->inp_ip_tos;
}

struct mbuf *tfi_m_prepend(struct mbuf *m, int size, int how) {
	M_PREPEND(m, sizeof(struct xudpiphdr), M_DONTWAIT);
	if (m == NULL)
		return NULL;
	return m;
}

int tfi_soreserve(struct socket *so, unsigned long sndcc, unsigned long rcvcc)
{
	return soreserve(so, sndcc, rcvcc);
}

void tfi_soisconnected(struct socket *so)
{
	soisconnected(so);
}

void tfi_m_free(struct mbuf *m)
{
	m_free(m);
}

void tfi_m_freem(struct mbuf *m)
{
	m_freem(m);
}

void tfi_m_adj(struct mbuf *mp, int req_len) { m_adj(mp, req_len); }
struct mbuf *tfi_mpullup(struct mbuf *m, int len) { return m_pullup(m, len); }
int tfi_splnet() { return splnet(); }
void tfi_splx(int s) { splx(s); }

int tfi_in_pcbconnect(struct inpcb *inp, struct sockaddr *nam,
			     struct proc *p) {
	if (p == NULL) {
		p = &proc0;
	}
	return in_pcbconnect(inp, nam, p);
}

void tfi_in_pcbdisconnect(struct inpcb *inp) {
	in_pcbdisconnect(inp);
}

int tfi_in_pcbbind(struct inpcb *inp, struct sockaddr *nam, struct proc *p) {
	return in_pcbbind(inp, nam, p);
}

int tfi_in_cksum(struct mbuf *m, int len) {
	return in_cksum(m, len);
}

u_short tfi_in_pseudo(u_int32_t a, u_int32_t b, u_int32_t c) {
	return in_pseudo(a, b, c);
}

void *tfi_calloc(int size, int mult) {
	void *retval;
	retval = malloc(size * mult, M_TFCPROTO, M_ZERO | M_NOWAIT);
	/* printf("calloc %p %d\n", retval, size); */
	return retval;
}

void tfi_free(void *ptr) {
	/* printf("FREEING %p!!!\n", ptr); */
	free(ptr, M_TFCPROTO);
}

void tfi_exit(int status) {
	printf("Exit called \n");
	return;
}
u_int32_t tfi_htonl(u_int32_t x) {
	return htonl(x);
}
u_int32_t tfi_ntohl(u_int32_t x) {
	return ntohl(x);
}
u_int16_t tfi_htons(u_int16_t x) {
	return htons(x);
}
u_int16_t tfi_ntohs(u_int16_t x) {
	return ntohs(x);
}

int tfi_printf(const char *x) {
	return printf(x);
}

struct rt_metrics *tfi_getrtmx(struct rtentry *rte)
{
	return &rte->rt_rmx;
}

/* BEGIN -- Parveen */
int tfi_soshiwat(struct socket *so)
{
	return so->so_snd.sb_hiwat;
}

int tfi_sorhiwat(struct socket *so)
{
	return so->so_rcv.sb_hiwat;
}

struct rtentry *
tfi_rtlookup(inc)
	struct in_conninfo *inc;
{
	struct route *ro;
	struct rtentry *rt;

	ro = &inc->inc_route;
	rt = ro->ro_rt;
	if (rt == NULL || !(rt->rt_flags & RTF_UP)) {
		/* No route yet, so try to acquire one */
		if (inc->inc_faddr.s_addr != INADDR_ANY) {
			ro->ro_dst.sa_family = AF_INET;
			ro->ro_dst.sa_len = sizeof(struct sockaddr_in);
			((struct sockaddr_in *) &ro->ro_dst)->sin_addr =
			    inc->inc_faddr;
			rtalloc(ro);
			rt = ro->ro_rt;
		}
	}
	return rt;
}

int tfi_isrtup(struct rtentry *rt)
{
	return ((rt->rt_flags & (RTF_UP|RTF_HOST)) == (RTF_UP|RTF_HOST));
}

int tfi_issolinger(struct socket *so)
{
	return (((so->so_options & SO_LINGER) && so->so_linger == 0));
}

void tfi_setsolinger(struct socket *so, int period)
{
	if ((so->so_options & SO_LINGER) && so->so_linger == 0)
		so->so_linger = period;
}

void tfi_sbrflush(struct socket * so)
{
	sbflush(&so->so_rcv);
}

void tfi_soisdisconnecting(struct socket * so)
{
	soisdisconnecting(so);
}

void tfi_soisdisconnected(struct socket * so)
{
	soisdisconnected(so);
}

void tfi_soisconnecting(struct socket * so)
{
	soisconnecting(so);
}

void tfi_socantsendmore(struct socket * so)
{
	socantsendmore(so);
}

int tfi_sogetscc(struct socket *so)
{
	return so->so_snd.sb_cc;
}

int tfi_sogetrcc(struct socket *so)
{
	return so->so_rcv.sb_cc;
}

int tfi_in_pcbladdr(struct inpcb *inp, struct sockaddr *nam,
		    struct sockaddr_in **ifaddr)
{
	return in_pcbladdr(inp, nam, ifaddr);
}

struct inpcb *tfi_in_pcblookup_hash(struct inpcb *inp, struct in_addr faddr, u_int fport,
				    struct in_addr laddr, u_int lport, u_int wildcard,
				    struct ifnet *ifp)
{
	return in_pcblookup_hash(inp->inp_pcbinfo, faddr, fport,
				 laddr, lport, wildcard,
				 ifp);
}
void tfi_in_pcbrehash(struct inpcb *inp)
{
	in_pcbrehash(inp);
}


/* END - Parveen */

short tfi_sogetoptions(struct socket *so)
{
	return so->so_options;
}

struct mbuf *tfi_xtcptemptom(struct xtcptemp *xtt)
{
	/* XXX: Hard-coded structure size */
	return dtom(xtt);
}
short tfi_getsostate(struct socket * so)
{
	return so->so_state;
}
short tfi_mflags(struct mbuf *m)
{
	return m->m_flags;
}
int tfi_in_broadcast(struct in_addr in, struct ifnet *ifp)
{
	return in_broadcast(in, ifp);
}

int tfi_in_localaddr(struct in_addr in)
{
	return in_localaddr(in);
}

void tfi_sbsdrop(struct socket *so, int len)
{
	sbdrop(&so->so_snd, len);
}
void tfi_sohasoutofband(struct socket *so)
{
	sohasoutofband(so);
}
void tfi_socantrcvmore(struct socket *so)
{
	socantrcvmore(so);
}
struct mbuf *tfi_mgetnext(struct mbuf *m)
{
	return m->m_next;
}

int tfi_rtgetmtu(struct rtentry *rt)
{
	return rt->rt_ifp->if_mtu;
}

void tfi_sbsreserve(struct socket *so, u_long cc)
{
	sbreserve(&so->so_snd, cc, so, NULL);
}
void tfi_sbrreserve(struct socket *so, u_long cc)
{
	sbreserve(&so->so_rcv, cc, so, NULL);
}

long tfi_getrt_refcnt(struct rtentry *rte)
{
	return rte->rt_refcnt;
}

u_long tfi_getrt_flags(struct rtentry *rte)
{
	return rte->rt_flags;
}

struct sockaddr *tfi_getrt_gateway(struct rtentry *rte)
{
	return rte->rt_gateway;
}

void tfi_rt_free(struct rtentry *rte)
{
	RTFREE(rte);
}

struct sockaddr *tfi_rt_key(struct rtentry *rte)
{
	return rt_key(rte);
}

struct sockaddr *tfi_rt_mask(struct rtentry *rte)
{
	return rt_mask(rte);
}

struct _dyneither_ptr tfi_m_copydata(struct socket *so, int off, int len,
					   struct _dyneither_ptr data)
{
	if (data.last_plus_one - data.curr < len) {
		data.base = data.curr = data.last_plus_one = NULL;
		return data;
	}
	m_copydata(so->so_snd.sb_mb, off, len, data.curr);
	data.curr += len;
	return data;
}

struct mbuf *tfi_m_copy(struct socket *so, int off, int len)
{
	return m_copy(so->so_snd.sb_mb, off, len);
}

void tfi_m_setnext(struct mbuf *m, struct mbuf *n)
{
	m->m_next = n;
}

u_short  tfi_in_addword(u_short sum, u_short w)
{
	return in_addword(sum, w);
}

TFI_INLINE
struct socket *tfi_sonewconn(struct socket *head, int connstatus)
{
	struct socket *retval;
	struct tfc_proto *xp;
	int error;
	
	retval = sonewconn(head, connstatus);

	/*
	 * This indicates a new accept socket.
	 * Directly invoke the user attach function
	 */
	xp = sototfccb(head)->xp;
	error = tfc_soattachproto(retval, xp->dm->proto, xp, NULL);
	if( error )
	{
		printf("fuck %d\n", error);
	}
	return retval;
}

TFI_INLINE
struct socket *tfi_sonewconn1(struct socket *head, int connstatus,
			      struct tfi_proto *proto)
{
	struct socket *retval;
	struct tfc_proto *xp;
	int error;

	retval = sonewconn(head, connstatus);

	/*
	 * This indicates a new accept socket.
	 * Directly invoke the user attach function
	 */
	if (proto == NULL) {
		xp = sototfccb(head)->xp;
	} else
		xp = proto->p;
	
	error = tfc_soattachproto(retval, xp->dm->proto, xp, NULL);
	if( error ) {
		printf("soattach proto filed with error %d\n", error);
	}
	return retval;
}


TFI_INLINE void tfi_soabort(struct socket *so) { soabort(so); }

void tfi_minithdr(struct mbuf *m, int offset, int len)
{
	/* m->m_data += offset; */
	m->m_len = len;
	m->m_pkthdr.len = len;
	m->m_pkthdr.rcvif = NULL;
}

/* end of -- These definitions should go into cmap.c */

int tfi_pru_accept_notsupp(struct socket *so, struct sockaddr **nam, struct tfi_protosw_cyc *nu)
{
	return EOPNOTSUPP;
}

int tfi_pru_listen_notsupp(struct socket *so, struct proc *p, struct tfi_protosw_cyc *nu)
{
	return EOPNOTSUPP;
}

int tfi_pru_rcvd_notsupp(struct socket *so, int flags, struct tfi_protosw_cyc *nu)
{
	return EOPNOTSUPP;
}

int tfi_pru_rcvoob_notsupp(struct socket *so, struct mbuf *m, int flags, struct tfi_protosw_cyc *nu)
{
	return EOPNOTSUPP;
}

/* set_handler macro should set the exception handler, so that null exception is
 not thrown
*/

int Cyc_pru_abort(struct socket *so, struct tfi_protosw_cyc *psw)
{
	set_handler(retval = psw->pr_usrreqs->pru_abort(so));
}

int Cyc_pru_accept(struct socket *so, struct sockaddr **nam, struct tfi_protosw_cyc *psw)
{
	set_handler(retval = psw->pr_usrreqs->pru_accept(so, nam));
}

int Cyc_pru_attach( struct socket *so, int proto, struct proc *p, struct tfi_protosw_cyc *psw)
{
	set_handler(retval = psw->pr_usrreqs->pru_attach(so, proto, p));
}

int Cyc_pru_connect( struct socket *so, struct sockaddr *nam,
	       struct proc *p, struct tfi_protosw_cyc *psw)
{
	set_handler(retval = psw->pr_usrreqs->pru_connect(so, nam, p));
}

int Cyc_pru_detach( struct socket *so, struct tfi_protosw_cyc *psw)
{
	set_handler(retval = psw->pr_usrreqs->pru_detach(so));
}

int Cyc_pru_disconnect( struct socket *so, struct tfi_protosw_cyc *psw)
{
	set_handler(retval = psw->pr_usrreqs->pru_disconnect(so));
}

int Cyc_pru_listen( struct socket *so, struct proc *p, struct tfi_protosw_cyc *psw)
{
	set_handler(retval = psw->pr_usrreqs->pru_listen(so, p));
}

int Cyc_pru_rcvd( struct socket *so, int flags, struct tfi_protosw_cyc *psw)
{
	set_handler(retval = psw->pr_usrreqs->pru_rcvd(so, flags));
}

int Cyc_pru_rcvoob( struct socket *so, struct mbuf *m, int flags, struct tfi_protosw_cyc *psw)
{
	set_handler(retval = psw->pr_usrreqs->pru_rcvoob(so, m, flags));
}

int Cyc_pru_send( struct socket *so, int flags, struct mbuf *m, 
				 struct sockaddr *addr, struct mbuf *control,
				 struct proc *p, struct tfi_protosw_cyc *psw)
{
	set_handler(retval = psw->pr_usrreqs->pru_send(so,
						       flags,
						       m,
						       addr,
						       control,
						       p));
}

int Cyc_pru_shutdown( struct socket *so, struct tfi_protosw_cyc *psw)
{
	set_handler(retval = psw->pr_usrreqs->pru_shutdown(so));
}

int Cyc_pr_input(struct mbuf *m, int off, int proto, struct socket * so, struct tfi_protosw_cyc *psw)
{
	set_handler(retval = psw->pr_input(m, off, proto, so));
}
int Cyc_pr_output(struct mbuf *m, struct socket *so, struct tfi_protosw_cyc *psw)
{
	set_handler(retval = psw->pr_output(m, so));
}
					/* output to protocol (from above) */
int Cyc_pr_ctlinput(struct socket * so, int flags, struct tfi_protosw_cyc *psw)
{
	set_handler({ psw->pr_ctlinput(so, flags); retval = 0; });
}

int Cyc_pr_ctloutput(struct socket *so, struct sockopt *sopt, struct tfi_protosw_cyc *psw)
{
	set_handler(retval = psw->pr_ctloutput(so, sopt));
}

int Cyc_pr_init(struct tfi_protosw_cyc *psw)
{
	set_handler(retval = psw->pr_init());
}

int Cyc_pr_fasttimo(struct tfi_protosw_cyc *psw)
{
	set_handler({ psw->pr_fasttimo(); retval = 0; });
}

int Cyc_pr_slowtimo(struct tfi_protosw_cyc *psw)
{
	set_handler({ psw->pr_slowtimo(); retval = 0; });
}

int Cyc_pr_drain(struct tfi_protosw_cyc *psw)
{
	set_handler({ psw->pr_drain(); retval = 0; });
}

int Cyc_pr_copystate(struct tfi_tcpcb *ctp, struct socket *so,
		     struct tfi_protosw_cyc *psw)
{
	set_handler(retval = psw->pr_copystate(ctp, so));
}

int Cyc_pr_synsocket(struct tfi_syncache *sc, struct socket *so,
		     struct socket **nso, struct tfi_protosw_cyc *psw)
{
	set_handler(retval = psw->pr_synsocket(sc, so, nso));
}
