/*
 * tfi.h
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * for any purpose with or without fee is hereby granted, provided
 * that the above copyright notice and this permission notice
 * appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE UNIVERSITY OF UTAH
 * DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING
 * ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO
 * EVENT SHALL THE UNIVERSITY OF UTAH BE LIABLE FOR ANY SPECIAL,
 * DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER
 * IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
 * THIS SOFTWARE.
 */
#ifndef TFI_PROTO_H
#define TFI_PROTO_H

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/malloc.h>
#include <sys/kernel.h>
#include <sys/sysctl.h>
#include <sys/mbuf.h>
#include <sys/socket.h>
#include <sys/socketvar.h>
#include <sys/protosw.h>
#include <sys/proc.h>
#include <sys/jail.h>
#include <sys/queue.h>

#include <vm/vm_zone.h> 

#include <net/if.h>
#include <net/route.h>

#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/in_pcb.h>
#include <netinet/in_var.h>
#include <netinet/ip_var.h>
#include <netinet/tcp.h>
#include <netinet/tcp_fsm.h>
#include <netinet/tcp_seq.h>
#include <netinet/tcp_timer.h>
#include <netinet/tcp_var.h>
#include <netinet/tcpip.h>


#include <netinet/xtcp/tfc.h>
#include <netinet/xtcp/tfci.h>
#include <netinet/xtcp/tfc_stats.h>
#include <netinet/xtcp/tfc_error.h>
#include <netinet/xtcp/lossfraction.h>
#include <netinet/xtcp/scoreboard.h>
#include <netinet/xtcp/tfc_cb.h>

#include <netinet/ip.h> /* for definition of struct ip */
#include <machine/in_cksum.h>

/* No inlining will work for now */
#define TFI_INLINE 

int tfi_pcbfree(void *state);

/* The next structure is used for connection negotiation */
/* This should change as and when xtcp_extensions/tcp/ changes */
struct xtcp_syncache_dummy_t {
	void *a, *b;
};
struct tfi_syncache {
	inp_gen_t       sc_inp_gencnt;		/* pointer check */
	struct tfi_tcpcb  *sc_tp;		/* tcb for listening socket */
	struct		mbuf *sc_ipopts;	/* source route */
	struct 		in_conninfo sc_inc;	/* addresses */
#define sc_route	sc_inc.inc_route
#define sc_route6	sc_inc.inc6_route
	u_int32_t	sc_tsrecent;
	tcp_cc		sc_cc_send;		/* holds CC or CCnew */
	tcp_cc		sc_cc_recv;
	tcp_seq 	sc_irs;			/* seq from peer */
	tcp_seq 	sc_iss;			/* our ISS */
	u_long		sc_rxttime;		/* retransmit time */
	u_int16_t	sc_rxtslot; 		/* retransmit counter */
	u_int16_t	sc_peer_mss;		/* peer's MSS */
	u_int16_t	sc_wnd;			/* advertised window */
	u_int8_t 	sc_requested_s_scale:4,
			sc_request_r_scale:4;
	u_int8_t	sc_flags;
#define SCF_NOOPT	0x01			/* no TCP options */
#define SCF_WINSCALE	0x02			/* negotiated window scaling */
#define SCF_TIMESTAMP	0x04			/* negotiated timestamps */
#define SCF_CC		0x08			/* negotiated CC */
#define SCF_SACK_PERMIT	0x40			/* negotiated SACK */
#define SCF_UNREACH	0x10			/* icmp unreachable received */
#define SCF_KEEPROUTE	0x20			/* keep cloned route */
#define SCF_TFCNEGOTIATE_USE	0x80		/* remote end requested a protocol */
	struct xtcp_syncache_dummy_t sc_hash, sc_timerq;
        u_char          sc_nonce_sum;
	u_long          sc_tiwin;
        u_long          sc_resends;             /* How many resends? */
	u_long          sc_rhash;             /* hash of the requested protocol */
	u_long          sc_lhash;             /* hash of the locally acceptable protocol */
};

/*
  o struct socket should be treated as opaque object by the user protocol.
  o struct inp should be treated read-only.
*/
struct tfi_usrreqs {
	int	(*pru_abort) __P((struct socket *so, struct tfi_protosw_cyc *psw));
	int	(*pru_accept) __P((struct socket *so, struct sockaddr **nam, struct tfi_protosw_cyc *psw));
	int	(*pru_attach) __P((struct socket *so, int proto,
				   struct proc *p, struct tfi_protosw_cyc *psw));
	/* Not needed. Handled completely by TFC */
	/* int	(*pru_bind) __P((struct socket *so, struct sockaddr *nam,
	   struct proc *p)); */

	int	(*pru_connect) __P((struct socket *so, struct sockaddr *nam,
				    struct proc *p, struct tfi_protosw_cyc *psw));

	/* Not needed, anyway for now */
        /* int	(*pru_connect2) __P((struct socket *so1, struct socket *so2)); */
	/* it's always in_control */
	/* int	(*pru_control) __P((struct socket *so, u_long cmd, caddr_t data,
	   struct ifnet *ifp, struct proc *p)); */
	
	int	(*pru_detach) __P((struct socket *so, struct tfi_protosw_cyc *psw));
	int	(*pru_disconnect) __P((struct socket *so, struct tfi_protosw_cyc *psw));
	int	(*pru_listen) __P((struct socket *so, struct proc *p, struct tfi_protosw_cyc *psw));

	/* Not needed because in_setpeeraddr will take care of it. */
	/* int	(*pru_peeraddr) __P((struct socket *so, 
	   struct sockaddr **nam)); */

	int	(*pru_rcvd) __P((struct socket *so, int flags, struct tfi_protosw_cyc *psw));
	int	(*pru_rcvoob) __P((struct socket *so, struct mbuf *m,
				   int flags, struct tfi_protosw_cyc *psw));
	int	(*pru_send) __P((struct socket *so, int flags, struct mbuf *m, 
				 struct sockaddr *addr, struct mbuf *control,
				 struct proc *p, struct tfi_protosw_cyc *psw));
#define	TFI_PRUS_OOB	0x1
#define	TFI_PRUS_EOF	0x2
#define	TFI_PRUS_MORETOCOME	0x4

	int	(*pru_shutdown) __P((struct socket *so, struct tfi_protosw_cyc *psw));
};

struct tfi_tcpcb; /* replicate the xtcpcb struct here */

/* structure that a tfc protocol will provide */
struct tfi_protosw {
	/* short	pr_type;	*/	/* socket type used for */
	/* short	pr_protocol;	*/	/* protocol number */
	/* short	pr_flags;	*/	/* see below */
	/* protocol-protocol hooks */
	int	(*pr_input) __P((struct mbuf *, int off, int proto, struct socket *,
				 struct tfi_protosw_cyc *psw));
	/* input to protocol (from below) */
	int	(*pr_output) __P((struct mbuf *m, struct socket *so, struct tfi_protosw_cyc *psw));
	/* output to protocol (from above) */
	int	(*pr_ctlinput) __P((register struct socket *, int, struct tfi_protosw_cyc *psw));
	/* control input (from below) */
	int	(*pr_ctloutput) __P((struct socket *, struct sockopt *,
				     struct tfi_protosw_cyc *psw));
	/* control output (from above) */
	/* utility hooks */
	int	(*pr_init) __P((struct tfi_protosw_cyc *psw));	/* initialization hook */
	int	(*pr_fasttimo) __P((struct tfi_protosw_cyc *psw));
	/* fast timeout (200ms) */
	int	(*pr_slowtimo) __P((struct tfi_protosw_cyc *psw));
	/* slow timeout (500ms) */
	int	(*pr_drain) __P((struct tfi_protosw_cyc *psw));
	/* flush any excess space possible */
	struct	tfi_usrreqs *pr_usrreqs;	/* supersedes pr_usrreq() */
	int     (*pr_copystate) __P((struct tfi_tcpcb *s, struct socket *so, struct tfi_protosw_cyc *psw));
	int     (*pr_synsocket) __P((struct tfi_syncache *, struct socket *so, struct socket **nso, struct tfi_protosw_cyc *psw));
};

struct tfc_proto;

/* Opaque handle for the user protocol */
struct tfi_proto {
	struct tfc_proto *p;
};

#define TFC_DISCONNECT 1

extern void tfi_downcallenter(void);
extern void tfi_downcallleave(void);

extern int tfi_addproto(u_short domain, u_long hash, struct tfi_protosw *sw,
			struct tfi_proto *pr);
extern int tfi_addproto_cyc(u_short proto, u_long hash,
			    struct tfi_protosw_cyc *sw,
			    struct tfi_proto *pr);

extern int tfi_delproto(struct tfi_proto *pr);

/* For now these will just map to the original functions */
/* socket functions */
#define tfi_sbappendaddr sbappendaddr
extern int tfi_sbrappend(struct socket *so, struct mbuf *m);
extern int tfi_sbsappend(struct socket *so, struct mbuf *m);
extern long tfi_sbrspace(struct socket *so);
extern long tfi_sbsspace(struct socket *so);
extern void tfi_sorwakeup(struct socket *so);
extern void tfi_sowwakeup(struct socket *so);

extern void tfi_soseterror(struct socket *so, int err);
extern void tfi_sosetstate(struct socket *so, short state);
extern void tfi_sounsetstate(struct socket *so, short state);
extern struct inpcb *tfi_sotoinpcb(struct socket *so);

/* #define tfi_soreserve soreserve *//* This should check for limits */
/* #define tfi_soisconnected soisconnected */
extern int tfi_soreserve(struct socket *so, unsigned long sndcc, unsigned long rcvcc);
extern void tfi_soisconnected(struct socket *so);

struct socket *tfi_sonewconn(struct socket *head, int connstatus);
struct socket *tfi_sonewconn1(struct socket *head, int connstatus,
			      struct tfi_proto *proto);

/* utility functions */
extern struct xtcpcb *tfi_sotouserpcb(struct socket *so);
extern struct xtcpcb *tfi_intouserpcb(struct inpcb *inp);
extern int tfi_pcballoc(struct socket *so, int size);
extern int tfi_pcbdelete(struct socket *so);

/* mbuf functions */
/* #define tfi_m_freem m_freem */
#define tfi_mtod mtod
/* #define tfi_m_adj m_adj */
/* #define tfi_mtod mtod */
#define tfi_m_pullup tfi_mpullup
void tfi_m_free(struct mbuf *m);
void tfi_m_freem(struct mbuf *m);
void tfi_m_adj(struct mbuf *mp, int req_len);
struct mbuf *tfi_mpullup(struct mbuf *m, int len);
extern int tfi_mlen(struct mbuf *m);
extern struct mbuf *tfi_m_prepend(struct mbuf *m, int size, int how);

/* Get an mbuf with header */
extern struct mbuf * tfi_mgethdr(void);

/* set length of an mbuf. For now it only works for plen less
   than 100 (single mbuf) */
extern int tfi_msetlen(struct mbuf *m, int plen);
extern int tfi_msethlen(struct mbuf *m, int plen);

/* interrupt routines */
int tfi_splnet(void);
void tfi_splx(int s);
/*
#define tfi_splnet splnet
#define tfi_splx splx
*/
/* inet functions */
/* #define tfi_in_pcbconnect in_pcbconnect
#define tfi_in_pcbdisconnect in_pcbdisconnect
*/
extern int tfi_in_pcbconnect(struct inpcb *inp, struct sockaddr *nam,
			     struct proc *p);
extern void tfi_in_pcbdisconnect(struct inpcb *inp);
extern int tfi_in_pcbbind(struct inpcb *inp, struct sockaddr *nam, struct proc *p);

/* TFI send functions will free the mbuf, so the user
   protocol should not attempt
*/
extern int tfi_netsend(struct mbuf *m, struct inpcb *inp, seq_t seq_no);
extern int tfi_netresend(struct mbuf *m, struct inpcb *inp, seq_t seq_no);
extern int tfi_sendack(struct mbuf *m, struct inpcb *inp);
extern int tfi_acksum(seq_t seq_no, u_char nonce_sum, struct socket *so);
extern int tfi_acksum_leq(seq_t seq_no, u_char nonce_sum, struct socket *so);
extern int tfi_nack(seq_t seq_no, struct socket *so);

extern int tfi_netsend_range(struct mbuf *m, struct inpcb *inp, seq_t bseq,
			     seq_t eseq);
extern int tfi_netresend_range(struct mbuf *m, struct inpcb *inp, seq_t bseq, seq_t eseq);
extern int tfi_acksum_range(seq_t bseq, seq_t eseq, u_char nonce_sum,
			    struct socket *so);

/*
#define tfi_in_pcbdisconnect in_pcbdisconnect
#define tfi_in_pcbbind in_pcbbind
#define tfi_in_pcbconnect in_pcbconnect
*/

/* support for unsupported functions */
extern int tfi_pru_accept_notsupp(struct socket *so, struct sockaddr **nam, struct tfi_protosw_cyc *nu);
extern int tfi_pru_listen_notsupp(struct socket *so, struct proc *p, struct tfi_protosw_cyc *nu);
extern int tfi_pru_rcvd_notsupp(struct socket *so, int flags, struct tfi_protosw_cyc *nu);
extern int tfi_pru_rcvoob_notsupp(struct socket *so, struct mbuf *m, int flags, struct tfi_protosw_cyc *nu);

struct tfi_usrreqs_cyc {
	int	(*pru_abort) __P((struct socket *so));
	int	(*pru_accept) __P((struct socket *so, struct sockaddr **nam));
	int	(*pru_attach) __P((struct socket *so, int proto,
				   struct proc *p));
	/* Not needed. Handled completely by TFC */
	/* int	(*pru_bind) __P((struct socket *so, struct sockaddr *nam,
	   struct proc *p)); */

	int	(*pru_connect) __P((struct socket *so, struct sockaddr *nam,
				    struct proc *p));

	/* Not needed, anyway for now */
        /* int	(*pru_connect2) __P((struct socket *so1, struct socket *so2)); */
	/* it's always in_control */
	/* int	(*pru_control) __P((struct socket *so, u_long cmd, caddr_t data,
	   struct ifnet *ifp, struct proc *p)); */
	
	int	(*pru_detach) __P((struct socket *so));
	int	(*pru_disconnect) __P((struct socket *so));
	int	(*pru_listen) __P((struct socket *so, struct proc *p));

	/* Not needed because in_setpeeraddr will take care of it. */
	/* int	(*pru_peeraddr) __P((struct socket *so, 
	   struct sockaddr **nam)); */

	int	(*pru_rcvd) __P((struct socket *so, int flags));
	int	(*pru_rcvoob) __P((struct socket *so, struct mbuf *m,
				   int flags));
	int	(*pru_send) __P((struct socket *so, int flags, struct mbuf *m, 
				 struct sockaddr *addr, struct mbuf *control,
				 struct proc *p));
	int	(*pru_shutdown) __P((struct socket *so));	
};

/* structure that a tfc protocol will provide */
struct tfi_protosw_cyc {
	/* protocol-protocol hooks */
	int	(*pr_input) __P((struct mbuf *, int off, int proto, struct socket *));
	/* input to protocol (from below) */
	int	(*pr_output) __P((struct mbuf *m, struct socket *so));
	/* output to protocol (from above) */
	void	(*pr_ctlinput) __P((struct socket *, int));
	/* control input (from below) */
	int	(*pr_ctloutput) __P((struct socket *, struct sockopt *));
	/* control output (from above) */
	/* utility hooks */
	int	(*pr_init) __P((void));	/* initialization hook */
	void	(*pr_fasttimo) __P((void));
	/* fast timeout (200ms) */
	void	(*pr_slowtimo) __P((void));
	/* slow timeout (500ms) */
	void	(*pr_drain) __P((void));
	/* flush any excess space possible */
	struct	tfi_usrreqs_cyc *pr_usrreqs;	/* supersedes pr_usrreq() */
	int (*pr_copystate) __P((struct tfi_tcpcb *ctp, struct socket *so));
	int (*pr_synsocket) __P((struct tfi_syncache *sc, struct socket *so,
				 struct socket **nso));
};

#endif /* TFI_PROTO_H */
