/*
 * tfi_cyc.h
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * for any purpose with or without fee is hereby granted, provided
 * that the above copyright notice and this permission notice
 * appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE UNIVERSITY OF UTAH
 * DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING
 * ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO
 * EVENT SHALL THE UNIVERSITY OF UTAH BE LIABLE FOR ANY SPECIAL,
 * DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER
 * IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
 * THIS SOFTWARE.
 */
#ifndef TFI_CYC_H
#define TFI_CYC_H

typedef unsigned long seq_t;
struct xtcpcb;

/* cyclone headers */
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>

/* local headers */
#include <netinet/xtcp/ip_cyc.h>
#include <netinet/xtcp/tfc_error.h>

#define IPPROTO_XTCP 114
#define IPPROTO_TFC 115
#define TFC_ATTACH_PROTOCOL 1
#define XUDP_CYC_HASH 179 /* 179 is for XUDP cyclone code */

/* XXX */
#define abs(a) ((a) < 0 ? -(a) : (a))
#define max(a,b) ((a) >= (b) ? (a) : (b))
#define min(a,b) ((a) < (b) ? (a) : (b))

#define NONCE_SUM(a, b) (((a)+(b)) & 0x03)

// from param.h
#define	roundup(x, y)	((((x)+((y)-1))/(y))*(y))  /* to any y */
#define	roundup2(x, y)	(((x)+((y)-1))&(~((y)-1))) /* if y is powers of two */
#define powerof2(x)	((((x)-1)&(x))==0)

// from mbuf.h
/*
 * mbuf pkthdr flags (also stored in m_flags).
 */
#define	M_BCAST		0x0100	/* send/received as link-level broadcast */
#define	M_MCAST		0x0200	/* send/received as link-level multicast */
#define	M_FRAG		0x0400	/* packet is a fragment of a larger packet */
#define	M_FIRSTFRAG	0x0800	/* packet is first fragment */
#define	M_LASTFRAG	0x1000	/* packet is last fragment */

#define	CSUM_IP			0x0001		/* will csum IP */
#define	CSUM_TCP		0x0002		/* will csum TCP */
#define	CSUM_UDP		0x0004		/* will csum UDP */
#define	CSUM_IP_FRAGS		0x0008		/* will csum IP fragments */
#define	CSUM_FRAGMENT		0x0010		/* will do IP fragmentation */

#define	CSUM_IP_CHECKED		0x0100		/* did csum IP */
#define	CSUM_IP_VALID		0x0200		/*   ... the csum is valid */
#define	CSUM_DATA_VALID		0x0400		/* csum_data field is valid */
#define	CSUM_PSEUDO_HDR		0x0800		/* csum_data has pseudo hdr */

#define	CSUM_DELAY_DATA		(CSUM_TCP | CSUM_UDP)
#define	CSUM_DELAY_IP		(CSUM_IP)	/* XXX add ipv6 here too? */


#define	IP_MAXPACKET	65535		/* maximum packet size */
#define	INADDR_ANY		(u_int32_t)0x00000000 /* in.h */
#define	M_DONTWAIT	0x0001		/* do not block from malloc.h */

#define	PR_SLOWHZ	2		/* 2 slow timeouts per second */
#define	PR_FASTHZ	5		/* 5 fast timeouts per second */

/* socketvar.h */
#define	SS_NOFDREF		0x0001	/* no file table ref any more */
#define	SS_ISCONNECTED		0x0002	/* socket connected to a peer */
#define	SS_ISCONNECTING		0x0004	/* in process of connecting to peer */
#define	SS_ISDISCONNECTING	0x0008	/* in process of disconnecting */
#define	SS_CANTSENDMORE		0x0010	/* can't send more data to peer */
#define	SS_CANTRCVMORE		0x0020	/* can't receive more data from peer */
#define	SS_RCVATMARK		0x0040	/* at mark on input */

#define __P(x) x

/* Use Cyclone's sockaddr structure */
typedef struct sockaddr sockaddr_t;

typedef	u_int32_t tcp_seq;
typedef u_int32_t tcp_cc;		/* connection count per rfc1644 */

struct rmxp_tao {
	tcp_cc	tao_cc;			/* latest CC in valid SYN */
	tcp_cc	tao_ccsent;		/* latest CC sent to peer */
	u_short	tao_mssopt;		/* peer's cached MSS */
	u_short unused; // PP: added for padding
#ifdef notyet
	u_short	tao_flags;		/* cache status flags */
#define	TAOF_DONT	0x0001		/* peer doesn't understand rfc1644 */
#define	TAOF_OK		0x0002		/* peer does understand rfc1644 */
#define	TAOF_UNDEF	0		/* we don't know yet */
#endif /* notyet */
};

struct rt_metrics {
	u_long	rmx_locks;	/* Kernel must leave these values alone */
	u_long	rmx_mtu;	/* MTU for this path */
	u_long	rmx_hopcount;	/* max hops expected */
	u_long	rmx_expire;	/* lifetime for route, e.g. redirect */
	u_long	rmx_recvpipe;	/* inbound delay-bandwidth product */
	u_long	rmx_sendpipe;	/* outbound delay-bandwidth product */
	u_long	rmx_ssthresh;	/* outbound gateway buffer limit */
	u_long	rmx_rtt;	/* estimated round trip time */
	u_long	rmx_rttvar;	/* estimated rtt variance */
	u_long	rmx_pksent;	/* packets sent using this route */
	struct rmxp_tao rmx_tao;	/* will be used for T/TCP later */
};

/* BEGIN parveen */
#define	PRUS_OOB	0x1
#define	PRUS_EOF	0x2
#define	PRUS_MORETOCOME	0x4
/* END parveen */

/* BEGIN tim */

/* END tim */

/* Make a nice wrapper for the unsafe C code */
namespace UnsafeTFI {
	extern "C include" {  /* tell Cyclone that <tfi.h> is C code */
                #include <netinet/xtcp/tfia.h>  /* include the tfi C interface */
	} export {
		/* export these functions */
		tfi_downcallenter, tfi_downcallleave,
	        tfi_addproto_cyc, tfi_delproto, tfi_sbrappend, tfi_sbsappend, tfi_sbrspace,
			tfi_sbsspace, tfi_sorwakeup, tfi_sowwakeup, tfi_soseterror,
			tfi_sosetstate, tfi_sounsetstate, tfi_sotoinpcb, tfi_soreserve,
			tfi_soisconnected, tfi_sotouserpcb,
			tfi_intouserpcb, tfi_pcballoc, tfi_pcbdelete,
			tfi_m_freem, tfi_m_free, tfi_mtodfat, tfi_m_adj, tfi_mpullup, tfi_mgethdr,
			tfi_msetlen, tfi_mlen, tfi_mtoip, tfi_m_pkthdr, tfi_m_prepend,
			tfi_msetdata, tfi_msethlen,
			tfi_splnet, tfi_splx, tfi_in_pcbconnect, tfi_in_pcbdisconnect,
			tfi_in_cksum, tfi_inpladdr, tfi_inpfaddr,tfi_inplport, tfi_inpfport,
			tfi_inpttl, tfi_inptos,
			tfi_netsend, tfi_netsend_range, tfi_sendack, tfi_acksum, tfi_acksum_leq, tfi_acksum_range, tfi_nack,
			tfi_in_pcbbind, tfi_in_pseudo,
			tfi_htonl, tfi_ntohl, tfi_htons, tfi_ntohs,
			tfi_printf, printf, hz, ticks,
			tfi_getrtmx,
			max_protohdr, max_linkhdr, tfi_mhlen, panic,
			ip_defttl, callout_init, callout_reset, callout_stop,
			read_random_unlimited,
			MD5Init, MD5Update, MD5Final,
			// BEGIN -- Tim
			tfi_sonewconn, tfi_sonewconn1, tfi_soabort,
			ip_srcroute,
			ip_output,
			in_losing, in_pcbinshash,
			tfi_sogetoptions,
			tfi_getrt_refcnt, tfi_getrt_flags, tfi_getrt_gateway,
			arc4random,
			rtrequest,
			tfi_rt_key, tfi_rt_mask, tfi_rt_free,
			tfi_minithdr,
			tfi_pru_accept_notsupp,
			// BEBIN -- Parveen
			tfi_soisdisconnected,
			tfi_soshiwat, tfi_sorhiwat, tfi_rtlookup, tfi_isrtup, tfi_issolinger,
			tfi_setsolinger, tfi_sbrflush, tfi_soisconnecting,
			tfi_soisdisconnecting, tfi_mflags, tfi_in_broadcast, 
			tfi_socantsendmore, tfi_sogetscc, tfi_sogetrcc, tfi_in_pcbladdr,
			tfi_in_pcblookup_hash, tfi_in_pcbrehash, tfi_getsostate,
			tfi_sbsreserve, tfi_sbrreserve,
			tfi_rtgetmtu, tfi_socantrcvmore, tfi_mgetnext, 
			tfi_sohasoutofband, sb_max,
			tfi_sbsdrop, tfi_m_copydata, tfi_m_copy, tfi_m_setnext,
			tfi_in_addword, tfi_netresend, tfi_netresend_range,
			tfi_in_localaddr;
	}
	/*
	extern "C include" {
                #include <netinet/xtcp/tfc_error.h>
	} export {
		TFC_NOERROR, TFC_ENOSUPPORT, TFC_EINUSE,
			TFC_INVALIDPROTO,
			TFC_EXISTS, TFC_ENOMEM, TFC_EARGS, TFC_EISCONN, TFC_EINVAL,
			TFC_ERROR, TFI_ERROR, TFC_NOPROTO, TFC_BADNONCE, TFC_EINVALSEQ,
			TFC_WAIT, TFC_GOAWAY, TFC_SBOF
			;
	}
	*/
}

typedef struct UnsafeTFI::tfi_protosw tfi_protosw_t;
typedef struct UnsafeTFI::tfi_usrreqs tfi_usrreqs_t;
typedef struct UnsafeTFI::tfi_proto tfi_proto_t;
typedef struct UnsafeTFI::socket socket_t;
typedef struct UnsafeTFI::mbuf mbuf_t;
typedef struct UnsafeTFI::m_pkthdr m_pkthdr_t;
typedef struct UnsafeTFI::inpcb inpcb_t;
typedef struct UnsafeTFI::proc proc_t;
typedef struct UnsafeTFI::callout callout_t;
typedef struct UnsafeTFI::ifnet ifnet_t;
typedef struct UnsafeTFI::tfi_tcpcb tfi_tcpcb_t;
typedef struct UnsafeTFI::tfi_syncache tfi_syncache_t;

// For now #define these. Later on there should be checks to ensure that these variables can only be modified in certain limits.
#define hz UnsafeTFI::hz
#define ticks UnsafeTFI::ticks
#define sb_max UnsafeTFI::sb_max
#define max_protohdr UnsafeTFI::max_protohdr
#define max_linkhdr UnsafeTFI::max_linkhdr
#define ip_defttl UnsafeTFI::ip_defttl
#define tfi_mhlen UnsafeTFI::tfi_mhlen

// ensure that len is <= 32 in calls to read_random
#define read_random_unlimited UnsafeTFI::read_random_unlimited

// functions
#define SAFE_UNSAFE(x) #x UnsafeTFI::#x

#define tfi_pru_accept_notsupp UnsafeTFI::tfi_pru_accept_notsupp

// Callout functions
#define callout_init UnsafeTFI::callout_init
#define callout_reset UnsafeTFI::callout_reset
#define callout_stop UnsafeTFI::callout_stop
#define	callout_pending(c) ((c)->c_flags & CALLOUT_PENDING)

#define panic(x) UnsafeTFI::panic(x)

#ifdef TFC_MASK
#define __P(x) x
/*
  o struct socket should be treated as opaque object by the user protocol.
  o struct inp should be treated read-only.
*/
struct tfi_usrreqs {
	int	(*pru_abort) __P((struct socket @so));
	int	(*pru_accept) __P((struct socket @so, struct sockaddr **nam));
	int	(*pru_attach) __P((struct socket @so, int proto,
				   struct proc *p));
	/* Not needed. Handled completely by TFC */
	/* int	(*pru_bind) __P((struct socket *so, struct sockaddr *nam,
	   struct proc *p)); */

	int	(*pru_connect) __P((struct socket @so, struct sockaddr *nam,
				    struct proc *p));

	/* Not needed, anyway for now */
        /* int	(*pru_connect2) __P((struct socket *so1, struct socket *so2)); */
	/* it's always in_control */
	/* int	(*pru_control) __P((struct socket *so, unsigned long cmd, caddr_t data,
	   struct ifnet *ifp, struct proc *p)); */
	
	int	(*pru_detach) __P((struct socket @so));
	int	(*pru_disconnect) __P((struct socket @so));
	int	(*pru_listen) __P((struct socket @so, struct proc *p));

	/* Not needed because in_setpeeraddr will take care of it. */
	/* int	(*pru_peeraddr) __P((struct socket *so, 
	   struct sockaddr **nam)); */

	int	(*pru_rcvd) __P((struct socket @so, int flags));
	int	(*pru_rcvoob) __P((struct socket @so, struct mbuf *m,
				   int flags));
	int	(*pru_send) __P((struct socket @so, int flags, struct mbuf @m, 
				 struct sockaddr *addr, struct mbuf *control,
				 struct proc *p));
#define	TFI_PRUS_OOB	0x1
#define	TFI_PRUS_EOF	0x2
#define	TFI_PRUS_MORETOCOME	0x4

	int	(*pru_shutdown) __P((struct socket @so));	
};

/* structure that a tfc protocol will provide */
struct tfi_protosw {
	short	pr_type;		/* socket type used for */
	short	pr_protocol;		/* protocol number */
	short	pr_flags;		/* see below */
	/* protocol-protocol hooks */
	int	(*pr_input) __P((struct mbuf @, int off, int proto, struct socket @));
					/* input to protocol (from below) */
	int	(*pr_output) __P((struct mbuf @m, struct socket *so));
					/* output to protocol (from above) */
	void	(*pr_ctlinput) __P((struct socket @, int));
					/* control input (from below) */
	int	(*pr_ctloutput) __P((struct socket @, struct sockopt *));
					/* control output (from above) */
	/* utility hooks */
	int	(*pr_init) __P((void));	/* initialization hook */
	void	(*pr_fasttimo) __P((void));
					/* fast timeout (200ms) */
	void	(*pr_slowtimo) __P((void));
					/* slow timeout (500ms) */
	void	(*pr_drain) __P((void));
					/* flush any excess space possible */
	struct	tfi_usrreqs @pr_usrreqs;	/* supersedes pr_usrreq() */
};

/* Opaque handle for the user protocol */
struct tfi_proto {
	const struct tfc_proto *p;
};

#endif /* TFC_MASK */

#define TFC_DISCONNECT 1

extern int tfi_addproto(unsigned short domain, unsigned long hash,
			tfi_protosw_t @sw, tfi_proto_t @pr);

extern int tfi_delproto(tfi_proto_t @pr);

/* For now these will just map to the original functions */
/* socket functions */
extern int tfi_sbrappend(socket_t @so, mbuf_t @m);
extern int tfi_sbsappend(socket_t @so, mbuf_t @m);
extern long tfi_sbrspace(socket_t @so);
extern long tfi_sbsspace(socket_t @so);
extern void tfi_sorwakeup(socket_t @so);
extern void tfi_sowwakeup(socket_t @so);
extern void tfi_soseterror(socket_t @so, int err);
extern void tfi_sosetstate(socket_t @so, short state);
extern void tfi_sounsetstate(socket_t @so, short state);
extern inpcb_t *tfi_sotoinpcb(socket_t @so);
extern int tfi_soreserve(socket_t @so, unsigned long sndcc, unsigned long rcvcc);
extern void tfi_soisconnected(socket_t @so);
extern void tfi_soisdisconnected(socket_t @so);
extern void tfi_soisconnecting(socket_t @so);
extern void tfi_soisdisconnecting(socket_t @so);
extern void tfi_sbrreserve(socket_t @so, u_long cc);
extern int tfi_issolinger(socket_t @so);
extern void tfi_setsolinger(socket_t @so, int period);
extern void tfi_sbrflush(socket_t @so);
extern int tfi_sogetscc(socket_t @so);
extern int tfi_sogetrcc(socket_t @so);
extern void tfi_socantsendmore(socket_t @so);
extern short tfi_sogetoptions(socket_t @so);
extern int tfi_sorhiwat(socket_t @so);
extern int tfi_soshiwat(socket_t @so);
extern short tfi_getsostate(socket_t @so);
extern void tfi_sbsreserve(socket_t @so, u_long cc);
extern void tfi_socantrcvmore(socket_t @so);
extern void tfi_sbsdrop(socket_t @so, int len);
extern void tfi_sohasoutofband(socket_t @so);

/* utility functions */
/* This should be dynamically defined so that its return type matches a pcb for
   each user protocol */
extern struct xtcpcb *tfi_sotouserpcb(socket_t @so);

/* same as above */
extern struct xtcpcb *tfi_intouserpcb(inpcb_t @inp);

extern int tfi_pcballoc(socket_t @so, int size);
extern int tfi_pcbdelete(socket_t @so);

extern struct rt_metrics *tfi_getrtmx(struct rtentry *rte);
extern int tfi_rtgetmtu(struct rtentry *);

/* mbuf functions */
extern void tfi_m_free(mbuf_t @m);
extern void tfi_m_freem(mbuf_t @m);
extern char ? @nozeroterm tfi_mtod(mbuf_t @m);
extern void tfi_m_adj(mbuf_t @mp, int req_len);
extern mbuf_t *tfi_mpullup(mbuf_t @m, int len);
extern mbuf_t *tfi_mgethdr();
extern int tfi_mlen(mbuf_t @m);
struct ip *tfi_mtoip(mbuf_t @m);
extern m_pkthdr_t *tfi_m_pkthdr(mbuf_t @m);
extern char ?tfi_m_copydata(socket_t @so, int off, int len, char ?@nozeroterm data);
extern mbuf_t *tfi_m_copy(socket_t @so, int off, int len);
extern void tfi_m_setnext(mbuf_t @m, mbuf_t @n);


/* set length of an mbuf. For now it only works for plen less
   than 100 (single mbuf) */
extern int tfi_msetlen(mbuf_t @m, int plen);
extern mbuf_t *tfi_m_prepend(mbuf_t *m, int size, int how);
extern int tfi_msetdata(mbuf_t @m, void *data);
extern int tfi_msethlen(mbuf_t @m, int plen);
extern short tfi_mflags(mbuf_t @m);
extern mbuf_t *tfi_mgetnext(mbuf_t @m);

/* interrupt routines */
extern int tfi_splnet();
extern void tfi_splx(int s);
extern struct xudphdr *tfi_strtoxudphdr(char ? @nozeroterm data);

/* inet functions */
extern void tfi_in_pcbdisconnect(inpcb_t *inp);
extern int tfi_in_pcbbind(inpcb_t *inp, sockaddr_t *nam, proc_t *p);
extern u_short tfi_in_addword(u_short sum, u_short w);
extern int tfi_in_pcbconnect(inpcb_t *inp, struct sockaddr_in *nam, proc_t *p);
extern int tfi_in_cksum(mbuf_t *m, int len);
extern void tfi_in_pcbrehash(inpcb_t *inp);
extern inpcb_t *tfi_in_pcblookup_hash(inpcb_t *inp, struct in_addr faddr, u_int fport, struct in_addr laddr, u_int lport, u_int wildcard, ifnet_t *ifp);
extern int tfi_in_pcbladdr(inpcb_t *inp, struct sockaddr_in *nam, struct sockaddr_in **ifaddr);
extern int tfi_in_broadcast(struct in_addr in, ifnet_t *ifp);
extern int tfi_in_localaddr(struct in_addr in);

extern struct in_addr tfi_inpladdr(inpcb_t *inp);

extern struct in_addr tfi_inpfaddr(inpcb_t *inp);

extern u_int16_t tfi_inplport(inpcb_t *inp);

extern u_int16_t tfi_inpfport(inpcb_t *inp);

extern u_short tfi_in_pseudo(u_int32_t a, u_int32_t b, u_int32_t c);

extern u_int16_t tfi_inpttl(inpcb_t *inp);

extern u_int16_t tfi_inptos(inpcb_t *inp);


/* TFI send functions will free the mbuf, so the user
   protocol should not attempt
*/
extern int tfi_netsend(mbuf_t @m, inpcb_t @inp, seq_t seq_no);
extern int tfi_netresend(mbuf_t @m, inpcb_t @inp, seq_t seq_no);
extern int tfi_netsend_range(mbuf_t @m, inpcb_t @inp, seq_t bseq, seq_t eseq);
extern int tfi_netresend_range(mbuf_t @m, inpcb_t @inp, seq_t bseq, seq_t eseq);
extern int tfi_sendack(mbuf_t @m, inpcb_t @inp);
extern int tfi_acksum(seq_t seq_no, unsigned char nonce_sum, socket_t @so);
extern int tfi_acksum_leq(seq_t seq_no, unsigned char nonce_sum, socket_t @so);
extern int tfi_acksum_range(seq_t bseq, seq_t eseq, unsigned char nonce_sum, socket_t @so);
extern int tfi_nack(seq_t seq_no, socket_t @so);

/* some helper functions */
int uccopy(unsigned char ?@nozeroterm src, unsigned char ?@nozeroterm dst, int len);
int uczero(unsigned char ?@nozeroterm str, int len);
u_int32_t htonl(u_int32_t x);
u_int16_t htons(u_int16_t x);
u_int32_t ntohl(u_int32_t x);
u_int16_t ntohs(u_int16_t x);

extern int pru_accept_notsupp(socket_t *so, struct sockaddr **nam);
extern int pru_listen_notsupp(socket_t *so, proc_t *p);
extern int pru_rcvd_notsupp(socket_t *so, int flags);
extern int pru_rcvoob_notsupp(socket_t *so, mbuf_t *m, int flags);


/* BEGIN parveen */
/* from <i386/include/param.h> */

#ifndef	MCLSHIFT
#define MCLSHIFT	11		/* convert bytes to m_buf clusters */
#endif	/* MCLSHIFT */

#define MCLBYTES	(1 << MCLSHIFT)	/* size of an m_buf cluster */
#define MCLOFSET	(MCLBYTES - 1)	/* offset within an m_buf cluster */

#define tfi_sotoxtcpcb tfi_sotouserpcb
extern int imax(int a, int b);
int ulmin(u_long a, u_long b);
/* END parveen */

/* BEGIN tim */

int ip_output(mbuf_t *, mbuf_t *, struct conn_route *, int, void *);

/* END tim */

#endif /* TFI_CYC_H */
