/*
 * tfi_pcb.c
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * for any purpose with or without fee is hereby granted, provided
 * that the above copyright notice and this permission notice
 * appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE UNIVERSITY OF UTAH
 * DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING
 * ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO
 * EVENT SHALL THE UNIVERSITY OF UTAH BE LIABLE FOR ANY SPECIAL,
 * DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER
 * IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
 * THIS SOFTWARE.
 */
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/malloc.h>
#include <sys/kernel.h>
#include <sys/sysctl.h>
#include <sys/errno.h>
#include <sys/mbuf.h>
#include <sys/socket.h>
#include <sys/socketvar.h>
#include <sys/protosw.h>

#include <vm/vm_zone.h> /* memory for tfc_cb and the likes */

#include <net/if.h>
#include <net/route.h>

#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/in_pcb.h>
#include <netinet/in_var.h>
#include <netinet/ip.h>
#include <netinet/ip_var.h>
#include <netinet/tcp.h>
#include <netinet/tcp_fsm.h>
#include <netinet/tcp_seq.h>
#include <netinet/tcp_timer.h>
#include <netinet/tcp_var.h>
#include <netinet/tcpip.h>

#include <netinet/xtcp/tfc.h>
#include <netinet/xtcp/tfci.h>
#include <netinet/xtcp/tfi.h>
#include <netinet/xtcp/tfc_stats.h>
#include <netinet/xtcp/tfc_error.h>
#include <netinet/xtcp/lossfraction.h>
#include <netinet/xtcp/scoreboard.h>
#include <netinet/xtcp/tfc_cb.h>


/* User PCB functions */
#include <netinet/xtcp/tfi_pcb.h>


/*
 * Should all operations involving pcb be run at splnet?
 */
int
tfi_pcballoc(struct socket *so, int size)
{
	int s = splnet();
	int error = 0;
	struct tfc_cb *tp;
	void *pcb;
	MALLOC(pcb, void *, size, M_TFCPROTO, M_NOWAIT);
	if (pcb == NULL) {
		error = ENOBUFS;
		goto out;
	}
	bzero(pcb, size);
	
	tp = sototfccb(so);
	if (tp == NULL) {
		error = EINVAL;
		goto release;
	}
	tp->pcb = pcb;
	goto out;
	
 release:
	if (pcb)
		FREE(pcb, M_TFCPROTO);
 out:
	splx(s);
	return error;
}

int
tfi_pcbdelete(struct socket *so)
{
	int error = 0;
	struct tfc_cb *tp;

	if (sotoinpcb(so) == NULL) {
		error = EINVAL;
		goto out;
	}
	tp = sototfccb(so);
	if (tp == NULL || tp->pcb == NULL) {
		error = EINVAL;
		goto out;
	}
	FREE(tp->pcb, M_TFCPROTO);
	tp->pcb = NULL;
 out:
	return error;
}

int
tfi_pcbfree(void *state)
{
	FREE(state, M_TFCPROTO);
	return 0;
}
