/*
 * tfia.h
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * for any purpose with or without fee is hereby granted, provided
 * that the above copyright notice and this permission notice
 * appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE UNIVERSITY OF UTAH
 * DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING
 * ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO
 * EVENT SHALL THE UNIVERSITY OF UTAH BE LIABLE FOR ANY SPECIAL,
 * DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER
 * IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
 * THIS SOFTWARE.
 */
#ifndef TFIA_PROTO_H
#define TFIA_PROTO_H

/* cyclone sys/types.h> */
#include <sys/types.h>

struct mbuf;
// struct socket;
struct proc;
/* struct sockaddr; Cyclone has its own */
/* extern struct xtcpcb; */
// struct sockopt;
struct tfc_proto;
struct ifnet;

// We have to ensure that the protocol code cannot write its fields.
// Therefore, all functions sould be changed from socket * to const socket *  
struct socket {
	short	so_type;		/* generic type, see socket.h */
	short	so_options;		/* from socket call, see socket.h */
	short	so_linger;		/* time to linger while closing */
	short	so_state;		/* internal state flags SS_*, below */
	void   	*so_pcb;		/* protocol control block */
	void    *so_proto;	/* protocol handle */
/*
 * Variables for connection queuing.
 * Socket where accepts occur is so_head in all subsidiary sockets.
 * If so_head is 0, socket is not related to an accept.
 * For head socket so_incomp queues partially completed connections,
 * while so_comp is a queue of connections ready to be accepted.
 * If a connection is aborted and it has so_head set, then
 * it has to be pulled out of either so_incomp or so_comp.
 * We allow connections to queue up based on current queue lengths
 * and limit on number of queued connections for this socket.
 */
	struct	socket *so_head;	/* back pointer to accept socket */
	struct {
	    void *a;
	    void *b;
	} so_incomp;	/* queue of partial unaccepted connections */
	struct {
	    void *a;
	    void *b;
	} so_comp;	/* queue of complete unaccepted connections */
	struct {
	    void *a;
	    void *b;
	} so_list;	/* list of unaccepted connections */
	short	so_qlen;		/* number of unaccepted connections */
	short	so_incqlen;		/* number of unaccepted incomplete
					   connections */
	short	so_qlimit;		/* max number queued connections */
	short	so_timeo;		/* connection timeout */
	u_short	so_error;		/* error affecting connection */
	void *so_sigio;	/* information for async I/O or
					   out of band data (SIGURG) */
	u_long	so_oobmark;		/* chars to oob mark */
	struct {
	    void *a;
	    void *b;
	} so_aiojobq; /* AIO ops waiting on socket */
};

// Placeholder for callout. Taken from callout.h
struct callout {
	struct {
		void *a;
		void *b;
	} link;
	int	c_time;				/* ticks to the event */
	void	*c_arg;				/* function argument */
	void	(*c_func) __P((void *));	/* function to call */
	int	c_flags;			/* state of this entry */
};

#define	CALLOUT_LOCAL_ALLOC	0x0001 /* was allocated from callfree */
#define	CALLOUT_ACTIVE		0x0002 /* callout is currently active */
#define	CALLOUT_PENDING		0x0004 /* callout is waiting for timeout */

struct callout_handle {
	struct callout *callout;
};

#define	callout_active(c)	((c)->c_flags & CALLOUT_ACTIVE)
#define	callout_deactivate(c)	((c)->c_flags &= ~CALLOUT_ACTIVE)
void	callout_init __P((struct callout *));
#define	callout_pending(c)	((c)->c_flags & CALLOUT_PENDING)
// redefine in terms of a type variable
void	callout_reset __P((struct callout *, int, void (*)(`a), `a));
int	callout_stop __P((struct callout *));

/* mbuf pkthdr for cyclone code */
struct m_pkthdr {
	struct ifnet *rcvif;		/* rcv interface */
	int	len;			/* total packet length */
	/* variables for ip and tcp reassembly */
	void	*header;		/* pointer to packet header */
	/* variables for hardware checksum */
	int	csum_flags;		/* flags regarding checksum */
	int	csum_data;		/* data field used by csum routines */
	void *aux;		/* extra data buffer; ipsec/others */
};

struct inpcb {
	struct {
		void *a;
		void *b;
	} inp_hash; /* hash list */
	struct {
		void *a;
		void *b;
	} inp_list; /* list for all PCBs of this proto */
	u_int32_t	inp_flow;

	/* local and foreign ports, local and foreign addr */
	struct	in_conninfo inp_inc;
	
	void *inp_ppcb;	/* pointer to per-protocol pcb */
	void *inp_pcbinfo;	/* PCB list info */
	struct	socket *inp_socket;	/* back pointer to socket */
	/* list for this PCB's local port */
	int	inp_flags;		/* generic IP/datagram flags */
	
	void *inp_sp; /* for IPSEC */
	u_char	inp_vflag;
#define	INP_IPV4	0x1
#define	INP_IPV6	0x2
	u_char	inp_ip_ttl;		/* time to live proto */
	u_char	inp_ip_p;		/* protocol proto */
	
	/* protocol dependent part; options */
	struct {
		u_char	inp4_ip_tos;		/* type of service proto */
		struct	mbuf *inp4_options;	/* IP options */
		void *inp4_moptions; /* IP multicast options */
	} inp_depend4;
#define inp_fport	inp_inc.inc_fport
#define inp_lport	inp_inc.inc_lport
#define	inp_faddr	inp_inc.inc_faddr
#define	inp_laddr	inp_inc.inc_laddr
#define	inp_route	inp_inc.inc_route
#define	inp_ip_tos	inp_depend4.inp4_ip_tos
#define	inp_options	inp_depend4.inp4_options
#define	inp_moptions	inp_depend4.inp4_moptions
	struct {
		/* IP options */
		void *inp6_options;
		/* IP6 options for outgoing packets */
		void *inp6_outputopts;
		/* IP multicast options */
		void *inp6_moptions;
		/* ICMPv6 code type filter */
		void *inp6_icmp6filt;
		/* IPV6_CHECKSUM setsockopt */
		int	inp6_cksum;
		u_short	inp6_ifindex;
		short	inp6_hops;
		u_int8_t	inp6_hlim;
	} inp_depend6;
	struct {
		void *a;
		void *b;
	} inp_portlist;
	void *inp_phd;	/* head of this list */
	inp_gen_t	inp_gencnt;	/* generation count of this instance */
};

struct	xtcp_tsegqe_head_dummy_t {
	void *unused;
};

struct tfi_tcpcb {
	struct	xtcp_tsegqe_head_dummy_t t_segq;
	int	t_dupacks;		/* consecutive dup acks recd */
	void	*unused;	/* unused */

	void* tt_rexmt;	/* retransmit timer */
	void* tt_persist;	/* retransmit persistence */
	void* tt_keep;	/* keepalive */
	void* tt_2msl;	/* 2*msl TIME_WAIT timer */
	void* tt_delack;	/* delayed ACK timer */

	struct inpcb *t_inpcb;		/* back pointer to internet pcb */
	int	t_state;		/* state of this connection */
	u_int	t_flags;

	int	t_force;		/* 1 if forcing out a byte */

	tcp_seq	snd_una;		/* send unacknowledged */
	tcp_seq	snd_max;		/* highest sequence number sent;
					 * used to recognize retransmits
					 */
	tcp_seq	snd_nxt;		/* send next */
	tcp_seq	snd_up;			/* send urgent pointer */

	tcp_seq	snd_wl1;		/* window update seg seq number */
	tcp_seq	snd_wl2;		/* window update seg ack number */
	tcp_seq	iss;			/* initial send sequence number */
	tcp_seq	irs;			/* initial receive sequence number */

	tcp_seq	rcv_nxt;		/* receive next */
	tcp_seq	rcv_adv;		/* advertised window */
	u_long	rcv_wnd;		/* receive window */
	tcp_seq	rcv_up;			/* receive urgent pointer */

	u_long	snd_wnd;		/* send window */
	u_long	snd_cwnd;		/* congestion-controlled window */
	u_long	snd_bwnd;		/* bandwidth-controlled window */
	u_long	snd_ssthresh;		/* snd_cwnd size threshold for
					 * for slow start exponential to
					 * linear switch
					 */
	u_long	snd_bandwidth;		/* calculated bandwidth or 0 */
	tcp_seq	snd_recover;		/* for use in fast recovery */

	u_int	t_maxopd;		/* mss plus options */

	u_long	t_rcvtime;		/* inactivity time */
	u_long	t_starttime;		/* time connection was established */
	int	t_rtttime;		/* round trip time */
	tcp_seq	t_rtseq;		/* sequence number being timed */

	int	t_bw_rtttime;		/* used for bandwidth calculation */
	tcp_seq	t_bw_rtseq;		/* used for bandwidth calculation */

	int	t_rxtcur;		/* current retransmit value (ticks) */
	u_int	t_maxseg;		/* maximum segment size */
	int	t_srtt;			/* smoothed round-trip time */
	int	t_rttvar;		/* variance in round-trip time */

	int	t_rxtshift;		/* log(2) of rexmt exp. backoff */
	u_int	t_rttmin;		/* minimum rtt allowed */
	u_int	t_rttbest;		/* best rtt we've seen */
	u_long	t_rttupdated;		/* number of times rtt sampled */
	u_long	max_sndwnd;		/* largest window peer has offered */

	int	t_softerror;		/* possible error not yet reported */
/* out-of-band data */
	char	t_oobflags;		/* have some */
	char	t_iobc;			/* input character */

/* RFC 1323 variables */
	u_char	snd_scale;		/* window scaling for send window */
	u_char	rcv_scale;		/* window scaling for recv window */
	u_char	request_r_scale;	/* pending window scaling */
	u_char	requested_s_scale;
	u_long	ts_recent;		/* timestamp echo data */

	u_long	ts_recent_age;		/* when last updated */
	tcp_seq	last_ack_sent;
/* RFC 1644 variables */
	tcp_cc	cc_send;		/* send connection count */
	tcp_cc	cc_recv;		/* receive connection count */
	/* experimental */
	u_long	snd_cwnd_prev;		/* cwnd prior to retransmit */
	u_long	snd_ssthresh_prev;	/* ssthresh prior to retransmit */
	u_long	t_badrxtwin;		/* window for retransmit recovery */

	/* AJW: Added to avoid sharing with the inp_tp structure */
	struct	callout  inp_tp_persist, inp_tp_keep, inp_tp_2msl;
	struct	callout inp_tp_delack, inp_tp_rexmt;


	struct	callout* tt_bogus;     
	struct callout inp_tp_bogus;

	/* Per-socket nonce sum */
	u_char nonce_sum;

	/* Next TFC sequence number to allocate */
	/*  seq_t next_tfc_seq; */

	/* First TFC sequence number to ACK */
	/*  seq_t first_tfc_seq;*/

	/* 
	 * Largest TFI nonce sequence number; needed to differentiate
	 * between sends and re-sends
	 */
     
	tcp_seq largest_nonce_seq;

	/* Hash to return to the listener side */
	u_long t_tfchash;
};

struct xtcp_syncache_dummy_t {
	void *a, *b;
};
/* The next two structures are used for connection negotiation */ 
struct tfi_syncache {
	inp_gen_t       sc_inp_gencnt;		/* pointer check */
	struct tfi_tcpcb  *sc_tp;		/* tcb for listening socket */
	struct		mbuf *sc_ipopts;	/* source route */
	struct 		in_conninfo sc_inc;	/* addresses */
#define sc_route	sc_inc.inc_route
#define sc_route6	sc_inc.inc6_route
	u_int32_t	sc_tsrecent;
	tcp_cc		sc_cc_send;		/* holds CC or CCnew */
	tcp_cc		sc_cc_recv;
	tcp_seq 	sc_irs;			/* seq from peer */
	tcp_seq 	sc_iss;			/* our ISS */
	u_long		sc_rxttime;		/* retransmit time */
	u_int16_t	sc_rxtslot; 		/* retransmit counter */
	u_int16_t	sc_peer_mss;		/* peer's MSS */
	u_int16_t	sc_wnd;			/* advertised window */
	u_int8_t 	sc_requested_s_scale:4,
			sc_request_r_scale:4;
	u_int8_t	sc_flags;
#define SCF_NOOPT	0x01			/* no TCP options */
#define SCF_WINSCALE	0x02			/* negotiated window scaling */
#define SCF_TIMESTAMP	0x04			/* negotiated timestamps */
#define SCF_CC		0x08			/* negotiated CC */
#define SCF_SACK_PERMIT	0x40			/* negotiated SACK */
#define SCF_UNREACH	0x10			/* icmp unreachable received */
#define SCF_KEEPROUTE	0x20			/* keep cloned route */
#define SCF_TFCNEGOTIATE_USE	0x80		/* remote end requested a protocol */
	struct xtcp_syncache_dummy_t sc_hash, sc_timerq;
        u_char          sc_nonce_sum;
	u_long          sc_tiwin;
        u_long          sc_resends;             /* How many resends? */
	u_long          sc_rhash;             /* hash of the requested protocol */
	u_long          sc_lhash;             /* hash of the locally acceptable protocol */
};


/* typedef unsigned int seq_t;*/
#define __P(x) x

/*
  o struct socket should be treated as opaque object by the user protocol.
  o struct inp should be treated read-only.
*/
struct tfi_usrreqs {
	int	(@pru_abort) __P((struct socket *so));
	int	(@pru_accept) __P((struct socket *so, struct sockaddr **nam));
	int	(@pru_attach) __P((struct socket *so, int proto,
				   struct proc *p));
	/* Not needed. Handled completely by TFC */
	/* int	(*pru_bind) __P((struct socket *so, struct sockaddr *nam,
	   struct proc *p)); */

	int	(@pru_connect) __P((struct socket *so, struct sockaddr_in *nam,
				    struct proc *p));

	/* Not needed, anyway for now */
        /* int	(*pru_connect2) __P((struct socket *so1, struct socket *so2)); */
	/* it's always in_control */
	/* int	(*pru_control) __P((struct socket *so, u_long cmd, caddr_t data,
	   struct ifnet *ifp, struct proc *p)); */
	
	int	(@pru_detach) __P((struct socket *so));
	int	(@pru_disconnect) __P((struct socket *so));
	int	(@pru_listen) __P((struct socket *so, struct proc *p));

	/* Not needed because in_setpeeraddr will take care of it. */
	/* int	(*pru_peeraddr) __P((struct socket *so, 
	   struct sockaddr **nam)); */

	int	(@pru_rcvd) __P((struct socket *so, int flags));
	int	(@pru_rcvoob) __P((struct socket *so, struct mbuf *m,
				   int flags));
	int	(@pru_send) __P((struct socket *so, int flags, struct mbuf *m, 
				 struct sockaddr_in *addr, struct mbuf *control,
				 struct proc *p));
#define	TFI_PRUS_OOB	0x1
#define	TFI_PRUS_EOF	0x2
#define	TFI_PRUS_MORETOCOME	0x4

	int	(@pru_shutdown) __P((struct socket *so));	
};

/* structure that a tfc protocol will provide */
struct tfi_protosw {
	/* protocol-protocol hooks */
	int	(@pr_input) __P((struct mbuf *, int off, int proto, struct socket *));
	/* input to protocol (from below) */
	int	(*pr_output) __P((struct mbuf *m, struct socket *so));
	/* output to protocol (from above) */
	void	(*pr_ctlinput) __P((struct socket *, int));
	/* control input (from below) */
	int	(*pr_ctloutput) __P((struct socket *, datatype SockOpt *));
	/* control output (from above) */
	/* utility hooks */
	int	(*pr_init) __P((void));	/* initialization hook */
	void	(*pr_fasttimo) __P((void));
	/* fast timeout (200ms) */
	void	(*pr_slowtimo) __P((void));
	/* slow timeout (500ms) */
	void	(*pr_drain) __P((void));
	/* flush any excess space possible */
	struct	tfi_usrreqs *pr_usrreqs;	/* supersedes pr_usrreq() */
	int (*pr_copystate) __P((struct tfi_tcpcb *ctp, struct socket *so));
	int (*pr_synsocket) __P((struct tfi_syncache *, struct socket @so,
				 struct socket **nso));
};

struct tfc_proto;

/* Opaque handle for the user protocol */
struct tfi_proto {
	struct tfc_proto *p;
};

#define TFC_DISCONNECT 1

extern void tfi_downcallenter(void);
extern void tfi_downcallleave(void);

extern int tfi_addproto_cyc(unsigned short domain, unsigned long hash, struct tfi_protosw *sw,
		   struct tfi_proto *pr);

extern int tfi_delproto(struct tfi_proto *pr);

/* For now these will just map to the original functions */
/* socket functions */
/* #define tfi_sbappendaddr sbappendaddr */
extern int tfi_sbrappend(struct socket *so, struct mbuf *m);
extern int tfi_sbsappend(struct socket *so, struct mbuf *m);
extern long tfi_sbrspace(struct socket *so);
extern long tfi_sbsspace(struct socket *so);

/* #define tfi_sbsspace(so) sbspace(&(so)->so_snd)*/
extern void tfi_sorwakeup(struct socket *so);
extern void tfi_sowwakeup(struct socket *so);
extern void tfi_soseterror(struct socket *so, int err);
extern void tfi_sosetstate(struct socket *so, short state);
extern void tfi_sounsetstate(struct socket *so, short state);
extern struct inpcb *tfi_sotoinpcb(struct socket *so);
extern int tfi_soreserve(struct socket *so, unsigned long sndcc, unsigned long rcvcc);
extern void tfi_soisconnected(struct socket *so);

extern int tfi_pcballoc(struct socket *so, int size);
extern int tfi_pcbdelete(struct socket *so);
extern struct xtcpcb *tfi_sotouserpcb(struct socket *so);
extern struct xtcpcb *tfi_intouserpcb(struct inpcb *inp);


/* mbuf functions */
void tfi_m_free(struct mbuf *m);
void tfi_m_freem(struct mbuf *m);
extern char ? @nozeroterm tfi_mtodfat(struct mbuf *m);
void tfi_m_adj(struct mbuf *mp, int req_len);
struct mbuf *tfi_mpullup(struct mbuf *m, int len);
extern int tfi_mlen(struct mbuf *m);
extern struct ip *tfi_mtoip(struct mbuf *m);

/* Get an mbuf with header */
extern struct mbuf *tfi_mgethdr(void);
/* set length of an mbuf. For now it only works for plen less
   than 100 (single mbuf) */
extern int tfi_msetlen(struct mbuf *m, int plen);
extern struct m_pkthdr* tfi_m_pkthdr(struct mbuf *m);
extern struct mbuf *tfi_m_prepend(struct mbuf *m, int size, int how);
extern int tfi_msetdata(struct mbuf @m, void *data);

/* interrupt routines */
int tfi_splnet();
void tfi_splx(int s);

/* inet functions */
extern int tfi_in_pcbconnect(struct inpcb *inp, struct sockaddr_in *nam,
			     struct proc *p);
extern void tfi_in_pcbdisconnect(struct inpcb *inp);
extern int tfi_in_pcbbind(struct inpcb *inp, struct sockaddr *nam, struct proc *p);
extern int tfi_in_cksum(struct mbuf *m, int len);
extern u_short tfi_in_pseudo(u_int32_t a, u_int32_t b, u_int32_t c);

extern struct in_addr tfi_inpladdr(struct inpcb *inp);
extern struct in_addr tfi_inpfaddr(struct inpcb *inp);
extern u_int16_t tfi_inplport(struct inpcb *inp);
extern u_int16_t tfi_inpfport(struct inpcb *inp);
extern u_char tfi_inpttl(struct inpcb *inp);
extern u_char tfi_inptos(struct inpcb *inp);

/* TFI send functions will free the mbuf, so the user
   protocol should not attempt
*/
extern int tfi_netsend(struct mbuf *m, struct inpcb *inp, seq_t seq_no);
extern int tfi_netresend(struct mbuf *m, struct inpcb *inp, seq_t seq_no);
extern int tfi_sendack(struct mbuf *m, struct inpcb *inp);
extern int tfi_acksum(seq_t seq_no, unsigned char nonce_sum, struct socket *so);
extern int tfi_acksum_leq(seq_t seq_no, unsigned char nonce_sum, struct socket *so);
extern int tfi_nack(seq_t seq_no, struct socket *so);

extern int tfi_netsend_range(struct mbuf *m, struct inpcb *inp, seq_t bseq,
			     seq_t eseq);
extern int tfi_netresend_range(struct mbuf *m, struct inpcb *inp, seq_t bseq, seq_t eseq);
extern int tfi_acksum_range(seq_t bseq, seq_t eseq, u_char nonce_sum,
			    struct socket *so);

/* cyclone runtime support */
extern u_int32_t tfi_htonl(u_int32_t x);
extern u_int16_t tfi_htons(u_int16_t x);
extern u_int32_t tfi_ntohl(u_int32_t x);
extern u_int16_t tfi_ntohs(u_int16_t x);

/* Only for debugging purposes */
extern int tfi_printf(const char *x);
extern int printf(const char *x, ...);

extern const int hz;
extern const int max_linkhdr;
extern int max_protohdr; // XXX: Modification of this should not be allowed.
extern const int tfi_mhlen;
extern void panic(const char *str);
extern const int ip_defttl;
extern const int ticks;
u_int read_random_unlimited(`a buf, u_int nbytes);

typedef struct MD5Context {
  u_int32_t state[4];	/* state (ABCD) */
  u_int32_t count[2];	/* number of bits, modulo 2^64 (lsb first) */
  unsigned char buffer[64];	/* input buffer */
} MD5_CTX;

void   MD5Init (MD5_CTX *);
void   MD5Update (MD5_CTX *, `a, unsigned int);
void   MD5Pad (MD5_CTX *);
void   MD5Final (`a, MD5_CTX *);
char * MD5End(MD5_CTX *, char *);
char * MD5File(const char *, char *);
char * MD5Data(const unsigned char *, unsigned int, char *);

struct rt_metrics *tfi_getrtmx(struct rtentry *rte);

/* BEGIN parveen */

extern int tfi_soshiwat(struct socket *so);
extern int tfi_sorhiwat(struct socket *so);
struct rtentry *tfi_rtlookup(struct in_conninfo *inc);
int tfi_isrtup(struct rtentry *rt);
int tfi_issolinger(struct socket *so);
void tfi_setsolinger(struct socket *so, int period);
void tfi_sbrflush(struct socket * so);
void tfi_soisdisconnecting(struct socket * so);
void tfi_soisconnecting(struct socket * so);
void tfi_socantsendmore(struct socket * so);
int tfi_sogetscc(struct socket *so);
int tfi_sogetrcc(struct socket *so);
int tfi_msethlen(struct mbuf *m, int plen);
int tfi_in_pcbladdr(struct inpcb *inp, struct sockaddr_in *nam, struct sockaddr_in **ifaddr);

// This is very unsafe, we cannot let a transport see other inpcb. It can use those and
// screw other connections...
struct inpcb *tfi_in_pcblookup_hash(struct inpcb *inp, struct in_addr faddr, u_int fport, struct in_addr laddr, u_int lport, u_int wildcard, struct ifnet *ifp);
void tfi_in_pcbrehash(struct inpcb *inp);
void tfi_soisdisconnected(struct socket * so);
short tfi_getsostate(struct socket * so);
short tfi_mflags(struct mbuf *m);
int tfi_in_broadcast(struct in_addr in, struct ifnet *ifp);
int tfi_in_localaddr(struct in_addr in);
void tfi_sbsdrop(struct socket *so, int len);
void tfi_sohasoutofband(struct socket *so);
void tfi_socantrcvmore(struct socket *so);
struct mbuf *tfi_mgetnext(struct mbuf *);
extern const u_long sb_max;
extern int tfi_rtgetmtu(struct rtentry *);
extern void tfi_sbsreserve(struct socket *so, u_long cc);
extern void tfi_sbrreserve(struct socket *so, u_long cc);
extern char ?tfi_m_copydata(struct socket *so, int off, int len, char ?@nozeroterm data);
extern struct mbuf *tfi_m_copy(struct socket *so, int off, int len);
extern void tfi_m_setnext(struct mbuf *, struct mbuf *);
extern u_short  tfi_in_addword(u_short, u_short);
extern int tfi_netresend(struct mbuf *m, struct inpcb *inp, seq_t seq_no);

/* END parveen */

/* BEGIN tim */

extern int tfi_pru_accept_notsupp(struct socket *so, struct sockaddr **nam);
void	in_losing __P((struct inpcb *));
int	in_pcbinshash __P((struct inpcb *));
struct mbuf *ip_srcroute(void);
int ip_output(struct mbuf *, struct mbuf *, struct conn_route *, int, void *);
short tfi_sogetoptions(struct socket *so);
long tfi_getrt_refcnt(struct rtentry *rte);
u_long tfi_getrt_flags(struct rtentry *rte);
struct sockaddr *tfi_getrt_gateway(struct rtentry *rte);

void tfi_rt_free(struct rtentry *);
struct sockaddr *tfi_rt_key(struct rtentry *);
struct sockaddr *tfi_rt_mask(struct rtentry *);

u_int32_t arc4random __P((void));

int	 rtrequest __P((int, struct sockaddr *,
	    struct sockaddr *, struct sockaddr *, int, struct rtentry **));

extern struct socket *tfi_sonewconn(struct socket *, int);
extern struct socket *tfi_sonewconn1(struct socket *, int, struct tfi_proto *);
extern void tfi_soabort(struct socket *so);
void tfi_minithdr(struct mbuf *m, int offset, int len);

#define	RTF_UP		0x1		/* route usable */
#define	RTF_GATEWAY	0x2		/* destination is a gateway */
#define	RTF_HOST	0x4		/* host entry (net otherwise) */
#define	RTF_REJECT	0x8		/* host or net unreachable */
#define	RTF_DYNAMIC	0x10		/* created dynamically (by redirect) */
#define	RTF_MODIFIED	0x20		/* modified dynamically (by redirect) */
#define RTF_DONE	0x40		/* message confirmed */
/*			0x80		   unused, was RTF_DELCLONE */
#define RTF_CLONING	0x100		/* generate new routes on use */
#define RTF_XRESOLVE	0x200		/* external daemon resolves name */
#define RTF_LLINFO	0x400		/* generated by link layer (e.g. ARP) */
#define RTF_STATIC	0x800		/* manually added */
#define RTF_BLACKHOLE	0x1000		/* just discard pkts (during updates) */
#define RTF_PROTO2	0x4000		/* protocol specific routing flag */
#define RTF_PROTO1	0x8000		/* protocol specific routing flag */

#define RTF_PRCLONING	0x10000		/* protocol requires cloning */
#define RTF_WASCLONED	0x20000		/* route generated through cloning */
#define RTF_PROTO3	0x40000		/* protocol specific routing flag */
/*			0x80000		   unused */
#define RTF_PINNED	0x100000	/* future use */
#define	RTF_LOCAL	0x200000 	/* route represents a local address */
#define	RTF_BROADCAST	0x400000	/* route represents a bcast address */
#define	RTF_MULTICAST	0x800000	/* route represents a mcast address */
	
/*
 * Message types.
 */
#define RTM_ADD		0x1	/* Add Route */
#define RTM_DELETE	0x2	/* Delete Route */
#define RTM_CHANGE	0x3	/* Change Metrics or flags */
#define RTM_GET		0x4	/* Report Metrics */
#define RTM_LOSING	0x5	/* Kernel Suspects Partitioning */
#define RTM_REDIRECT	0x6	/* Told to use different route */
#define RTM_MISS	0x7	/* Lookup failed on this address */
#define RTM_LOCK	0x8	/* fix specified metrics */
#define RTM_OLDADD	0x9	/* caused by SIOCADDRT */
#define RTM_OLDDEL	0xa	/* caused by SIOCDELRT */
#define RTM_RESOLVE	0xb	/* req to resolve dst to LL addr */
#define RTM_NEWADDR	0xc	/* address being added to iface */
#define RTM_DELADDR	0xd	/* address being removed from iface */
#define RTM_IFINFO	0xe	/* iface going up/down etc. */
#define	RTM_NEWMADDR	0xf	/* mcast group membership being added to if */
#define	RTM_DELMADDR	0x10	/* mcast group membership being deleted */
#define	RTM_IFANNOUNCE	0x11	/* iface arrival/departure */
				/* 0x1000000 and up unassigned */
/* END tim */

#endif /* TFIA_PROTO_H */
