/*
 * xldev.h
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * for any purpose with or without fee is hereby granted, provided
 * that the above copyright notice and this permission notice
 * appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE UNIVERSITY OF UTAH
 * DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING
 * ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO
 * EVENT SHALL THE UNIVERSITY OF UTAH BE LIABLE FOR ANY SPECIAL,
 * DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER
 * IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
 * THIS SOFTWARE.
 */

#ifndef __XLDEV_H
#define __XLDEV_H

/*
 * Structure for individual requests.
 *
 * xr_link - List node slot.
 * xr_data - Stores the request data, currently, just an opaque byte array.
 * xr_data.index - The current index into the data buffer.
 * xr_data.length - The actual amount of data in the buffer.
 * xr_data.buffer - The buffer itself.
 */
typedef struct xld_request {
    STAILQ_ENTRY(xld_request) xr_link;
    struct {
	u_long index;
	u_long length;
#define XLD_MAX_REQUEST_LENGTH 128
	char buffer[XLD_MAX_REQUEST_LENGTH];
    } xr_data;
} xld_request_t;

/*
 * Header for lists of xld_request's.
 */
STAILQ_HEAD(xld_request_head, xld_request);

/*
 * Descriptor associated with each open xldev file.
 *
 * xd_requests - The current set of posted requests.
 */
typedef struct xld_d {
    struct xld_request_head xd_requests;
} xld_d_t;

/*
 * The xldev device structure allocated on module load.
 */
extern dev_t xld_sdev;

/* BEGIN Device functions */
d_open_t xldev_open;
d_close_t xldev_close;
d_ioctl_t xldev_ioctl;
d_read_t xldev_read;
d_write_t xldev_write;
/* END Device functions */

/*
 * Append a request to the descriptor associated with the given device.
 *
 * dev - The device to operate on.  (e.g. xld_sdev)
 * ia - The IP address of the requester.
 * hash - The requested protocol hash.
 * returns - Zero on success, otherwise one of the following error codes:
 *           EINVAL: One or more of the arguments is invalid.
 *           ENOTCONN: The device is not opened by a process.
 */
int xldev_req_append_in(dev_t dev, struct in_addr *ia, u_long hash);

#endif
