/*
 * xldevmod.c
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * for any purpose with or without fee is hereby granted, provided
 * that the above copyright notice and this permission notice
 * appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE UNIVERSITY OF UTAH
 * DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING
 * ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO
 * EVENT SHALL THE UNIVERSITY OF UTAH BE LIABLE FOR ANY SPECIAL,
 * DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER
 * IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
 * THIS SOFTWARE.
 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/module.h>
#include <sys/conf.h>

#include <netinet/in.h>

#include <netinet/xtcp/xldev.h>

#define XLDEV_MAJOR 32

static struct cdevsw xld_devsw = {
	/* open */	xldev_open,
	/* close */	xldev_close,
	/* read */	xldev_read,
	/* write */	xldev_write,
	/* ioctl */	xldev_ioctl,
	/* poll */	nopoll,
	/* mmap */	nommap,
	/* strategy */	nostrategy,
	/* name */	"xldev",
	/* maj */	XLDEV_MAJOR,
	/* dump */	nodump,
	/* psize */	nopsize,
	/* flags */	D_TTY,
	/* bmaj */	-1
};

/* 
 * Used as the variable that is the reference to our device
 * in devfs... we must keep this variable sane until we 
 * call kldunload.
 */
dev_t xld_sdev;

/*
 * This function is called each time the module is loaded or unloaded.
 * Since we are a miscellaneous module, we have to provide whatever
 * code is necessary to patch ourselves into the area we are being
 * loaded to change.
 *
 * The stat information is basically common to all modules, so there
 * is no real issue involved with stat; we will leave it lkm_nullcmd(),
 * since we don't have to do anything about it.
 */

static int
xldev_load(module_t mod, int cmd, void *arg)
{
    int  err = 0;

    switch (cmd) {
    case MOD_LOAD:
	
	/* Do any initialization that you should do with the kernel */
	
	/* if we make it to here, print copyright on console*/
	printf("\nXTCP protocol loader device driver\n");
	printf("Copyright (c) 2003\n");
	printf("University of Utah and the Flux Group\n");
	printf("All rights reserved\n");
	xld_sdev = make_dev(&xld_devsw, 0, UID_ROOT, GID_WHEEL, 0600, "xldev");
	break;		/* Success*/

    case MOD_UNLOAD:
	printf("Unloaded XTCP protocol loader device driver\n");
	destroy_dev(xld_sdev);
	break;		/* Success*/

    default:	/* we only understand load/unload*/
	err = EINVAL;
	break;
    }

    return(err);
}

/* Now declare the module to the system */

DEV_MODULE(xldev, xldev_load, NULL);
