/*
 * Copyright (c) 1982, 1986, 1989, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	@(#)udp_var.h	8.1 (Berkeley) 6/10/93
 * $FreeBSD: src/sys/netinet/udp_var.h,v 1.22.2.1 2001/02/18 07:12:25 luigi Exp $
 */

#ifndef _NETINET_XUDP_VAR_H_
#define _NETINET_XUDP_VAR_H_

/* XUDP packet types */
#define XUDP_DATA 2
#define XUDP_ACK  4
#define XUDP_CE   8

/*
 * UDP kernel structures and variables.
 */
struct	xudpiphdr {
	struct 	ipovly ui_i;		/* overlaid ip structure */
	struct	xudphdr ui_u;		/* udp header */
};
#define	ui_x1		ui_i.ih_x1
#define	ui_pr		ui_i.ih_pr
#define	ui_len		ui_i.ih_len
#define	ui_src		ui_i.ih_src
#define	ui_dst		ui_i.ih_dst
#define	ui_sport	ui_u.uh_sport
#define	ui_dport	ui_u.uh_dport
#define	ui_ulen		ui_u.uh_ulen
#define	ui_sum		ui_u.uh_sum
#define	ui_seq		ui_u.uh_seq
#define	ui_noncesum	ui_u.uh_noncesum
#define	ui_type	        ui_u.uh_type

struct	xudpstat_t {
				/* input statistics: */
	unsigned long	udps_ipackets;		/* total input packets */
	unsigned long	udps_hdrops;		/* packet shorter than header */
	unsigned long	udps_badsum;		/* checksum error */
	unsigned long	udps_nosum;		/* no checksum */
	unsigned long	udps_badlen;		/* data length larger than packet */
	unsigned long	udps_noport;		/* no socket on port */
	unsigned long	udps_noportbcast;	/* of above, arrived as broadcast */
	unsigned long	udps_fullsock;		/* not delivered, input socket full */
	unsigned long	udpps_pcbcachemiss;	/* input packets missing pcb cache */
	unsigned long	udpps_pcbhashmiss;	/* input packets not for hashed pcb */
				/* output statistics: */
	unsigned long	udps_opackets;		/* total output packets */
	unsigned long	udps_fastout;		/* output packets on fast path */
	/* of no socket on port, arrived as multicast */
	unsigned long	udps_noportmcast;
};

/*
 * Names for UDP sysctl objects
 */
#define	UDPCTL_CHECKSUM		1	/* checksum UDP packets */
#define UDPCTL_STATS		2	/* statistics (read-only) */
#define	UDPCTL_MAXDGRAM		3	/* max datagram size */
#define	UDPCTL_RECVSPACE	4	/* default receive buffer space */
#define	UDPCTL_PCBLIST		5	/* list of PCBs for UDP sockets */
#define UDPCTL_MAXID		6

#define UDPCTL_NAMES { \
	{ 0, 0 }, \
	{ "checksum", CTLTYPE_INT }, \
	{ "stats", CTLTYPE_STRUCT }, \
	{ "maxdgram", CTLTYPE_INT }, \
	{ "recvspace", CTLTYPE_INT }, \
	{ "pcblist", CTLTYPE_STRUCT }, \
}

#ifdef _KERNEL
/* SYSCTL_DECL(_net_inet_udp); */

/*
extern struct	pr_usrreqs udp_usrreqs;
extern struct	inpcbhead udb;
extern struct	inpcbinfo udbinfo;
*/
extern unsigned long	xudp_sendspace;
extern unsigned long	xudp_recvspace;
extern struct xudpstat_t xudpstat;
extern int	x_log_in_vain;

void	xudp_ctlinput __P((int, sockaddr_t *, void *));
int	xudp_init __P((void));
int     xudp_input __P((mbuf_t *, int, int, socket_t *));
void	xudp_notify __P((socket_t *so, int err));
int	xudp_shutdown __P((socket_t *so));
#endif

#endif
