/*
 * jail.h
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * for any purpose with or without fee is hereby granted, provided
 * that the above copyright notice and this permission notice
 * appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE UNIVERSITY OF UTAH
 * DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING
 * ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO
 * EVENT SHALL THE UNIVERSITY OF UTAH BE LIABLE FOR ANY SPECIAL,
 * DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER
 * IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
 * THIS SOFTWARE.
 */
/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.org> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 *
 * $FreeBSD: src/sys/sys/jail.h,v 1.8.2.2 2000/11/01 17:58:06 rwatson Exp $
 *
 */

#ifndef _SYS_JAIL_H_
#define _SYS_JAIL_H_

/* original (version 0) */
struct jail {
	u_int32_t	version;
	char		*path;
	char		*hostname;
	u_int32_t	ip_number;
};

/* version 1 */
struct jail1 {
	u_int32_t	version;
	char		*path;
	char		*hostname;
	u_int32_t	ip_number;
	u_int16_t	port_low, port_high;
	int		options;
};

/* Per-jail options.  Dependent on global sysctl settings below */
#define JAIL_SYSVIPC	1	/* sys5 IPC allowed */
#define JAIL_INETRAW	2	/* PF_INET/SOCK_RAW sockets allowed */
#define JAIL_BPFRO	4	/* read-only BPF allowed */

#define JAIL_ALLOPTS	7

#ifndef _KERNEL

int jail __P((struct jail *));

#else /* _KERNEL */

#ifdef MALLOC_DECLARE
MALLOC_DECLARE(M_PRISON);
#endif

/*
 * This structure describes a prison.  It is pointed to by all struct
 * proc's of the inmates.  pr_ref keeps track of them and is used to
 * delete the struture when the last inmate is dead.
 */

struct prison {
	int		pr_ref;
	char 		pr_host[MAXHOSTNAMELEN];
	u_int32_t	pr_ip;
	u_int16_t	pr_lport, pr_hport;
	void		*pr_linux;
	int		pr_flags;
};

/*
 * Sysctl-set variables that determine global jail policy
 */
extern int	jail_set_hostname_allowed;
extern int	jail_socket_unixiproute_only;

struct proc;
int	prison_allows __P((struct proc *p, int flag));
void	prison_log __P((struct proc *p, const char *fmt, ...));

#endif /* !_KERNEL */
#endif /* !_SYS_JAIL_H_ */
