/*
 * sink.c
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * for any purpose with or without fee is hereby granted, provided
 * that the above copyright notice and this permission notice
 * appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE UNIVERSITY OF UTAH
 * DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING
 * ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO
 * EVENT SHALL THE UNIVERSITY OF UTAH BE LIABLE FOR ANY SPECIAL,
 * DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER
 * IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
 * THIS SOFTWARE.
 */
/* fpont 1/00 */
/* pont.net    */
/* tcpServer.c */

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <stdio.h>
#include <unistd.h> /* close */

#include <sys/time.h>
#include <unistd.h>

/*#define IPPROTO_XTCP 114*/
#define TFC_ATTACH_PROTOCOL 1
#define XTCP_HASH 170

#define SERVER_PORT 1505
#define MAX_MSG 100

#ifdef XTCP
#define PROTOCOL 114
#else
#define PROTOCOL 0
#endif


#define SUCCESS 0
#define ERROR   1

#define END_LINE 0x0A

/* function readline */
int read_line();

int main (int argc, char *argv[]) {
  
  int sd, newSd, cliLen;

  struct sockaddr_in cliAddr, servAddr;
  char buffer[4096];

  struct timeval periodicstart, now;
  unsigned long long int periodicbytesread = 0ULL;
  int ret;

  /* create socket */
  sd = socket(AF_INET, SOCK_STREAM, PROTOCOL);
   if(sd<0) {
    perror("cannot open socket ");
    return ERROR;
  }
   
#ifdef XTCP
   {
     int pval = XTCP_HASH;
     if (setsockopt(sd, PROTOCOL, TFC_ATTACH_PROTOCOL, &pval, sizeof pval)
	 == -1) {
       perror("setsockopt");
       exit(1);
     }
     printf("sockopt...");
   }
#endif
   
  /* bind server port */
  servAddr.sin_family = AF_INET;
  servAddr.sin_addr.s_addr = htonl(INADDR_ANY);
  servAddr.sin_port = htons(SERVER_PORT);
  
  if(bind(sd, (struct sockaddr *) &servAddr, sizeof(servAddr))<0) {
    perror("cannot bind port ");
    return ERROR;
  }

  listen(sd,5);

  printf("%s: waiting for data on port TCP %u\n",argv[0],SERVER_PORT);
  
  cliLen = sizeof(cliAddr);
  newSd = accept(sd, (struct sockaddr *) &cliAddr, &cliLen);
  if(newSd<0) {
    perror("cannot accept connection ");
      return ERROR;
  }
  
  while(1) {
    ret = recv(newSd, buffer, 4096, 0);
    if (ret <= 0) {
      printf("  ** read failed :(  %d\n", ret);
      break;
    }
    /*printf("read %d bytes\n");*/

    periodicbytesread += ret;

    if (periodicbytesread >= 100000000ULL) {
      unsigned long diffmicros;
      double        throughput;

      gettimeofday(&now, NULL);
      diffmicros = 1000000L*(now.tv_sec - periodicstart.tv_sec);
      diffmicros += (now.tv_usec - periodicstart.tv_usec);

      throughput = ((double) periodicbytesread) / ( (double) diffmicros);
      printf("Thr%d: throughput %f bytes/us  ==  %f bytes/s == %f Mb/s\n",
		 newSd, throughput, 1000000L*throughput, 8*throughput);
      periodicbytesread = 0ULL;
      gettimeofday(&periodicstart, NULL);
    }
  }

  return 0;
}


/* WARNING WARNING WARNING WARNING WARNING WARNING WARNING       */
/* this function is experimental.. I don't know yet if it works  */
/* correctly or not. Use Steven's readline() function to have    */
/* something robust.                                             */
/* WARNING WARNING WARNING WARNING WARNING WARNING WARNING       */

/* rcv_line is my function readline(). Data is read from the socket when */
/* needed, but not byte after bytes. All the received data is read.      */
/* This means only one call to recv(), instead of one call for           */
/* each received byte.                                                   */
/* You can set END_CHAR to whatever means endofline for you. (0x0A is \n)*/
/* read_lin returns the number of bytes returned in line_to_return       */
int read_line(int newSd, char *line_to_return) {
  
  static int rcv_ptr=0;
  static char rcv_msg[MAX_MSG];
  static int n;
  int offset;

  offset=0;

  while(1) {
    if(rcv_ptr==0) {
      /* read data from socket */
      memset(rcv_msg,0x0,MAX_MSG); /* init buffer */
      n = recv(newSd, rcv_msg, MAX_MSG, 0); /* wait for data */
      if (n<0) {
	perror(" cannot receive data ");
	return ERROR;
      } else if (n==0) {
	printf(" connection closed by client\n");
	close(newSd);
	return ERROR;
      }
    }
  
    /* if new data read on socket */
    /* OR */
    /* if another line is still in buffer */

    /* copy line into 'line_to_return' */
    while(*(rcv_msg+rcv_ptr)!=END_LINE && rcv_ptr<n) {
      memcpy(line_to_return+offset,rcv_msg+rcv_ptr,1);
      offset++;
      rcv_ptr++;
    }
    
    /* end of line + end of buffer => return line */
    if(rcv_ptr==n-1) { 
      /* set last byte to END_LINE */
      *(line_to_return+offset)=END_LINE;
      rcv_ptr=0;
      return ++offset;
    } 
    
    /* end of line but still some data in buffer => return line */
    if(rcv_ptr <n-1) {
      /* set last byte to END_LINE */
      *(line_to_return+offset)=END_LINE;
      rcv_ptr++;
      return ++offset;
    }

    /* end of buffer but line is not ended => */
    /*  wait for more data to arrive on socket */
    if(rcv_ptr == n) {
      rcv_ptr = 0;
    } 
    
  } /* while */
}
  
  

