/*
 * source.c
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * for any purpose with or without fee is hereby granted, provided
 * that the above copyright notice and this permission notice
 * appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE UNIVERSITY OF UTAH
 * DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING
 * ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO
 * EVENT SHALL THE UNIVERSITY OF UTAH BE LIABLE FOR ANY SPECIAL,
 * DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER
 * IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
 * THIS SOFTWARE.
 */
/* fpont 12/99 */
/* pont.net    */
/* tcpClient.c */

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <stdio.h>
#include <unistd.h> /* close */

/*#define IPPROTO_XTCP 114*/
#define TFC_ATTACH_PROTOCOL 1
#define XTCP_HASH 170

#define SERVER_PORT 1505
#define SERVER_ADDRESS 0x91db5f80   /* 128.95.219.145 */
#define MAX_MSG 100

#ifdef XTCP
#define PROTOCOL 114
#else
#define PROTOCOL 0
#endif

int main (int argc, char *argv[]) {

  int sd, rc, i;

#ifdef XTCP
  int pval = XTCP_HASH;
#endif

  struct sockaddr_in localAddr, servAddr;
  /*  struct hostent *h;*/

  printf("starting source!\n");

  servAddr.sin_family = AF_INET;
  servAddr.sin_addr.s_addr = SERVER_ADDRESS;
  servAddr.sin_port = htons(SERVER_PORT);

  printf("opening sock...\n");
  
  /* create socket */
  sd = socket(AF_INET, SOCK_STREAM, PROTOCOL);
  if(sd<0) {
    perror("cannot open socket ");
    exit(1);
  }
  printf("socket...\n");
  
#ifdef XTCP
  
  if (setsockopt(sd, PROTOCOL, TFC_ATTACH_PROTOCOL, &pval, sizeof pval)
      == -1) {
    perror("setsockopt");
    exit(1);
  }
  printf("sockopt...\n");
  
#endif
  
  /* bind any port number */
  localAddr.sin_family = AF_INET;
  localAddr.sin_addr.s_addr = htonl(INADDR_ANY);
  localAddr.sin_port = htons(0);
  
  rc = bind(sd, (struct sockaddr *) &localAddr, sizeof(localAddr));
  if(rc<0) {
    printf("%s: cannot bind port TCP %u\n",argv[0],SERVER_PORT);
    perror("error ");
    exit(1);
  }
  printf("bind successful...\n ");
  
  /* connect to server */
  rc = connect(sd, (struct sockaddr *) &servAddr, sizeof(servAddr));
  if(rc<0) {
    perror("cannot connect ");
    exit(1);
  }
  printf ("connect successful\n");

  {
    char data [4096];
    int cycles = 0;
    int bytes = 0;

    while (1) {
      rc = send(sd, data, 4096, 0);
    
      if(rc<0) {
	perror("cannot send data ");
	close(sd);
	exit(1);    
      }     

      bytes+= rc;
      cycles++;
      if ( (cycles % (64*1024)) == 0)
	printf("bytes: %d\n",bytes);
    }
  }

  close (sd);      
  return 0;
}
