/*
 * STP-related changes are Copyright (c) 2005
 * The University of Utah and the Flux Group
 * All rights reserved.
 *
 * Copyright (c) 1982, 1986, 1993, 1994, 1995
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	@(#)tcp_var.h	8.4 (Berkeley) 5/24/95
 * $FreeBSD: src/sys/netinet/tcp_var.h,v 1.56.2.12 2002/08/24 18:40:26 dillon Exp $
 */

/*
 * AJW: We borrow the constants from tcp_var.h, but rename all globals
 * and functions to avoid name clashes.
 *
 * also includes a small number of definitions from other .h files
 */

#ifndef _NETINET_XTCP_VAR_H_
#define _NETINET_XTCP_VAR_H_

#define TFC_NEGOTIATE

/* CYC:
#include <netinet/in_pcb.h>
#include <netinet/tcp_var.h>
#include <netinet/xtcp/tfi.h>
*/

// ip_compat.h
#ifndef	TH_FIN
# define	TH_FIN	0x01
#endif
#ifndef	TH_SYN
# define	TH_SYN	0x02
#endif
#ifndef	TH_RST
# define	TH_RST	0x04
#endif
#ifndef	TH_PUSH
# define	TH_PUSH	0x08
#endif
#ifndef	TH_ACK
# define	TH_ACK	0x10
#endif
#ifndef	TH_URG
# define	TH_URG	0x20
#endif

/* AJW: from tcp.h; added nonce sum field */

/*
 * TCP header.
 * Per RFC 793, September, 1981.
 */
struct xtcphdr {
	u_short	th_sport;		/* source port */
	u_short	th_dport;		/* destination port */
	tcp_seq	th_seq;			/* sequence number */
	tcp_seq	th_ack;			/* acknowledgement number */
#if BYTE_ORDER == LITTLE_ENDIAN
	u_int	th_x2:4,		/* (unused) */
		th_off:4;		/* data offset */
#endif
#if BYTE_ORDER == BIG_ENDIAN
	u_int	th_off:4,		/* data offset */
		th_x2:4;		/* (unused) */
#endif
	u_char	th_flags;

	u_short	th_win;			/* window */
	u_short	th_sum;			/* checksum */
	u_short	th_urp;			/* urgent pointer */

  /*
        seq_t th_data_nonce_seq;         
        seq_t th_ack_nonce_seq;
	*/

        u_char  th_nonce_sum;           /* TFC nonce sum */
};


struct xtcp_syncache {
	inp_gen_t	sc_inp_gencnt;		/* pointer check */
	struct 		xtcpcb *sc_tp;		/* tcb for listening socket */
	mbuf_t		*sc_ipopts;	/* source route */
	struct 		in_conninfo sc_inc;	/* addresses */
#define sc_route	sc_inc.inc_route
#define sc_route6	sc_inc.inc6_route
	u_int32_t	sc_tsrecent;
	tcp_cc		sc_cc_send;		/* holds CC or CCnew */
	tcp_cc		sc_cc_recv;
	tcp_seq 	sc_irs;			/* seq from peer */
	tcp_seq 	sc_iss;			/* our ISS */
	u_long		sc_rxttime;		/* retransmit time */
	u_int16_t	sc_rxtslot; 		/* retransmit counter */
	u_int16_t	sc_peer_mss;		/* peer's MSS */
	u_int16_t	sc_wnd;			/* advertised window */
	u_int8_t 	sc_requested_s_scale:4,
			sc_request_r_scale:4;
	u_int8_t	sc_flags;
#define SCF_NOOPT	0x01			/* no TCP options */
#define SCF_WINSCALE	0x02			/* negotiated window scaling */
#define SCF_TIMESTAMP	0x04			/* negotiated timestamps */
#define SCF_CC		0x08			/* negotiated CC */
#define SCF_SACK_PERMIT	0x40			/* negotiated SACK */
#define SCF_UNREACH	0x10			/* icmp unreachable received */
#define SCF_KEEPROUTE	0x20			/* keep cloned route */
	// Cyclone Queue can be used here but see that later
	TAILQ_ENTRY(xtcp_syncache)	sc_hash;
	TAILQ_ENTRY(xtcp_syncache)	sc_timerq;

        u_char          sc_nonce_sum;
        u_long          sc_resends;             /* How many resends? */
};

struct xtcp_syncache_head {
	TAILQ_HEAD(, xtcp_syncache)	sch_bucket;
	u_int		sch_length;
};

/* AJW: From tcpip.h; reference xtcpcb  */

/*
 * Tcp+ip header, after ip options removed.
 */
struct xtcpiphdr {
	struct 	ipovly ti_i;		/* overlaid ip structure */
	struct	xtcphdr ti_t;		/* tcp header */
};
#ifdef notyet
/*
 * Tcp+ip header, after ip options removed but including TCP options.
 */
struct full_xtcpiphdr {
	struct 	ipovly ti_i;		/* overlaid ip structure */
	struct	xtcphdr ti_t;		/* tcp header */
	char	ti_o[TCP_MAXOLEN];	/* space for tcp options */
};
#endif /* notyet */


/*
 * Kernel variables for tcp.
 */
extern int	xtcp_do_rfc1323;
extern int	xtcp_do_rfc1644;


/* TCP segment queue entry */
struct xtcp_tseg_qent {
	LIST_ENTRY(xtcp_tseg_qent) tqe_q;
	int	tqe_len;		/* TCP segment data length */
	struct	xtcphdr *tqe_th;	/* a pointer to tcp header */
	mbuf_t	*tqe_m;		/* mbuf contains packet */
        u_char  tqe_tos;                /* IP TOS (for nonce sum) */
};
// PP: Use Cyclone queue 
LIST_HEAD(xtcp_tsegqe_head, xtcp_tseg_qent);

/*
#ifdef MALLOC_DECLARE
MALLOC_DECLARE(M_TSEGQ);
#endif
*/

struct xtcptemp {
	struct ip tt_ipgen;
	// u_char	tt_ipgen[40]; /* the size must be of max ip header, now IPv6 */
	u_char tt_iprest[20];
	struct	xtcphdr tt_t;
};

#define xtcp6cb		xtcpcb  /* for KAME src sync over BSD*'s */

/*
 * Tcp control block, one per tcp; fields:
 * Organized for 16 byte cacheline efficiency.
 */
struct xtcpcb {
	struct	xtcp_tsegqe_head t_segq;
	int	t_dupacks;		/* consecutive dup acks recd */
	struct	xtcptemp	*unused;	/* unused */

	callout_t* tt_rexmt;	/* retransmit timer */
	callout_t* tt_persist;	/* retransmit persistence */
	callout_t* tt_keep;	/* keepalive */
	callout_t* tt_2msl;	/* 2*msl TIME_WAIT timer */
	callout_t* tt_delack;	/* delayed ACK timer */

	inpcb_t *t_inpcb;		/* back pointer to internet pcb */
	int	t_state;		/* state of this connection */
	u_int	t_flags;
	
	int	t_force;		/* 1 if forcing out a byte */

	tcp_seq	snd_una;		/* send unacknowledged */
	tcp_seq	snd_max;		/* highest sequence number sent;
					 * used to recognize retransmits
					 */
	tcp_seq	snd_nxt;		/* send next */
	tcp_seq	snd_up;			/* send urgent pointer */

	tcp_seq	snd_wl1;		/* window update seg seq number */
	tcp_seq	snd_wl2;		/* window update seg ack number */
	tcp_seq	iss;			/* initial send sequence number */
	tcp_seq	irs;			/* initial receive sequence number */

	tcp_seq	rcv_nxt;		/* receive next */
	tcp_seq	rcv_adv;		/* advertised window */
	u_long	rcv_wnd;		/* receive window */
	tcp_seq	rcv_up;			/* receive urgent pointer */

	u_long	snd_wnd;		/* send window */
	u_long	snd_cwnd;		/* congestion-controlled window */
	u_long	snd_bwnd;		/* bandwidth-controlled window */
	u_long	snd_ssthresh;		/* snd_cwnd size threshold for
					 * for slow start exponential to
					 * linear switch
					 */
	u_long	snd_bandwidth;		/* calculated bandwidth or 0 */
	tcp_seq	snd_recover;		/* for use in fast recovery */

	u_int	t_maxopd;		/* mss plus options */

	u_long	t_rcvtime;		/* inactivity time */
	u_long	t_starttime;		/* time connection was established */
	int	t_rtttime;		/* round trip time */
	tcp_seq	t_rtseq;		/* sequence number being timed */

	int	t_bw_rtttime;		/* used for bandwidth calculation */
	tcp_seq	t_bw_rtseq;		/* used for bandwidth calculation */

	int	t_rxtcur;		/* current retransmit value (ticks) */
	u_int	t_maxseg;		/* maximum segment size */
	int	t_srtt;			/* smoothed round-trip time */
	int	t_rttvar;		/* variance in round-trip time */

	int	t_rxtshift;		/* log(2) of rexmt exp. backoff */
	u_int	t_rttmin;		/* minimum rtt allowed */
	u_int	t_rttbest;		/* best rtt we've seen */
	u_long	t_rttupdated;		/* number of times rtt sampled */
	u_long	max_sndwnd;		/* largest window peer has offered */

	int	t_softerror;		/* possible error not yet reported */
#define	TCPOOB_HAVEDATA	0x01
#define	TCPOOB_HADDATA	0x02
/* out-of-band data */
	char	t_oobflags;		/* have some */
	char	t_iobc;			/* input character */

/* RFC 1323 variables */
	u_char	snd_scale;		/* window scaling for send window */
	u_char	rcv_scale;		/* window scaling for recv window */
	u_char	request_r_scale;	/* pending window scaling */
	u_char	requested_s_scale;
	u_long	ts_recent;		/* timestamp echo data */

	u_long	ts_recent_age;		/* when last updated */
	tcp_seq	last_ack_sent;
/* RFC 1644 variables */
	tcp_cc	cc_send;		/* send connection count */
	tcp_cc	cc_recv;		/* receive connection count */
/* experimental */
	u_long	snd_cwnd_prev;		/* cwnd prior to retransmit */
	u_long	snd_ssthresh_prev;	/* ssthresh prior to retransmit */
	u_long	t_badrxtwin;		/* window for retransmit recovery */

	/* AJW: Added to avoid sharing with the inp_tp structure */
	callout_t inp_tp_persist, inp_tp_keep, inp_tp_2msl;
	callout_t inp_tp_delack, inp_tp_rexmt;
	
	callout_t *tt_bogus;     
	callout_t inp_tp_bogus;
	
	/* Per-socket nonce sum */
	u_char nonce_sum;
	
	/* Next TFC sequence number to allocate */
	/*  seq_t next_tfc_seq; */

	/* First TFC sequence number to ACK */
	/*  seq_t first_tfc_seq;*/
	
	/* 
	 * Largest TFI nonce sequence number; needed to differentiate
	 * between sends and re-sends
	 */
	
	tcp_seq largest_nonce_seq;

#ifdef TFC_NEGOTIATE
	/* Negotiate: Hash to return to the listener side */
	u_long t_tfchash;
#endif	
	struct  xtcp_tsegqe_head segq_freelist;
};

/*
 * Structure to hold TCP options that are only used during segment
 * processing (in tcp_input), but not held in the tcpcb.
 * It's basically used to reduce the number of parameters
 * to tcp_dooptions.
 */
struct tcpopt {
	u_long		to_flags;	/* which options are present */
#define TOF_TS			0x0001	/* timestamp */
#define TOF_CC			0x0002	/* CC and CCnew are exclusive */
#define TOF_CCNEW		0x0004
#define	TOF_CCECHO		0x0008
#define	TOF_MSS			0x0010
#define	TOF_SCALE		0x0020
#define	TOF_SACK_PERMIT		0x0040
#define	TOF_MIGRATE_PERMIT	0x0080
#define	TOF_MIGRATE_PERMIT_SHORT	0x0100
#define	TOF_MIGRATE		0x0200
#define	TOF_MIGRATE_PRELOAD	0x0400
#ifdef TFC_NEGOTIATE
#define TOF_TFCNEGOTIATE_USE	0x0800
#define TOF_TFCNEGOTIATE_TRY	0x1000
#define TOF_TFCNEGOTIATE_OK	0x2000
	u_int32_t       to_tfchash;
#endif
	u_int32_t	to_tsval;
	u_int32_t	to_tsecr;
	tcp_cc		to_cc;		/* holds CC or CCnew */
	tcp_cc		to_ccecho;
	u_int16_t	to_mss;
	u_int8_t 	to_requested_s_scale;
	u_int8_t 	to_pad;
};

#define	TF_ACKNOW	0x000001	/* ack peer immediately */
#define	TF_DELACK	0x000002	/* ack, but try to delay it */
#define	TF_NODELAY	0x000004	/* don't delay packets to coalesce */
#define	TF_NOOPT	0x000008	/* don't use tcp options */
#define	TF_SENTFIN	0x000010	/* have sent FIN */
#define	TF_REQ_SCALE	0x000020	/* have/will request window scaling */
#define	TF_RCVD_SCALE	0x000040	/* other side has requested scaling */
#define	TF_REQ_TSTMP	0x000080	/* have/will request timestamps */
#define	TF_RCVD_TSTMP	0x000100	/* a timestamp was received in SYN */
#define	TF_SACK_PERMIT	0x000200	/* other side said I could SACK */
#define	TF_NEEDSYN	0x000400	/* send SYN (implicit state) */
#define	TF_NEEDFIN	0x000800	/* send FIN (implicit state) */
#define	TF_NOPUSH	0x001000	/* don't push */
#define	TF_REQ_CC	0x002000	/* have/will request CC */
#define	TF_RCVD_CC	0x004000	/* a CC was received in SYN */
#define	TF_SENDCCNEW	0x008000	/* send CCnew instead of CC in SYN */
#define	TF_MORETOCOME	0x010000	/* More data to be appended to sock */
#define	TF_LQ_OVERFLOW	0x020000	/* listen queue overflow */
#define	TF_LASTIDLE	0x040000	/* connection was previously idle */
#define TF_RXWIN0SENT	0x080000	/* sent a receiver win 0 in response */
#define TF_MIGRATE	0x100000	/* negotiated migrate */
#define TF_NOMIGRATE	0x200000	/* don't offer/accept migrate */
#ifdef TFC_NEGOTIATE
#define TF_TFC_OK       0x400000        /* Need to send an TFC_OK back to server */
#endif /* TFC_NEGOTIATE */

#define	TCP_RTT_SCALE		32	/* multiplier for srtt; 3 bits frac. */
#define	TCP_RTT_SHIFT		5	/* shift for srtt; 3 bits frac. */
#define	TCP_RTTVAR_SCALE	16	/* multiplier for rttvar; 2 bits */
#define	TCP_RTTVAR_SHIFT	4	/* shift for rttvar; 2 bits */
#define	TCP_DELTA_SHIFT		2	/* see tcp_input.c */

/* 
 * From tcp_seq.h.  Initialze the largest TFI sequence number
 */
#define	xtcp_sendseqinit(tp) \
	(tp)->snd_una = (tp)->snd_nxt = (tp)->snd_max = (tp)->snd_up = \
	    (tp)->snd_recover = (tp)->iss


#define	intoxtcpcb(ip)	((struct xtcpcb *) tfi_intouserpcb(ip))
#define	sotoxtcpcb(so)	(intoxtcpcb(tfi_sotoinpcb(so)))

/*
 * The initial retransmission should happen at rtt + 4 * rttvar.
 * Because of the way we do the smoothing, srtt and rttvar
 * will each average +1/2 tick of bias.  When we compute
 * the retransmit timer, we want 1/2 tick of rounding and
 * 1 extra tick because of +-1/2 tick uncertainty in the
 * firing of the timer.  The bias will give us exactly the
 * 1.5 tick we need.  But, because the bias is
 * statistical, we have to test that we don't drop below
 * the minimum feasible timer (which is 2 ticks).
 * This version of the macro adapted from a paper by Lawrence
 * Brakmo and Larry Peterson which outlines a problem caused
 * by insufficient precision in the original implementation,
 * which results in inappropriately large RTO values for very
 * fast networks.
 */
#define	TCP_REXMTVAL(tp) \
	max((tp)->t_rttmin, (((tp)->t_srtt >> (TCP_RTT_SHIFT - TCP_DELTA_SHIFT))  \
	  + (tp)->t_rttvar) >> TCP_DELTA_SHIFT)

// copied from tcp_var.h
/*
 * TCP statistics.
 * Many of these should be kept per connection,
 * but that's inconvenient at the moment.
 */
struct	tcpstat {
	u_long	tcps_connattempt;	/* connections initiated */
	u_long	tcps_accepts;		/* connections accepted */
	u_long	tcps_connects;		/* connections established */
	u_long	tcps_drops;		/* connections dropped */
	u_long	tcps_conndrops;		/* embryonic connections dropped */
	u_long	tcps_closed;		/* conn. closed (includes drops) */
	u_long	tcps_segstimed;		/* segs where we tried to get rtt */
	u_long	tcps_rttupdated;	/* times we succeeded */
	u_long	tcps_delack;		/* delayed acks sent */
	u_long	tcps_timeoutdrop;	/* conn. dropped in rxmt timeout */
	u_long	tcps_rexmttimeo;	/* retransmit timeouts */
	u_long	tcps_persisttimeo;	/* persist timeouts */
	u_long	tcps_keeptimeo;		/* keepalive timeouts */
	u_long	tcps_keepprobe;		/* keepalive probes sent */
	u_long	tcps_keepdrops;		/* connections dropped in keepalive */

	u_long	tcps_sndtotal;		/* total packets sent */
	u_long	tcps_sndpack;		/* data packets sent */
	u_long	tcps_sndbyte;		/* data bytes sent */
	u_long	tcps_sndrexmitpack;	/* data packets retransmitted */
	u_long	tcps_sndrexmitbyte;	/* data bytes retransmitted */
	u_long	tcps_sndacks;		/* ack-only packets sent */
	u_long	tcps_sndprobe;		/* window probes sent */
	u_long	tcps_sndurg;		/* packets sent with URG only */
	u_long	tcps_sndwinup;		/* window update-only packets sent */
	u_long	tcps_sndctrl;		/* control (SYN|FIN|RST) packets sent */

	u_long	tcps_rcvtotal;		/* total packets received */
	u_long	tcps_rcvpack;		/* packets received in sequence */
	u_long	tcps_rcvbyte;		/* bytes received in sequence */
	u_long	tcps_rcvbadsum;		/* packets received with ccksum errs */
	u_long	tcps_rcvbadoff;		/* packets received with bad offset */
	u_long	tcps_rcvmemdrop;	/* packets dropped for lack of memory */
	u_long	tcps_rcvshort;		/* packets received too short */
	u_long	tcps_rcvduppack;	/* duplicate-only packets received */
	u_long	tcps_rcvdupbyte;	/* duplicate-only bytes received */
	u_long	tcps_rcvpartduppack;	/* packets with some duplicate data */
	u_long	tcps_rcvpartdupbyte;	/* dup. bytes in part-dup. packets */
	u_long	tcps_rcvoopack;		/* out-of-order packets received */
	u_long	tcps_rcvoobyte;		/* out-of-order bytes received */
	u_long	tcps_rcvpackafterwin;	/* packets with data after window */
	u_long	tcps_rcvbyteafterwin;	/* bytes rcvd after window */
	u_long	tcps_rcvafterclose;	/* packets rcvd after "close" */
	u_long	tcps_rcvwinprobe;	/* rcvd window probe packets */
	u_long	tcps_rcvdupack;		/* rcvd duplicate acks */
	u_long	tcps_rcvacktoomuch;	/* rcvd acks for unsent data */
	u_long	tcps_rcvackpack;	/* rcvd ack packets */
	u_long	tcps_rcvackbyte;	/* bytes acked by rcvd acks */
	u_long	tcps_rcvwinupd;		/* rcvd window update packets */
	u_long	tcps_pawsdrop;		/* segments dropped due to PAWS */
	u_long	tcps_predack;		/* times hdr predict ok for acks */
	u_long	tcps_preddat;		/* times hdr predict ok for data pkts */
	u_long	tcps_pcbcachemiss;
	u_long	tcps_cachedrtt;		/* times cached RTT in route updated */
	u_long	tcps_cachedrttvar;	/* times cached rttvar updated */
	u_long	tcps_cachedssthresh;	/* times cached ssthresh updated */
	u_long	tcps_usedrtt;		/* times RTT initialized from route */
	u_long	tcps_usedrttvar;	/* times RTTVAR initialized from rt */
	u_long	tcps_usedssthresh;	/* times ssthresh initialized from rt*/
	u_long	tcps_persistdrop;	/* timeout in persist state */
	u_long	tcps_badsyn;		/* bogus SYN, e.g. premature ACK */
	u_long	tcps_mturesent;		/* resends due to MTU discovery */
	u_long	tcps_listendrop;	/* listen queue overflows */

	u_long	tcps_sc_added;		/* entry added to syncache */
	u_long	tcps_sc_retransmitted;	/* syncache entry was retransmitted */
	u_long	tcps_sc_dupsyn;		/* duplicate SYN packet */
	u_long	tcps_sc_dropped;	/* could not reply to packet */
	u_long	tcps_sc_completed;	/* successful extraction of entry */
	u_long	tcps_sc_bucketoverflow;	/* syncache per-bucket limit hit */
	u_long	tcps_sc_cacheoverflow;	/* syncache cache limit hit */
	u_long	tcps_sc_reset;		/* RST removed entry from syncache */
	u_long	tcps_sc_stale;		/* timed out or listen socket gone */
	u_long	tcps_sc_aborted;	/* syncache entry aborted */
	u_long	tcps_sc_badack;		/* removed due to bad ACK */
	u_long	tcps_sc_unreach;	/* ICMP unreachable received */
	u_long	tcps_sc_zonefail;	/* zalloc() failed */
	u_long	tcps_sc_sendcookie;	/* SYN cookie sent */
	u_long	tcps_sc_recvcookie;	/* SYN cookie received */
};

extern	struct tcpstat xtcpstat;	/* tcp statistics */
extern	int xtcp_mssdflt;	/* XXX */
extern	int xtcp_delack_enabled;
extern	int xtcp_do_newreno;
extern	int xtcp_path_mtu_discovery;
extern	int xtcp_ss_fltsz;
extern	int xtcp_ss_fltsz_local;


void	 xtcp_canceltimers __P((struct xtcpcb *));
struct xtcpcb *
	 xtcp_close __P((struct xtcpcb *));
/*void	 xtcp_ctlinput __P((int, struct sockaddr *, void *));*/

void	xtcp_ctlinput __P((socket_t *, int));

int	 xtcp_ctloutput __P((socket_t *, datatype SockOpt *));
struct xtcpcb *
	 xtcp_drop __P((struct xtcpcb *, int));
void	 xtcp_drain __P((void));
void	 xtcp_fasttimo __P((void));
struct rmxp_tao *
	 xtcp_gettaocache __P((struct in_conninfo *`H ));
int	 xtcp_init __P((void));

/* AJW: Prototype modified for TFC */
int	 xtcp_input __P((mbuf_t *`H, int, int,socket_t*));

void	 xtcp_mss __P((struct xtcpcb *, int));
int	 xtcp_mssopt __P((struct xtcpcb *));
void	 xtcp_drop_syn_sent __P((inpcb_t *, int));
void	 xtcp_mtudisc __P((inpcb_t *, int));
struct xtcpcb *
	 xtcp_newtcpcb __P((inpcb_t *));

/* AJW addition for TFC */
void init_xtcpcb (socket_t*);

int	 xtcp_output __P((struct xtcpcb *));
void	 xtcp_quench __P((inpcb_t *, int));
void	 xtcp_respond __P((struct xtcpcb *, struct ip *,
	    struct xtcphdr *, mbuf_t *, tcp_seq, tcp_seq, int));
struct rtentry *
	 xtcp_rtlookup __P((struct in_conninfo *`H ));
void	 xtcp_setpersist __P((struct xtcpcb *));
void	 xtcp_slowtimo __P((void));
struct xtcptemp *
	 xtcp_maketemplate __P((struct xtcpcb *));
void	 xtcp_fillheaders __P((struct xtcpcb *, struct ip *, struct xtcphdr *));
struct xtcpcb *
	 xtcp_timers __P((struct xtcpcb *, int));
void	 xtcp_trace __P((int, int, struct xtcpcb *, void *, struct xtcphdr *,
			int));
void	 xtcp_xmit_bandwidth_limit(struct xtcpcb *tp, tcp_seq ack_seq);
void	 xtcp_syncache_init(void);
void	 xtcp_syncache_unreach(struct in_conninfo *, struct xtcphdr *);
int	 xtcp_syncache_expand(struct in_conninfo *, struct xtcphdr *,
	     socket_t **, mbuf_t *);
int	 xtcp_syncache_add(struct in_conninfo *, struct tcpopt *,
	     struct xtcphdr *, socket_t **, mbuf_t *,u_char);
void	 xtcp_syncache_chkrst(struct in_conninfo *, struct xtcphdr *);
void	 xtcp_syncache_badack(struct in_conninfo *);

#ifdef TFC_NEGOTIATE

int xtcp_syncache_socket(tfi_syncache_t *sc, socket_t *lso,
			 socket_t **nso);
extern tfi_proto_t xtcp_pr;
#define TCPOLEN_TFCNEGOTIATE 6
#define TCPOPT_TFCNEGOTIATE_USE 16
#define TCPOPT_TFCNEGOTIATE_TRY 17
#define TCPOPT_TFCNEGOTIATE_OK 18
#define	   TCPOPT_TFCNEGOTIATE_HDR(tfcopt)		\
    (TCPOPT_NOP<<24|TCPOPT_NOP<<16|(tfcopt)<<8|TCPOLEN_TFCNEGOTIATE)

#endif /* TFC_NEGOTIATE */

/* Changed by AJW to work with TFC */
/*extern	struct pr_usrreqs tcp_usrreqs;*/
extern	tfi_usrreqs_t xtcp_usrreqs;

extern	u_long xtcp_sendspace;
extern	u_long xtcp_recvspace;
tcp_seq xtcp_new_isn __P((struct xtcpcb *));

extern int xtcp_rfc1323;
extern int xtcp_rfc1644;
extern int xtcp_v6mssdflt;

/* from tcp_seq.h */
extern tcp_cc	xtcp_ccgen;		/* global connection count */

/* not used:
	{ "pcblist", CTLTYPE_STRUCT }, \
	{ "stats", CTLTYPE_STRUCT }, \
	*/

#define BANDLIM_UNLIMITED -1
#define BANDLIM_ICMP_UNREACH 0
#define BANDLIM_ICMP_ECHO 1
#define BANDLIM_ICMP_TSTAMP 2
#define BANDLIM_RST_CLOSEDPORT 3 /* No connection, and no listeners */
#define BANDLIM_RST_OPENPORT 4   /* No connection, listener */
#define BANDLIM_MAX 4
#endif /* _NETINET_TCP_VAR_H_ */
