/*
 * STP-related changes are Copyright (c) 2005
 * The University of Utah and the Flux Group
 * All rights reserved.
 *
 * Copyright (c) 1982, 1986, 1993, 1994, 1995
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	@(#)tcp_var.h	8.4 (Berkeley) 5/24/95
 * $FreeBSD: src/sys/netinet/tcp_var.h,v 1.56.2.12 2002/08/24 18:40:26 dillon Exp $
 */

/*
 * AJW: We borrow the constants from tcp_var.h, but rename all globals
 * and functions to avoid name clashes.
 *
 * also includes a small number of definitions from other .h files
 */

#ifndef _NETINET_XTCP_VAR_H_
#define _NETINET_XTCP_VAR_H_

#define TFC_NEGOTIATE

#include <netinet/in_pcb.h>
#include <netinet/tcp_var.h>
#include <netinet/xtcp/tfi.h>


/* AJW: from tcp.h; added nonce sum field */

/*
 * TCP header.
 * Per RFC 793, September, 1981.
 */
struct xtcphdr {
	u_short	th_sport;		/* source port */
	u_short	th_dport;		/* destination port */
	tcp_seq	th_seq;			/* sequence number */
	tcp_seq	th_ack;			/* acknowledgement number */
#if BYTE_ORDER == LITTLE_ENDIAN
	u_int	th_x2:4,		/* (unused) */
		th_off:4;		/* data offset */
#endif
#if BYTE_ORDER == BIG_ENDIAN
	u_int	th_off:4,		/* data offset */
		th_x2:4;		/* (unused) */
#endif
	u_char	th_flags;

	u_short	th_win;			/* window */
	u_short	th_sum;			/* checksum */
	u_short	th_urp;			/* urgent pointer */

  /*
        seq_t th_data_nonce_seq;         
        seq_t th_ack_nonce_seq;
	*/

        u_char  th_nonce_sum;           /* TFC nonce sum */
};


struct xtcp_syncache {
	inp_gen_t	sc_inp_gencnt;		/* pointer check */
	struct 		xtcpcb *sc_tp;		/* tcb for listening socket */
	struct		mbuf *sc_ipopts;	/* source route */
	struct 		in_conninfo sc_inc;	/* addresses */
#define sc_route	sc_inc.inc_route
#define sc_route6	sc_inc.inc6_route
	u_int32_t	sc_tsrecent;
	tcp_cc		sc_cc_send;		/* holds CC or CCnew */
	tcp_cc		sc_cc_recv;
	tcp_seq 	sc_irs;			/* seq from peer */
	tcp_seq 	sc_iss;			/* our ISS */
	u_long		sc_rxttime;		/* retransmit time */
	u_int16_t	sc_rxtslot; 		/* retransmit counter */
	u_int16_t	sc_peer_mss;		/* peer's MSS */
	u_int16_t	sc_wnd;			/* advertised window */
	u_int8_t 	sc_requested_s_scale:4,
			sc_request_r_scale:4;
	u_int8_t	sc_flags;
#define SCF_NOOPT	0x01			/* no TCP options */
#define SCF_WINSCALE	0x02			/* negotiated window scaling */
#define SCF_TIMESTAMP	0x04			/* negotiated timestamps */
#define SCF_CC		0x08			/* negotiated CC */
#define SCF_SACK_PERMIT	0x40			/* negotiated SACK */
#define SCF_UNREACH	0x10			/* icmp unreachable received */
#define SCF_KEEPROUTE	0x20			/* keep cloned route */
	TAILQ_ENTRY(xtcp_syncache)	sc_hash;
	TAILQ_ENTRY(xtcp_syncache)	sc_timerq;

        u_char          sc_nonce_sum;
        u_long          sc_resends;             /* How many resends? */
};

struct xtcp_syncache_head {
	TAILQ_HEAD(, xtcp_syncache)	sch_bucket;
	u_int		sch_length;
};


/* AJW: From tcpip.h; reference xtcpcb  */

/*
 * Tcp+ip header, after ip options removed.
 */
struct xtcpiphdr {
	struct 	ipovly ti_i;		/* overlaid ip structure */
	struct	xtcphdr ti_t;		/* tcp header */
};
#ifdef notyet
/*
 * Tcp+ip header, after ip options removed but including TCP options.
 */
struct full_xtcpiphdr {
	struct 	ipovly ti_i;		/* overlaid ip structure */
	struct	xtcphdr ti_t;		/* tcp header */
	char	ti_o[TCP_MAXOLEN];	/* space for tcp options */
};
#endif /* notyet */


/*
 * Kernel variables for tcp.
 */
extern int	xtcps_do_rfc1323;
extern int	xtcps_do_rfc1644;


/* TCP segment queue entry */
struct xtcp_tseg_qent {
	LIST_ENTRY(xtcp_tseg_qent) tqe_q;
	int	tqe_len;		/* TCP segment data length */
	struct	xtcphdr *tqe_th;	/* a pointer to tcp header */
	struct	mbuf	*tqe_m;		/* mbuf contains packet */
        u_char  tqe_tos;                /* IP TOS (for nonce sum) */
};
LIST_HEAD(xtcp_tsegqe_head, xtcp_tseg_qent);

/*
#ifdef MALLOC_DECLARE
MALLOC_DECLARE(M_TSEGQ);
#endif
*/

struct xtcptemp {
	u_char	tt_ipgen[40]; /* the size must be of max ip header, now IPv6 */
	struct	xtcphdr tt_t;
};

#define xtcp6cb		xtcpcb  /* for KAME src sync over BSD*'s */

/*
 * Tcp control block, one per tcp; fields:
 * Organized for 16 byte cacheline efficiency.
 */
struct xtcpcb {
	struct	xtcp_tsegqe_head t_segq;
	int	t_dupacks;		/* consecutive dup acks recd */
	struct	xtcptemp	*unused;	/* unused */

	struct	callout* tt_rexmt;	/* retransmit timer */
	struct	callout* tt_persist;	/* retransmit persistence */
	struct	callout* tt_keep;	/* keepalive */
	struct	callout* tt_2msl;	/* 2*msl TIME_WAIT timer */
	struct	callout* tt_delack;	/* delayed ACK timer */

	struct	inpcb *t_inpcb;		/* back pointer to internet pcb */
	int	t_state;		/* state of this connection */
	u_int	t_flags;

	int	t_force;		/* 1 if forcing out a byte */

	tcp_seq	snd_una;		/* send unacknowledged */
	tcp_seq	snd_max;		/* highest sequence number sent;
					 * used to recognize retransmits
					 */
	tcp_seq	snd_nxt;		/* send next */
	tcp_seq	snd_up;			/* send urgent pointer */

	tcp_seq	snd_wl1;		/* window update seg seq number */
	tcp_seq	snd_wl2;		/* window update seg ack number */
	tcp_seq	iss;			/* initial send sequence number */
	tcp_seq	irs;			/* initial receive sequence number */

	tcp_seq	rcv_nxt;		/* receive next */
	tcp_seq	rcv_adv;		/* advertised window */
	u_long	rcv_wnd;		/* receive window */
	tcp_seq	rcv_up;			/* receive urgent pointer */

	u_long	snd_wnd;		/* send window */
	u_long	snd_cwnd;		/* congestion-controlled window */
	u_long	snd_bwnd;		/* bandwidth-controlled window */
	u_long	snd_ssthresh;		/* snd_cwnd size threshold for
					 * for slow start exponential to
					 * linear switch
					 */
	u_long	snd_bandwidth;		/* calculated bandwidth or 0 */
	tcp_seq	snd_recover;		/* for use in fast recovery */

	u_int	t_maxopd;		/* mss plus options */

	u_long	t_rcvtime;		/* inactivity time */
	u_long	t_starttime;		/* time connection was established */
	int	t_rtttime;		/* round trip time */
	tcp_seq	t_rtseq;		/* sequence number being timed */

	int	t_bw_rtttime;		/* used for bandwidth calculation */
	tcp_seq	t_bw_rtseq;		/* used for bandwidth calculation */

	int	t_rxtcur;		/* current retransmit value (ticks) */
	u_int	t_maxseg;		/* maximum segment size */
	int	t_srtt;			/* smoothed round-trip time */
	int	t_rttvar;		/* variance in round-trip time */

	int	t_rxtshift;		/* log(2) of rexmt exp. backoff */
	u_int	t_rttmin;		/* minimum rtt allowed */
	u_int	t_rttbest;		/* best rtt we've seen */
	u_long	t_rttupdated;		/* number of times rtt sampled */
	u_long	max_sndwnd;		/* largest window peer has offered */

	int	t_softerror;		/* possible error not yet reported */
/* out-of-band data */
	char	t_oobflags;		/* have some */
	char	t_iobc;			/* input character */

/* RFC 1323 variables */
	u_char	snd_scale;		/* window scaling for send window */
	u_char	rcv_scale;		/* window scaling for recv window */
	u_char	request_r_scale;	/* pending window scaling */
	u_char	requested_s_scale;
	u_long	ts_recent;		/* timestamp echo data */

	u_long	ts_recent_age;		/* when last updated */
	tcp_seq	last_ack_sent;
/* RFC 1644 variables */
	tcp_cc	cc_send;		/* send connection count */
	tcp_cc	cc_recv;		/* receive connection count */
/* experimental */
	u_long	snd_cwnd_prev;		/* cwnd prior to retransmit */
	u_long	snd_ssthresh_prev;	/* ssthresh prior to retransmit */
	u_long	t_badrxtwin;		/* window for retransmit recovery */

	/* AJW: Added to avoid sharing with the inp_tp structure */
	struct	callout  inp_tp_persist, inp_tp_keep, inp_tp_2msl;
	struct	callout inp_tp_delack, inp_tp_rexmt;
	
	
	struct	callout* tt_bogus;     
	struct callout inp_tp_bogus;
	
	/* Per-socket nonce sum */
	u_char nonce_sum;
	
	/* Next TFC sequence number to allocate */
	/*  seq_t next_tfc_seq; */
	
	/* First TFC sequence number to ACK */
	/*  seq_t first_tfc_seq;*/
	
	/* 
	 * Largest TFI nonce sequence number; needed to differentiate
	 * between sends and re-sends
	 */
	
	tcp_seq largest_nonce_seq;

#ifdef TFC_NEGOTIATE
	/* Negotiate: Hash to return to the listener side */
	u_long t_tfchash;
#endif	
	
	/* BEGIN TCP_SACK */
	int	sack_disable;		/* disable SACK for this connection */
	int	snd_numholes;		/* number of holes seen by sender */
	struct sackhole *snd_holes;	/* linked list of holes (sorted) */
	tcp_seq snd_last;		/* for use in fast recovery */

	tcp_seq rcv_laststart;		/* start of last segment recd. */
	tcp_seq rcv_lastend;		/* end of ... */
	tcp_seq rcv_lastsack;		/* last seq number(+1) sack'd by rcv'r*/
	int	rcv_numsacks;		/* # distinct sack blks present */
	
	struct sackblk sackblks[MAX_SACK_BLKS]; /* seq nos. of sack blocks */
	tcp_seq snd_fack;		/* for FACK congestion control */
	u_long	snd_awnd;		/* snd_nxt - snd_fack + */
					/* retransmitted data */
	int retran_data;		/* amount of outstanding retx. data  */
	/* END TCP_SACK */
};

#ifdef TFC_NEGOTIATE
/* Copy other flags from <netinet/tcp_var.h> and remove the dependency on it */ 
#define TF_TFC_OK       0x400000        /* Need to send an TFC_OK back to server */
#endif


/* 
 * From tcp_seq.h.  Initialze the largest TFI sequence number
 */
#define	xtcp_sendseqinit(tp) \
	(tp)->snd_una = (tp)->snd_nxt = (tp)->snd_max = (tp)->snd_up = \
	    (tp)->snd_recover = (tp)->iss


#define	intoxtcpcb(ip)	((struct xtcpcb *) tfi_intouserpcb(ip))
#define	sotoxtcpcb(so)	(intoxtcpcb(sotoinpcb(so)))


extern	struct tcpstat xtcpstat;	/* tcp statistics */
extern	int xtcps_mssdflt;	/* XXX */
extern	int xtcps_delack_enabled;
extern	int xtcps_do_newreno;
extern	int xtcps_path_mtu_discovery;
extern	int xtcps_ss_fltsz;
extern	int xtcps_ss_fltsz_local;


void	 xtcps_canceltimers __P((struct xtcpcb *));
struct xtcpcb *
	 xtcps_close __P((struct xtcpcb *));
/*void	 xtcp_ctlinput __P((int, struct sockaddr *, void *));*/

int	xtcps_ctlinput __P((register struct socket *, int, struct tfi_protosw *));

int	 xtcps_ctloutput __P((struct socket *, struct sockopt *, struct tfi_protosw *));
struct xtcpcb *
	 xtcps_drop __P((struct xtcpcb *, int));
int	 xtcps_drain __P((struct tfi_protosw *));
void	 xtcps_fasttimo __P((void));
struct rmxp_tao *
	 xtcps_gettaocache __P((struct in_conninfo *));
int	 xtcps_init __P((struct tfi_protosw *));

/* AJW: Prototype modified for TFC */
int	 xtcps_input __P((struct mbuf *, int, int,struct socket*, struct tfi_protosw *));

void	 xtcps_mss __P((struct xtcpcb *, int));
int	 xtcps_mssopt __P((struct xtcpcb *));
void	 xtcps_drop_syn_sent __P((struct inpcb *, int));
void	 xtcps_mtudisc __P((struct inpcb *, int));
struct xtcpcb *
	 xtcps_newtcpcb __P((struct inpcb *));

/* AJW addition for TFC */
void init_sack_xtcpcb (struct socket*);

int	 xtcps_output __P((struct xtcpcb *));
void	 xtcps_quench __P((struct inpcb *, int));
void	 xtcps_respond __P((struct xtcpcb *, void *,
	    struct xtcphdr *, struct mbuf *, tcp_seq, tcp_seq, int));
struct rtentry *
	 xtcps_rtlookup __P((struct in_conninfo *));
void	 xtcps_setpersist __P((struct xtcpcb *));
int	 xtcps_slowtimo __P((struct tfi_protosw *));
struct xtcptemp *
	 xtcps_maketemplate __P((struct xtcpcb *));
void	 xtcps_fillheaders __P((struct xtcpcb *, void *, void *));
struct xtcpcb *
	 xtcps_timers __P((struct xtcpcb *, int));
void	 xtcps_trace __P((int, int, struct tcpcb *, void *, struct tcphdr *,
			int));
void	 xtcps_xmit_bandwidth_limit(struct xtcpcb *tp, tcp_seq ack_seq);
void	 xtcps_syncache_init(void);
void	 xtcps_syncache_unreach(struct in_conninfo *, struct xtcphdr *);
int	 xtcps_syncache_expand(struct in_conninfo *, struct xtcphdr *,
	     struct socket **, struct mbuf *);
int	 xtcps_syncache_add(struct in_conninfo *, struct tcpopt *,
	     struct xtcphdr *, struct socket **, struct mbuf *,u_char);
void	 xtcps_syncache_chkrst(struct in_conninfo *, struct xtcphdr *);
void	 xtcps_syncache_badack(struct in_conninfo *);

int	 xtcps_sack_option __P((struct xtcpcb *,struct xtcphdr *,u_char *,int));
void	 xtcps_update_sack_list __P((struct xtcpcb *tp));
void	 xtcps_del_sackholes __P((struct xtcpcb *, struct xtcphdr *));
void	 xtcps_clean_sackreport __P((struct xtcpcb *tp));
void	 xtcps_sack_adjust __P((struct xtcpcb *tp));
struct sackhole *
	 xtcps_sack_output __P((struct xtcpcb *tp));
int	 xtcps_sack_partialack __P((struct xtcpcb *, struct xtcphdr *));
u_long	 xtcps_seq_subtract  __P((u_long, u_long ));

/* Changed by AJW to work with TFC */
/*extern	struct pr_usrreqs tcp_usrreqs;*/
extern	struct tfi_usrreqs xtcps_usrreqs;

extern	u_long xtcps_sendspace;
extern	u_long xtcps_recvspace;
tcp_seq xtcps_new_isn __P((struct xtcpcb *));

extern int xtcps_rfc1323;
extern int xtcps_rfc1644;
extern int xtcps_v6mssdflt;

/* from tcp_seq.h */
extern tcp_cc	xtcps_ccgen;		/* global connection count */

/* not used:
	{ "pcblist", CTLTYPE_STRUCT }, \
	{ "stats", CTLTYPE_STRUCT }, \
	*/

#ifdef TFC_NEGOTIATE
extern int xtcps_syncache_socket(struct tfi_syncache *, struct socket *,
			   struct socket **, struct tfi_protosw_cyc *psw);
extern struct tfi_proto xtcps_pr;
#endif

#endif /* _NETINET_TCP_VAR_H_ */
