/*
 * source.c
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * for any purpose with or without fee is hereby granted, provided
 * that the above copyright notice and this permission notice
 * appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE UNIVERSITY OF UTAH
 * DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING
 * ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO
 * EVENT SHALL THE UNIVERSITY OF UTAH BE LIABLE FOR ANY SPECIAL,
 * DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER
 * IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
 * THIS SOFTWARE.
 */
/* fpont 12/99 */
/* pont.net    */
/* tcpClient.c */

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <stdio.h>
#include <unistd.h> /* close */

/*#define IPPROTO_XTCP 114*/
#define TFC_ATTACH_PROTOCOL 1
#define XTCP_HASH 170
#define CYC_XTCP_HASH 180

#define SERVER_PORT 1505
#define SERVER_ADDRESS 0x91db5f80   /* 128.95.219.145 */
#define MAX_MSG 100

#define SERVER_NAME "192.168.1.3"

#define MAX_CONNECTIONS 512

#if defined(XTCP) || defined(CYC_XTCP)
#define PROTOCOL 114
#else
#define PROTOCOL 0
#endif

#ifdef REPORT
#include <sys/signal.h>
#include <sys/sysctl.h>
#include <sys/time.h>
int hz;

/*
 * Return the frequency of the kernel's statistics clock.
 */
static void
getstathz(void)
{
	struct clockinfo clockrate;
	int mib[2];
	size_t size;

	mib[0] = CTL_KERN;
	mib[1] = KERN_CLOCKRATE;
	size = sizeof clockrate;
	if (sysctl(mib, 2, &clockrate, &size, NULL, 0) == -1)
		err(1, "sysctl kern.clockrate");
	hz = clockrate.stathz;
}

static void
alarm_handler(int sig)
{
	static struct timeval lastnow;
	static unsigned long lastticks[5];
	struct timeval now, delta;
	unsigned long ticks[5], deltaticks[5];
	int rv;
	size_t bsize;

	gettimeofday(&now, 0);
	bsize = sizeof(ticks);
	rv = sysctlbyname("kern.cp_time", ticks, &bsize, 0, 0);
	if (rv == 0 && bsize == sizeof(ticks)) {
		for (rv = 0; rv < 5; rv++) {
			deltaticks[rv] = ticks[rv] - lastticks[rv];
			lastticks[rv] = ticks[rv];
		}
	}
	if (lastnow.tv_usec > now.tv_usec) {
		now.tv_usec += 1000000;
		now.tv_sec--;
	}
	delta.tv_usec = now.tv_usec - lastnow.tv_usec;
	delta.tv_sec = now.tv_sec - lastnow.tv_sec;
	lastnow = now;

	printf("%d.%03d: user %.3f sys %.3f idle %.3f\n",
	       delta.tv_sec, delta.tv_usec/1000,
	       (double)(deltaticks[0]+deltaticks[1]) / hz,
	       (double)(deltaticks[2]+deltaticks[3]) / hz,
	       (double)deltaticks[4] / hz);
}

static void
settimer(int sec)
{
	struct itimerval tv;

	if (sec == 0)
		return;

	tv.it_interval.tv_sec = sec;
	tv.it_interval.tv_usec = 0;
	tv.it_value = tv.it_interval;
	signal(SIGALRM, alarm_handler);
	setitimer(ITIMER_REAL, &tv, 0);
}

#endif

int main (int argc, char *argv[]) {

  int i;
  short port;
  int numConns;

  char data [4096];
  int cycles = 0;
  int bytes = 0;

  int socketDescrips[MAX_CONNECTIONS];

#ifdef XTCP
  int pval = XTCP_HASH;
#endif

#ifdef CYC_XTCP
  int pval = CYC_XTCP_HASH;
#endif

  struct sockaddr_in localAddr, servAddr;
  /*  struct hostent *h;*/

  printf("starting source!\n");

  /*
  if (argc > 1) {
    port = (short) atoi(argv[1]);
  }
  else
  */
  port = SERVER_PORT;

  if (argc > 1)
    numConns = atoi(argv[1]);
  else numConns = 1;

  servAddr.sin_family = AF_INET;
  servAddr.sin_addr.s_addr = inet_addr(SERVER_NAME);
  servAddr.sin_port = htons(port);

  printf("opening %d conns to server: %s   port: %d\n",numConns,
	 SERVER_NAME,port);

  for (i=0;i<numConns;i++) {
    int sd, rc;

    printf("%d ..",i);

    /* create socket */
    sd = socket(AF_INET, SOCK_STREAM, PROTOCOL);
    if(sd<0) {
      perror("cannot open socket ");
      exit(1);
    }

    socketDescrips[i] = sd;
  
#if defined(XTCP) || defined (CYC_XTCP)
  
    if (setsockopt(sd, PROTOCOL, TFC_ATTACH_PROTOCOL, &pval, sizeof pval)
	== -1) {
      perror("setsockopt");
      exit(1);
    }
    printf("sockopt...\n");
  
#endif
  
    /* bind any port number */
    localAddr.sin_family = AF_INET;
    localAddr.sin_addr.s_addr = htonl(INADDR_ANY);
    localAddr.sin_port = htons(0);
  
    rc = bind(sd, (struct sockaddr *) &localAddr, sizeof(localAddr));
    if(rc<0) {
      printf("%s: cannot bind port TCP %u\n",argv[0],SERVER_PORT);
      perror("error ");
      exit(1);
    }
    printf("bind successful...\n ");
  
    /* connect to server */
    rc = connect(sd, (struct sockaddr *) &servAddr, sizeof(servAddr));
    if(rc<0) {
      perror("cannot connect ");
      exit(1);
    }
  }
  
  printf("\nConnections established\n");
#ifdef REPORT
  getstathz();
  settimer(REPORT);
#endif

  while (1) {
    int rc;

    for (i=0; i < numConns; i++) {     
      rc = send(socketDescrips[i], data, 4096, 0);
    
      if(rc<0) {
	perror("cannot send data ");
	exit(1);    
      }     

      bytes+= rc;
      cycles++;
      if ( (cycles % (64*1024)) == 0)
	printf("bytes: %d\n",bytes);
    }
  }

  /*close (sd);      */
  return 0;
}
